unit CipherMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, SecUtils, SecComp;

type
  TfrmCipherMain = class(TForm)
    rgCipherAlg: TRadioGroup;
    rgCipherMode: TRadioGroup;
    Label1: TLabel;
    rgImplementation: TRadioGroup;
    edtIVector: TEdit;
    Label2: TLabel;
    edtKey: TEdit;
    Label3: TLabel;
    edtPlainText: TEdit;
    Label4: TLabel;
    edtCipherText: TEdit;
    Label5: TLabel;
    edtDecrPlainText: TEdit;
    SsAES1: TSsAES;
    SsTwoFish1: TSsTwoFish;
    SsDES1: TSsDES;
    SsBlowFish1: TSsBlowFish;
    SsRC41: TSsRC4;
    btnRandomKey: TButton;
    btnEncrypt: TButton;
    btnDecrypt: TButton;
    procedure rgCipherAlgClick(Sender: TObject);
    procedure rgCipherModeClick(Sender: TObject);
    procedure btnRandomKeyClick(Sender: TObject);
    procedure rgImplementationClick(Sender: TObject);
    procedure btnEncryptClick(Sender: TObject);
    procedure btnDecryptClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    procedure DisableIVectorField;
    procedure EnableIVectorField;
    function GetIVLength: Integer;
    function GetKeyLength: Integer;
    function GetCipherAlg: TCipherAlg;
    function GetCipherMode: TCipherMode;
  public
    procedure EncryptUsingComponent;
    procedure EncryptUsingClass;
    procedure EncryptUsingRoutine;      
    procedure DecryptUsingComponent;
    procedure DecryptUsingClass;
    procedure DecryptUsingRoutine;
    property CipherAlg: TCipherAlg read GetCipherAlg;
    property CipherMode: TCipherMode read GetCipherMode;
    property KeyLength: Integer read GetKeyLength;
    property IVLength: Integer read GetIVLength;
  end;

var
  frmCipherMain: TfrmCipherMain;

implementation

uses
  MpYarrow, Asn1;

{$R *.dfm}

function TfrmCipherMain.GetCipherAlg: TCipherAlg;
begin
  Result := caUnknown;
  case rgCipherAlg.ItemIndex of
    0: Result := caRijndael;
    1: Result := caTwoFish;
    2: Result := caDES;
    3: Result := caBlowFish;
    4: Result := caARC4;
  end;
end;

function TfrmCipherMain.GetCipherMode: TCipherMode;
begin
  case rgCipherMode.ItemIndex of
    1: Result := cmCBC;
    2: Result := cmCFB;
    3: Result := cmOFB;
    4: Result := cmCTR;
    5: Result := cmABC;
  else
    Result := cmECB;
  end;
end;

function TfrmCipherMain.GetIVLength: Integer;
var
  CipherAlg: SecUtils.TCipherAlg;
  CipherMode: SecUtils.TCipherMode;
  CC: TCipherClass;
begin
  Result := 0;
  CipherAlg := GetCipherAlg;
  if CipherAlg = caARC4 then
    Exit
  else if CipherAlg = caUnknown then begin
    Result := -1;
    Exit;
  end;
  CipherMode := GetCipherMode;
  if CipherMode <> cmECB then begin
    CC := FindCipherClass(CipherAlg,CipherMode);
    if CC = nil then
      Result := -1
    else
      Result := TBlockCipherClass(CC).BlockVectorSize * CC.BlockSize;
  end;
end;

function TfrmCipherMain.GetKeyLength: Integer;
begin
  case CipherAlg of
    caRijndael: Result := 16; // 128 bit
    caTwoFish:  Result := 16; // 128 bit
    caDES:      Result := 16; // 2 key 3DES-EDE
    caBlowFish: Result := 16; // 128 bit
    caARC4:     Result := 16; // 128 bit
  else
    Result := -1;
  end;
end;

procedure TfrmCipherMain.rgCipherAlgClick(Sender: TObject);
begin
  if CipherAlg = caARC4 then begin
    rgCipherMode.Enabled := False;
    rgCipherMode.ItemIndex := 3;
    DisableIVectorField;
  end else begin
    rgCipherMode.Enabled := True;
    rgImplementationClick(Sender);
  end;
end;

procedure TfrmCipherMain.rgCipherModeClick(Sender: TObject);
begin
  rgImplementationClick(Sender);
end;

procedure TfrmCipherMain.rgImplementationClick(Sender: TObject);
begin
  if CipherMode = cmECB then
    DisableIVectorField
  else
    EnableIVectorField;
end;

procedure TfrmCipherMain.btnRandomKeyClick(Sender: TObject);
var
  Len: Integer;
  IVector, Key: string;
begin
  Len := IVLength;
  if Len > 0 then begin
    if edtIVector.Enabled then begin
      SetLength(IVector,Len);
      MpYarrow.RawRandom(Pointer(IVector)^,Len*8)
    end else
      IVector := StringOfChar(#0,Len);
    edtIVector.Text := OSToHex(IVector);
  end;
  Len := KeyLength;
  SetLength(Key,Len);
  MpYarrow.RawRandom(Pointer(Key)^,Len*8);
  edtKey.Text := OSToHex(Key);
end;

procedure TfrmCipherMain.DisableIVectorField;
begin
  edtIVector.Text := '';
  edtIVector.Enabled := False;
end;

procedure TfrmCipherMain.EnableIVectorField;
begin
  edtIVector.Enabled := True;
end;

procedure TfrmCipherMain.EncryptUsingClass;
var
  CC: TCipherClass;
  C: TCipher;
  Key, IVector, CipherText: string;
begin
  CC := FindCipherClass(CipherAlg,CipherMode);
  Assert(Assigned(CC));
  Key := HexToOS(edtKey.Text);
  if edtIVector.Text <> '' then
    IVector := HexToOS(edtIVector.Text)
  else
    IVector := '';
  C := CC.Create(Pointer(Key)^,Length(Key),0);
  try
    if IVector <> '' then
      C.IVector := IVector;
    CipherText := C.EncryptStr(edtPlainText.Text);
  finally
    C.Free;
  end;
  edtCipherText.Text := OSToHex(CipherText);
end;

procedure TfrmCipherMain.EncryptUsingComponent;
var
  SecComp: TSsCustomCipherComponent;
  IVector, CipherText: string;
begin
  case CipherAlg of
    caRijndael:
      begin
        SecComp := SsAES1;
        SsAES1.Mode := CipherMode;
      end;
    caTwoFish:
      begin
        SecComp := SsTwoFish1;
        SsTwoFish1.Mode := CipherMode;
      end;
    caDES:
      begin
        SecComp := SsDES1;
        SsDES1.Mode := CipherMode;
      end;
    caBlowFish:
      begin
        SecComp := SsBlowFish1;
        SsBlowFish1.Mode := CipherMode;
      end;
    caARC4:     SecComp := SsRC41;
  else
    SecComp := nil;
  end;
  Assert(Assigned(SecComp));
  if edtIVector.Text <> '' then
    IVector := HexToOS(edtIVector.Text)
  else
    IVector := '';
  SecComp.SetKeyStr(edtKey.Text,sfHEX);
  if IVector <> '' then
    SecComp.IVector := IVector;
  CipherText := SecComp.EncryptStr(edtPlainText.Text);
  SecComp.CleanUp;
  edtCipherText.Text := OSToHex(CipherText);
end;

procedure TfrmCipherMain.EncryptUsingRoutine;
var
  Key, IVector, CipherText: string;
  CC: TCipherClass;
begin
  Key := HexToOS(edtKey.Text);
  if edtIVector.Text <> '' then begin
    IVector := HexToOS(edtIVector.Text);
    CC := FindCipherCLass(CipherAlg,CipherMode);
    CipherText := EncryptString(CC,Key,IVector,edtPlainText.Text);
  end else
    CipherText := EncryptString(CipherAlg,CipherMode,Key,edtPlainText.Text);
  edtCipherText.Text := OSToHex(CipherText);
end;

procedure TfrmCipherMain.btnEncryptClick(Sender: TObject);
begin
  case rgImplementation.ItemIndex of
    0: EncryptUsingComponent;
    1: EncryptUsingClass;
    2: EncryptUsingRoutine;
  end;
end;

procedure TfrmCipherMain.DecryptUsingClass;
var
  CC: TCipherClass;
  C: TCipher;
  Key, IVector, CipherText: string;
begin
  CC := FindCipherClass(CipherAlg,CipherMode);
  Assert(Assigned(CC));
  Key := HexToOS(edtKey.Text);
  if edtIVector.Text <> '' then
    IVector := HexToOS(edtIVector.Text)
  else
    IVector := '';
  CipherText := HexToOS(edtCipherText.Text);
  C := CC.Create(Pointer(Key)^,Length(Key),0);
  try
    if IVector <> '' then
      C.IVector := IVector;
    edtDecrPlainText.Text := C.DecryptStr(CipherText);
  finally
    C.Free;
  end;
end;

procedure TfrmCipherMain.DecryptUsingComponent;
var
  SecComp: TSsCustomCipherComponent;
  IVector, CipherText: string;
begin
  case CipherAlg of
    caRijndael:
      begin
        SecComp := SsAES1;
        SsAES1.Mode := CipherMode;
      end;
    caTwoFish:
      begin
        SecComp := SsTwoFish1;
        SsTwoFish1.Mode := CipherMode;
      end;
    caDES:
      begin
        SecComp := SsDES1;
        SsDES1.Mode := CipherMode;
      end;
    caBlowFish:
      begin
        SecComp := SsBlowFish1;
        SsBlowFish1.Mode := CipherMode;
      end;
    caARC4:     SecComp := SsRC41;
  else
    SecComp := nil;
  end;
  Assert(Assigned(SecComp));
  if edtIVector.Text <> '' then
    IVector := HexToOS(edtIVector.Text)
  else
    IVector := '';
  SecComp.SetKeyStr(edtKey.Text,sfHEX);
  if IVector <> '' then
    SecComp.IVector := IVector;
  CipherText := HexToOS(edtCipherText.Text);
  edtDecrPlainText.Text := SecComp.DecryptStr(CipherText);
  SecComp.CleanUp;
end;

procedure TfrmCipherMain.DecryptUsingRoutine;
var
  Key, IVector, CipherText: string;
  CC: TCipherClass;
begin
  Key := HexToOS(edtKey.Text);
  CipherText := HexToOS(edtCipherText.Text);
  if edtIVector.Text <> '' then begin
    IVector := HexToOS(edtIVector.Text);
    CC := FindCipherClass(CipherAlg,CipherMode);
    edtDecrPlainText.Text := DecryptString(CC,Key,IVector,CipherText);
  end else
    edtDecrPlainText.Text := DecryptString(CipherAlg,CipherMode,Key,CipherText);
end;

procedure TfrmCipherMain.btnDecryptClick(Sender: TObject);
begin
  case rgImplementation.ItemIndex of
    0: DecryptUsingComponent;
    1: DecryptUsingClass;
    2: DecryptUsingRoutine;
  end;
end;

procedure TfrmCipherMain.FormCreate(Sender: TObject);
begin
  btnRandomKeyClick(Sender);
end;

end.
