unit PFXLoaderMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, CertificateSelection, Pkix_Cert,
  Asn1;

type
  TForm1 = class(TForm)
    CertificateSelection1: TCertificateSelection;
    btnLoad: TButton;
    OpenDialog1: TOpenDialog;
    memToBeSigned: TMemo;
    memSignature: TMemo;
    btnSign: TButton;
    btnVerify: TButton;
    Label1: TLabel;
    Label2: TLabel;
    memDigest: TMemo;
    chbDigest: TCheckBox;
    procedure btnLoadClick(Sender: TObject);
    procedure CertificateSelection1CertChange(Sender: TObject;
      Cert: TCertificate);
    procedure btnSignClick(Sender: TObject);
    procedure btnVerifyClick(Sender: TObject);
  private
    procedure Sign(const AMessage: string; var ASignature: OctetString);
    procedure DigestAndSign(const AMessage: string;
                              var ADigest,
                                  ASignature: OctetString);
    function DigestAndVerify(const AMessage: string;
                             const ADigest,
                                   ASignature: OctetString): Boolean;
    function Verify(const AMessage: string;
                    const ASignature: OctetString): Boolean;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses
  SecUtils, MpPK, MpIF,
  CertModule;

{$R *.dfm}

procedure TForm1.btnLoadClick(Sender: TObject);
var
  PW: ISecretKey;
begin
  if OpenDialog1.Execute then begin
    PW := nil;
    dmCerts.KeysAndCerts.ImportFromPFX(OpenDialog1.FileName,PW);
    CertificateSelection1.CertSource := dmCerts.KeysAndCerts;
    CertificateSelection1.GroupIndex := 1;
    CertificateSelection1.DisplayAll;
  end;
end;

procedure TForm1.CertificateSelection1CertChange(Sender: TObject;
  Cert: TCertificate);
begin
  btnSign.Enabled := Assigned(CertificateSelection1.SelectedCert);
end;

procedure TForm1.btnSignClick(Sender: TObject);
var
  M, D, S: string;
begin
  M := memToBeSigned.Text;
  D := '';
  if chbDigest.Checked then
    DigestAndSign(M,D,S)
  else
    Sign(M,S);
  memSignature.Text := OSToHex(S);
  if D = '' then
    memDigest.Text := ''
  else
    memDigest.Text := OSToHex(D);
  btnVerify.Enabled := True;
end;

procedure TForm1.btnVerifyClick(Sender: TObject);
var
  M, S: string;
  Res: Boolean;
begin
  M := memToBeSigned.Text;
  S := HexToOS(memSignature.Text);
  if chbDigest.Checked then
    Res := DigestAndVerify(M,HexToOS(memDigest.Text),S)
  else
    Res := Verify(M,S);
  if Res then
    ShowMessage('OK')
  else
    ShowMessage('Fail');
end;

procedure TForm1.DigestAndSign(const AMessage: string;
                                 var ADigest,
                                     ASignature: OctetString);
var
  Cert: TCertificate;
  Priv: TMPPrivateKey;
  Len: Integer;
begin
  Cert := CertificateSelection1.SelectedCert;
  Priv := dmCerts.PrivateKeyRing.FindCreatePrivateKey(Cert.Data);
  if Priv = nil then
    raise Exception.Create('No private key for selected certificate');
  ADigest := DigestString(haSHA1,AMessage);
  SetLength(ASignature,Priv.SignatureLength + 1);
  Len := Priv.SignDig(Pointer(ADigest)^,Length(ADigest),
                      Pointer(ASignature)^,Length(ASignature),
                      haSHA1);
  SetLength(ASignature,Len);
end;

function TForm1.DigestAndVerify(const AMessage: string;
                                const ADigest,
                                      ASignature: OctetString): Boolean;
var
  Cert: TCertificate;
  Publ: TMPPublicKey;
begin
  Cert := CertificateSelection1.SelectedCert;
  Publ := TMPIFPublicKey.Create(Cert);
  try
    Result := ADigest = DigestString(haSHA1,AMessage);
    if Result then
      Result := Publ.VerDig(Pointer(ADigest)^,Length(ADigest),
                            Pointer(ASignature)^,Length(ASignature));
  finally
    Publ.Free;
  end;
end;

procedure TForm1.Sign(const AMessage: string; var ASignature: OctetString);
var
  Cert: TCertificate;
  Priv: TMPPrivateKey;
  Len: Integer;
begin
  Cert := CertificateSelection1.SelectedCert;
  Priv := dmCerts.PrivateKeyRing.FindCreatePrivateKey(Cert.Data);
  if Priv = nil then
    raise Exception.Create('No private key for selected certificate');
  SetLength(ASignature,Priv.SignatureLength + 1);
  Len := Priv.SignBuf(Pointer(AMessage)^,Length(AMessage),
                      Pointer(ASignature)^,Length(ASignature));
  SetLength(ASignature,Len);
end;

function TForm1.Verify(const AMessage: string;
                       const ASignature: OctetString): Boolean;
var
  Cert: TCertificate;
  Publ: TMPPublicKey;
begin
  Cert := CertificateSelection1.SelectedCert;
  Publ := TMPIFPublicKey.Create(Cert);
  try
    Result := Publ.VerBuf(Pointer(AMessage)^,Length(AMessage),
                          Pointer(ASignature)^,Length(ASignature));
  finally
    Publ.Free;
  end;
end;

end.
