unit CAPICOMMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdActns, ActnList, Menus, ComCtrls, SecComp, StreamSecII,
  MpX509, TlsInternalServer;

type
  TfrmCAPICOM = class(TForm)
    ListView: TListView;
    ActionList1: TActionList;
    MainMenu1: TMainMenu;
    actRefresh: TAction;
    actExtract: TAction;
    actFileSaveAs: TFileSaveAs;
    actFileOpen: TFileOpen;
    File1: TMenuItem;
    Open1: TMenuItem;
    SaveAs1: TMenuItem;
    Caption1: TMenuItem;
    Refresh1: TMenuItem;
    Extract1: TMenuItem;
    PrivateKeyRing: TSsPrivateKeyRingComponent;
    KeysAndCerts: TSimpleTLSInternalServer;
    procedure actRefreshExecute(Sender: TObject);
    procedure actExtractExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure actExtractUpdate(Sender: TObject);
    procedure actFileOpenAccept(Sender: TObject);
    procedure actFileSaveAsAccept(Sender: TObject);
  private                                 
    function GetCertificateEnc64(Index: Integer): String;
    function GetPKContainerName(Index: Integer): string;
    function GetPKProviderName(Index: Integer): string;
    function GetPKProviderType(Index: Integer): LongWord;
  public
    { Public declarations }
  end;

var
  frmCAPICOM: TfrmCAPICOM;

implementation

uses
  CAPICOM_TLB,
  SsBase64, SecUtils, Asn1, X509Base, MpIF, CAPIExtKey;

const
  CRYPT_VERIFYCONTEXT = 0;

{$R *.dfm}

procedure TfrmCAPICOM.actRefreshExecute(Sender: TObject);
var
  CertStore: IStore;
  i: Integer;
  ListItem: TListItem;
begin
  CertStore := CoStore.Create;
  CertStore.Open(CAPICOM_CURRENT_USER_STORE, 'My', CAPICOM_STORE_OPEN_READ_ONLY);
  ListView.Items.Clear;
  for i := 1 to CertStore.Certificates.Count do begin
    ListItem := ListView.Items.Add;
    ListItem.Data := Pointer(i);
    ListItem.Caption := CertStore.Certificates.Item[i].GetInfo(CAPICOM_CERT_INFO_SUBJECT_SIMPLE_NAME);
    ListItem.SubItems.Add(CertStore.Certificates.Item[i].GetInfo(CAPICOM_CERT_INFO_ISSUER_SIMPLE_NAME));
    ListItem.SubItems.Add(FormatDateTime('dd/mm/yyyy', CertStore.Certificates.Item[i].ValidToDate ));
    ListItem.SubItems.Add(CertStore.Certificates.Item[i].GetInfo(CAPICOM_CERT_INFO_SUBJECT_EMAIL_NAME));
    if CertStore.Certificates.Item[i].IsValid then
      ListItem.SubItems.Add('Yes')
    else
      ListItem.SubItems.Add('No');
  end;
end;

procedure TfrmCAPICOM.actExtractExecute(Sender: TObject);
var
  CSP: TCAPIIFPrivateKey;
  PrivKey: TIFPrivateKey;
  ListItem: TListItem;
  SelectedCert: Integer;
  SS: TStringStream;
  Cert: TASN1Struct;
  PKI: OctetString;
begin
  FillChar(PrivKey,SizeOf(PrivKey),0);
  try
    SelectedCert := Integer(ListView.Items[ListView.ItemIndex].Data);
    CSP := TCAPIIFPrivateKey.Create(True);
    try
      if CSP.AcquireContext(GetPKContainerName(SelectedCert),
                            GetPKProviderName(SelectedCert),
                            GetPKProviderType(SelectedCert),
                            CRYPT_VERIFYCONTEXT) then begin
        CSP.GetUserKey(1);
        CSP.RetrPrivKey(PrivKey)
      end;
    finally
      CSP.Free;
    end;
    SS := TStringStream.Create(MIME64ToStr(GetCertificateEnc64(SelectedCert)));
    try
      Cert := TASN1Struct.Create;
      try
        Cert.LoadFromStream(SS,fmtDER);
        PKI := ExtractSubjectKeyIdentifier(Cert);
        PrivateKeyRing.AddPrivateRSAKey(PrivKey,Pointer(PKI),Length(PKI),True);
        if Assigned(PrivateKeyRing.FindCreatePrivateKey(Cert)) then
          ShowMessage('Success')
        else
          ShowMessage('Failure');
      finally
        Cert.Free;
      end;
    finally
      SS.Free;
    end;
  finally
    DisposeIFPrivateKey(PrivKey);
  end;
end;

function TfrmCAPICOM.GetPKContainerName(Index: Integer): string;
var
  CertStore: IStore;
begin
  CertStore := CoStore.Create;
  CertStore.Open(CAPICOM_CURRENT_USER_STORE, 'My', CAPICOM_STORE_OPEN_READ_ONLY);
  Result := CertStore.Certificates.Item[Index].PrivateKey.ContainerName;
end;

function TfrmCAPICOM.GetPKProviderName(Index: Integer): string;
var
  CertStore: IStore;
begin
  CertStore := CoStore.Create;
  CertStore.Open(CAPICOM_CURRENT_USER_STORE, 'My', CAPICOM_STORE_OPEN_READ_ONLY);
  Result := CertStore.Certificates.Item[Index].PrivateKey.ProviderName;
end;

function TfrmCAPICOM.GetPKProviderType(Index: Integer): LongWord;
var
  CertStore: IStore;
begin
  CertStore := CoStore.Create;
  CertStore.Open(CAPICOM_CURRENT_USER_STORE, 'My', CAPICOM_STORE_OPEN_READ_ONLY);
  Result := CertStore.Certificates.Item[Index].PrivateKey.ProviderType;
end;

function TfrmCAPICOM.GetCertificateEnc64(Index: Integer): String;
var
  CertStore: IStore;
begin
  CertStore := CoStore.Create;
  CertStore.Open(CAPICOM_CURRENT_USER_STORE, 'My', CAPICOM_STORE_OPEN_READ_ONLY);
  if CertStore.Certificates.Item[Index].HasPrivateKey then
    Result := CertStore.Certificates.Item[Index].Export(CAPICOM_ENCODE_BASE64)
  else
    raise Exception.Create('The certificate doesnt have a private key associated with it.');
end;

procedure TfrmCAPICOM.FormCreate(Sender: TObject);
begin
  actRefresh.Execute;
end;

procedure TfrmCAPICOM.actExtractUpdate(Sender: TObject);
begin
  actExtract.Enabled := Assigned(ListView.Selected); 
end;

procedure TfrmCAPICOM.actFileOpenAccept(Sender: TObject);
var
  Pwd: ISecretKey;
begin
  Pwd := nil;
  PrivateKeyRing.LoadPrivateKeyRingFromFile(actFileOpen.Dialog.FileName,Pwd);
  actFileSaveAs.Dialog.FileName := ExtractFileName(actFileOpen.Dialog.FileName);
  actFileSaveAs.Dialog.InitialDir := ExtractFilePath(actFileOpen.Dialog.FileName);
end;

procedure TfrmCAPICOM.actFileSaveAsAccept(Sender: TObject);
begin
  KeysAndCerts.SavePrivateKeyRing(actFileSaveAs.Dialog.FileName);
end;

end.
