unit CMSDemoMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, ComCtrls, ExtCtrls;

type
  TFrmSign = class(TForm)
    OpenDialog1: TOpenDialog;
    Button2: TButton;
    Button4: TButton;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    SpeedButton3: TSpeedButton;
    Label3: TLabel;
    Label4: TLabel;
    memMessage: TMemo;
    Edit3: TEdit;
    Edit4: TEdit;
    SaveDialog1: TSaveDialog;
    chbDetachedMessage: TCheckBox;
    PageControl2: TPageControl;
    TabSheet3: TTabSheet;
    TabSheet4: TTabSheet;
    SpeedButton1: TSpeedButton;
    Label1: TLabel;
    SpeedButton2: TSpeedButton;
    Label2: TLabel;
    Label5: TLabel;
    SpeedButton5: TSpeedButton;
    edtMyCerts: TEdit;
    edtPrivateKeyRing: TEdit;
    edtRootCerts: TEdit;
    Label7: TLabel;
    edtCollectedCerts: TEdit;
    SpeedButton6: TSpeedButton;
    Label6: TLabel;
    edtSigningDescription: TEdit;
    chbIncludeRootCerts: TCheckBox;
    rgAction: TRadioGroup;
    Label8: TLabel;
    edtExtractedFile: TEdit;
    TabSheet5: TTabSheet;
    Label9: TLabel;
    Edit1: TEdit;
    SpeedButton4: TSpeedButton;
    procedure SpeedButton1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure SpeedButton2Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure SpeedButton3Click(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure chbDetachedMessageClick(Sender: TObject);
    procedure Edit3Change(Sender: TObject);
    procedure SpeedButton5Click(Sender: TObject);
    procedure SpeedButton6Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure rgActionClick(Sender: TObject);
    procedure SpeedButton4Click(Sender: TObject);
  private
    { Private declarations }
    function ProcessData(InStream, OutStream: TStream): Boolean;
    procedure LoadCerts();
  public
    { Public declarations }
  end;

var
  FrmSign: TFrmSign;
  InitialStream, SignedStream: TMemoryStream;

implementation

uses
  IniFiles,
  Cms, MpCms, CMSDemoDM;

{$R *.dfm}

const
  EXTATTACHEDDATA: string = '.p7m';
  EXTDETACHEDDATA: string = '.p7s';

procedure TFrmSign.LoadCerts();
begin
  dmCMSDemo.LoadRootCerts(edtRootCerts.Text);
  dmCMSDemo.LoadCollectedCerts(edtCollectedCerts.Text);
  dmCMSDemo.LoadMyCerts(edtMyCerts.Text);
end;

function TFrmSign.ProcessData(InStream, OutStream: TStream): Boolean;
var
  Actions: TContentTypeArray;
begin
  LoadCerts;
  if not dmCMSDemo.LoadPrivateKeyRing(edtPrivateKeyRing.Text) then
    raise Exception.Create('Unable to load User.pkr');
  case rgAction.ItemIndex of
    0: Actions := MakeContentTypeArray([pSignedData]);
    1: Actions := MakeContentTypeArray([pEnvelopedData]);
    2: Actions := MakeContentTypeArray([pSignedData,pEnvelopedData]);
    3: Actions := MakeContentTypeArray([pEnvelopedData,pSignedData]);
    4: Actions := MakeContentTypeArray([pSignedData,pEnvelopedData,pSignedData]);
  else
    SetLength(Actions,0);
  end;
  Result := dmCMSDemo.ProcessData(Actions,
                                  edtSigningDescription.Text,
                                  InStream,OutStream,
                                  chbIncludeRootCerts.Checked);
end;

procedure TFrmSign.Button2Click(Sender: TObject);
var
  FileDoc, FileSign: TFileStream;
begin
  case PageControl1.ActivePageIndex of
    0:
      begin
        if Assigned(InitialStream) then         
          InitialStream.Clear
        else
          InitialStream := TMemoryStream.Create;
        memMessage.Lines.SaveToStream(InitialStream);
        if Assigned(SignedStream) then
          SignedStream.Clear
        else
          SignedStream := TMemoryStream.Create;
        InitialStream.Position := 0;
        if ProcessData(InitialStream, SignedStream) then
          ShowMessage('Data processed correctly')
      end;
    1:
      begin
        FileDoc := TFileStream.Create(Edit3.Text, fmOpenRead);
        FileSign := TFileStream.Create(Edit4.Text, fmCreate);
        try
          ProcessData(FileDoc, FileSign);
        finally
          FileDoc.Free;
          FileSign.Free;
        end;
      end;
    2:    
      begin        
        LoadCerts;  
        if not dmCMSDemo.LoadPrivateKeyRing(edtPrivateKeyRing.Text) then
          raise Exception.Create('Unable to load User.pkr');
        FileSign := TFileStream.Create(Edit1.Text, fmOpenReadWrite);
        try
          dmCMSDemo.CounterSignData(FileSign,
                                    edtSigningDescription.Text,
                                    chbIncludeRootCerts.Checked);
        finally
          FileSign.Free;
        end;
      end;
  end;
end;

procedure TFrmSign.Button4Click(Sender: TObject);
var
  FileSign, FileExtracted: TFileStream;
  Actions: TStringList;
  Msg: string;
  DataStream: TStringStream;
begin
  Msg := '(no action)';
  LoadCerts;
  Actions := TStringList.Create;
  try
    case PageControl1.ActivePageIndex of
      0:
        begin
          DataStream := TStringStream.Create('');
          try
            SignedStream.Position := 0;
            if dmCMSDemo.DecodeData(SignedStream,DataStream,Actions) then
              Msg := 'Data verified correctly'#13#10 + Actions.Text
            else
              Msg := 'Failed verification!';
            memMessage.Text := DataStream.DataString;
          finally
            DataStream.Free;
          end;
        end;
      1:
        begin
          FileSign := TFileStream.Create(Edit4.Text, fmOpenRead);
          try
            if not Self.chbDetachedMessage.Checked then begin
              FileExtracted := TFileStream.Create(edtExtractedFile.Text, fmCreate);
              try
                if dmCMSDemo.DecodeData(FileSign,FileExtracted,Actions) then
                  Msg := 'File verified correctly'#13#10 + Actions.Text
                else
                  Msg := 'Failed verification!';
              finally
                FileExtracted.Free;
              end;
            end else begin
              FileExtracted := TFileStream.Create(Edit3.Text, fmOpenRead);
              try
                if dmCMSDemo.VerifyData(FileSign,FileExtracted,Actions) then
                  Msg := 'File verified correctly'#13#10 + Actions.Text
                else
                  Msg := 'Failed verification!';
              finally
                FileExtracted.Free;
              end;
            end;
          finally
            FileSign.Free;
          end;
        end;
      2:
        begin
          FileSign := TFileStream.Create(Edit1.Text, fmOpenRead);
          try
            if dmCMSDemo.VerifyData(FileSign,nil,Actions) then
              Msg := 'File verified correctly'#13#10 + Actions.Text
            else
              Msg := 'Failed verification!';
          finally
            FileSign.Free;
          end;
        end;
    end;
  finally
    Action.Free;
  end;
  ShowMessage(Msg);
end;

procedure TFrmSign.SpeedButton1Click(Sender: TObject);
begin
  OpenDialog1.InitialDir := ExtractFilePath(Application.ExeName);
  OpenDialog1.Filter := 'Signed Certificate List (*.scl)|*.scl';
  if OpenDialog1.Execute then
    edtMyCerts.Text := OpenDialog1.FileName;
end;

procedure TFrmSign.SpeedButton2Click(Sender: TObject);
begin
  OpenDialog1.InitialDir := ExtractFilePath(Application.ExeName);
  OpenDialog1.Filter := 'Private Key Ring (*.pkr)|*.pkr';
  if OpenDialog1.Execute then
    edtPrivateKeyRing.Text := OpenDialog1.FileName;
end;

procedure TFrmSign.SpeedButton3Click(Sender: TObject);
begin
  OpenDialog1.InitialDir := ExtractFilePath(Application.ExeName);
  OpenDialog1.Filter := 'Any file (*.*)|*.*';
  if OpenDialog1.Execute then
    Edit3.Text := OpenDialog1.FileName;
end;

procedure TFrmSign.FormDestroy(Sender: TObject);
begin
  InitialStream.Free;
  SignedStream.Free;
end;

procedure TFrmSign.chbDetachedMessageClick(Sender: TObject);
begin
  if chbDetachedMessage.Checked then
    begin
      dmCMSDemo.SignedData.DetachedSignature := True;
      Edit4.Text := ChangeFileExt(Edit4.Text, EXTDETACHEDDATA);
    end
  else
    begin
      dmCMSDemo.SignedData.DetachedSignature := False;
      Edit4.Text := ChangeFileExt(Edit4.Text, EXTATTACHEDDATA);
    end;
end;

procedure TFrmSign.Edit3Change(Sender: TObject);
var
  Ext: String;
begin
  if chbDetachedMessage.Checked then
    Ext := EXTDETACHEDDATA
  else
    Ext := EXTATTACHEDDATA;
  Edit4.Text := ChangeFileExt(Edit3.Text, Copy(ExtractFileExt(Edit3.Text), 2, length(Edit3.Text))) + Ext;
end;

procedure TFrmSign.SpeedButton5Click(Sender: TObject);
begin
  OpenDialog1.InitialDir := ExtractFilePath(Application.ExeName);
  OpenDialog1.Filter := 'Signed Certificate List (*.scl)|*.scl';
  if OpenDialog1.Execute then
    edtRootCerts.Text := OpenDialog1.FileName;
end;

procedure TFrmSign.SpeedButton6Click(Sender: TObject);
begin
  OpenDialog1.InitialDir := ExtractFilePath(Application.ExeName);
  OpenDialog1.Filter := 'Signed Certificate List (*.scl)|*.scl';
  if OpenDialog1.Execute then
    edtCollectedCerts.Text := OpenDialog1.FileName;
end;

procedure TFrmSign.FormShow(Sender: TObject);
var
  Ini: TIniFile;
begin
  Ini := TIniFile.Create(ExtractFilePath(Application.ExeName) + 'CMSDemo.ini');
  try
    edtMyCerts.Text := Ini.ReadString('Files','MyCerts',edtMyCerts.Text);
    edtRootCerts.Text := Ini.ReadString('Files','RootCerts',edtRootCerts.Text);
    edtCollectedCerts.Text := Ini.ReadString('Files','CollectedCerts',edtCollectedCerts.Text);
    edtPrivateKeyRing.Text := Ini.ReadString('Files','PrivateKeyRing',edtPrivateKeyRing.Text);

    chbDetachedMessage.Checked := Ini.ReadBool('Options','Detached',chbDetachedMessage.Checked);
    chbIncludeRootCerts.Checked := Ini.ReadBool('Options','IncludeRoot',chbIncludeRootCerts.Checked);
    rgAction.ItemIndex := Ini.ReadInteger('Options','Actions',rgAction.ItemIndex);
  finally
    Ini.Free;
  end;
end;

procedure TFrmSign.FormClose(Sender: TObject; var Action: TCloseAction);
var
  Ini: TIniFile;
begin
  Ini := TIniFile.Create(ExtractFilePath(Application.ExeName) + 'CMSDemo.ini');
  try
    Ini.WriteString('Files','MyCerts',edtMyCerts.Text);
    Ini.WriteString('Files','RootCerts',edtRootCerts.Text);
    Ini.WriteString('Files','CollectedCerts',edtCollectedCerts.Text);
    Ini.WriteString('Files','PrivateKeyRing',edtPrivateKeyRing.Text);

    Ini.WriteBool('Options','Detached',chbDetachedMessage.Checked);
    Ini.WriteBool('Options','IncludeRoot',chbIncludeRootCerts.Checked);
    Ini.WriteInteger('Options','Actions',rgAction.ItemIndex);
  finally
    Ini.Free;
  end;
end;

procedure TFrmSign.rgActionClick(Sender: TObject);
begin
  if rgAction.ItemIndex = 0 then
    chbDetachedMessage.Enabled := True
  else begin                          
    chbDetachedMessage.Enabled := False;
    chbDetachedMessage.Checked := False;
  end;
end;

procedure TFrmSign.SpeedButton4Click(Sender: TObject);
begin
  OpenDialog1.InitialDir := ExtractFilePath(Application.ExeName);
  OpenDialog1.Filter := 'PKCS#7 file (*.p7m,*.p7s)|*.p7m;*.p7s';
  if OpenDialog1.Execute then
    Edit1.Text := OpenDialog1.FileName;
end;

end.
