{*******************************************************}
{                                                       }
{     StreamSec Security Library for Borland Delphi     }
{     CertMgrMain Unit                                  }
{
  LICENSE

  Copyright (c) 2004, Henrick Wibell Hellstrm, StreamSec
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of StreamSec nor the names of its contributors may be
      used to endorse or promote products derived from this software without
      specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
{*******************************************************}
{$I ver.inc}
unit CertMgrMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdActns, ActnList, Menus, ComCtrls, StdCtrls, Pkix_Cert, Pkix_CRL,
  Asn1Data, CertificateViewer, ExtCtrls;

type
  TfrmCertMgrMain = class(TForm)
    MainMenu1: TMainMenu;
    ActionList1: TActionList;
    File1: TMenuItem;
    Open1: TMenuItem;
    SaveAs1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    TabControl1: TTabControl;
    ListView1: TListView;
    Actions1: TMenuItem;
    actNewCert: TAction;
    NewCertificate1: TMenuItem;
    PopupMenu1: TPopupMenu;
    actRevoke: TAction;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    RevokeCertificate1: TMenuItem;
    RevokeCertificate2: TMenuItem;
    actOpen: TAction;
    actSaveAs: TAction;
    actExit: TAction;
    SaveDialog2: TSaveDialog;
    actImport: TAction;
    ImportCertificate1: TMenuItem;
    actVerify: TAction;
    actVerify1: TMenuItem;
    actExport: TAction;
    SaveDialog3: TSaveDialog;
    ExportCertificate1: TMenuItem;
    ExportCertificate2: TMenuItem;
    CertificateViewer1: TCertificateViewer;
    actRemove: TAction;
    actRegister: TAction;
    CheckRequests1: TMenuItem;
    RemoveCertificate1: TMenuItem;
    actNewWindow: TAction;
    N2: TMenuItem;
    NewWindow1: TMenuItem;
    actActivate: TAction;
    Activate1: TMenuItem;
    OpenDialog2: TOpenDialog;
    procedure ListView1Data(Sender: TObject; Item: TListItem);
    procedure TabControl1Change(Sender: TObject);
    procedure ListView1SelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure actNewCertExecute(Sender: TObject);
    procedure actRevokeUpdate(Sender: TObject);
    procedure actRevokeExecute(Sender: TObject);
    procedure actExitExecute(Sender: TObject);
    procedure actOpenExecute(Sender: TObject);
    procedure actSaveAsExecute(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure actImportExecute(Sender: TObject);
    procedure actVerifyUpdate(Sender: TObject);
    procedure actVerifyExecute(Sender: TObject);
    procedure actExportUpdate(Sender: TObject);
    procedure actExportExecute(Sender: TObject);
    procedure actRemoveUpdate(Sender: TObject);
    procedure actRemoveExecute(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure SaveDialog1TypeChange(Sender: TObject);
    procedure actRegisterExecute(Sender: TObject);
    procedure actRegisterUpdate(Sender: TObject);
    procedure actNewWindowExecute(Sender: TObject);
    procedure actActivateExecute(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    FChain: TList;
    procedure Display;
  public
    procedure EventActivate(Cert: Pointer; CertLen: Integer);
  end;

var
  frmCertMgrMain: TfrmCertMgrMain;

implementation

uses
  SsBase64, Pkix, Asn1, CertModule, MpArith, X509Base, MpX509,
  SelectStoreDlg, NewCertDlg, RevokationDlg, MpPK, ImportDlg,
  ExportDlg, YarrowModule,
{$IFNDEF NO_REGISTER}
  RegistrationDlg,
{$ENDIF  NO_REGISTER}
  ShellAPI;

{$R *.dfm}

procedure TfrmCertMgrMain.ListView1Data(Sender: TObject; Item: TListItem);
var
{$IFDEF NEVER}
  RDName: TRdnsequence;
  F: PASN1Struct;
{$ELSE}
  RDName: TX501Name;
  Store: TX509TrustedCertificates;
{$ENDIF}
begin
{$IFDEF NEVER}
  RDName := TRdnsequence.Create(nil,nil);
  try
    try
      F := nil;
      case TabControl1.TabIndex of
        0: ExtractSubjectStruct(dmCert.MyCerts.Certs[Item.Index],F);
        1: ExtractSubjectStruct(dmCert.RootCerts.Certs[Item.Index],F);
        2: ExtractSubjectStruct(dmCert.IssuedCerts.Certs[Item.Index],F);
        3: ExtractSubjectStruct(dmCert.CollectedCerts.Certs[Item.Index],F);
        4: ExtractSubjectStruct(FChain[Item.Index],F);
      end;
      RDName.AssignStruct(F^);
    except
    end;
    Item.Caption := RDName.CommonName;
    Item.SubItems.Add(RDName.OrganizationName);
    Item.SubItems.Add(RDName.OrganizationalUnitName);
  finally
    RDName.Free;
  end;
{$ELSE}
  if TabControl1.TabIndex < 4 then begin
    case TabControl1.TabIndex of
      0: Store := dmCert.MyCerts;
      1: Store := dmCert.RootCerts;
      2: Store := dmCert.IssuedCerts;
      3: Store := dmCert.CollectedCerts;
    else
      raise Exception.Create('Internal error');
    end;
    if Store.Count <> ListView1.Items.Count then begin
      ListView1.Items.Count := Store.Count;
      Exit;
    end;
    ExtractSubject(Store.Certs[Item.Index],RDName,False);
  end else begin
    if FChain.Count <> ListView1.Items.Count then begin
      ListView1.Items.Count := FChain.Count;
      Exit;
    end;
    ExtractSubject(FChain[Item.Index],RDName,False);
  end;
  {
  case TabControl1.TabIndex of
    0: Store := dmCert.MyCerts;
    1: Store := dmCert.RootCerts;
    2: Store := dmCert.IssuedCerts;
    3: Store := dmCert.CollectedCerts;
    4:
  end;
  }
  Item.Caption := RDName.commonName.Str;
  Item.SubItems.Add(RDName.OrganizationName.Str);
  Item.SubItems.Add(RDName.OrganizationalUnitName.Str);
{$ENDIF}
end;

procedure TfrmCertMgrMain.TabControl1Change(Sender: TObject);
begin
  case TabControl1.TabIndex of
    0: ListView1.Items.Count := dmCert.MyCerts.Count;
    1: ListView1.Items.Count := dmCert.RootCerts.Count;
    2: ListView1.Items.Count := dmCert.IssuedCerts.Count;
    3: ListView1.Items.Count := dmCert.CollectedCerts.Count;
    4: ListView1.Items.Count := FChain.Count;
  end;
  ListView1.Selected := nil;
  ListView1.Repaint;
end;

procedure TfrmCertMgrMain.ListView1SelectItem(Sender: TObject;
  Item: TListItem; Selected: Boolean);
begin
  if not Selected then Exit;
  if TabControl1.TabIndex < 4 then begin
    dmCert.SelectCert(TabControl1.TabIndex,Item.Index);
    if TabControl1.TabIndex > 0 then
      actActivate.Enabled := True
    else
      actActivate.Enabled := False;
  end else begin
    dmCert.SelectCert(TASN1Struct(FChain[Item.Index]));
    actActivate.Enabled := False;
  end;
  Display;
end;

procedure TfrmCertMgrMain.Display;
var
  F, S: PASN1Struct;
  T: TASN1Struct;
  RC: TX509RevokedCertificate;
  C: TASN1Struct;
begin
  if TabControl1.TabIndex < 4 then begin
    FChain.Clear;
    F := nil;
    T := dmCert.SelectedCert.Data;
    FChain.Add(T);
    repeat
      F := nil;
      case TabControl1.TabIndex of
        0: dmCert.MyCerts.FindCACert(T,F);
        1: dmCert.RootCerts.FindCACert(T,F);
        2: dmCert.IssuedCerts.FindCACert(T,F);
        3: dmCert.CollectedCerts.FindCACert(T,F);
      end;
      if Assigned(F) then begin
        FChain.Insert(0,F^);
        T := F^;
      end;
    until F = nil;
  end;

  RC := nil;
  try
    if not dmCert.RootCerts.FindRevocation(dmCert.SelectedCert.Data,RC) then
      if not dmCert.MyCerts.FindRevocation(dmCert.SelectedCert.Data,RC) then
        if not dmCert.IssuedCerts.FindRevocation(dmCert.SelectedCert.Data,RC) then
          dmCert.CollectedCerts.FindRevocation(dmCert.SelectedCert.Data,RC);
    C := nil;
    dmCert.GetCRLs(C);
    S := nil;
    if not dmCert.CollectedCerts.FindSMIMECapabilities(dmCert.SelectedCert.Data,S) then
      if not dmCert.IssuedCerts.FindSMIMECapabilities(dmCert.SelectedCert.Data,S) then 
        if not dmCert.MyCerts.FindSMIMECapabilities(dmCert.SelectedCert.Data,S) then
          S := nil;
    if Assigned(S) then
      CertificateViewer1.Display(dmCert.SelectedCert,RC,C,S^)
    else
      CertificateViewer1.Display(dmCert.SelectedCert,RC,C,nil);
  finally
    RC.Free;
  end;
end;

procedure TfrmCertMgrMain.FormDestroy(Sender: TObject);
begin       
  Application.OnActivate := nil;
  Application.OnDeactivate := nil;
  FChain.Free;
end;

procedure TfrmCertMgrMain.FormCreate(Sender: TObject);
begin
  FChain := TList.Create;
end;

procedure TfrmCertMgrMain.actNewCertExecute(Sender: TObject);
var
  Dlg: TfrmNewCertDlg;
begin
  Dlg := TfrmNewCertDlg.Create(nil);
  try
    with Dlg do begin
      MyCerts := dmCert.MyCerts;
      RootCerts := dmCert.RootCerts;
      IssuedCerts := dmCert.IssuedCerts;
      PrivateKeyRing := dmCert.StreamSecII1;
      ShowModal;
    end;
  finally
    Dlg.Free;
  end;
  TabControl1Change(Sender);
end;

procedure TfrmCertMgrMain.actRevokeUpdate(Sender: TObject);
var
  CACert: PASN1Struct;
begin
  if dmCert.SelectedCert = nil then
    actRevoke.Enabled := False
  else if dmCert.MyCerts.FindCACert(dmCert.SelectedCert.Data,CACert) then begin
    actRevoke.Enabled := dmCert.StreamSecII1.FindCreatePrivateKey(CACert^) <> nil;
  end else
    actRevoke.Enabled := False;
end;

procedure TfrmCertMgrMain.actRevokeExecute(Sender: TObject);
var
  CACert, F: PASN1Struct;
  PrivKey: IMPPrivateKey;
  Reason: CRLReason;
  HoldInstruction: string;
  InvalidityDate, NextCRLUpdate: TDateTime;
  CRL: TCertificateList;
  URI: string;
  I: Integer;
begin
  if not dmCert.MyCerts.FindCACert(dmCert.SelectedCert.Data,CACert) then
    Exit;
  if dlgRevokation.ShowModal = mrOK then begin
    HoldInstruction := '';
    if dlgRevokation.chbKeyCompromise.Checked then
      Reason := keyCompromise
    else if dlgRevokation.chbCACompromise.Checked then
      Reason := caCompromise
    else if dlgRevokation.chbAffiliationChanged.Checked then
      Reason := affiliationChanged
    else if dlgRevokation.chbSuperseded.Checked then
      Reason := superseded
    else if dlgRevokation.chbCessationOfOperation.Checked then
      Reason := cessationOfOperation
    else if dlgRevokation.chbCertificateHold.Checked then begin
      Reason := certificateHold;
      if dlgRevokation.rbNone.Checked then
        HoldInstruction := id_holdinstruction_none
      else if dlgRevokation.rbCallIssuer.Checked then
        HoldInstruction := id_holdinstruction_callIssuer
      else if dlgRevokation.rbReject.Checked then
        HoldInstruction := id_holdinstruction_reject;
    end else if dlgRevokation.chbRemoveFromCRL.Checked then
      Reason := removeFromCRL
    else
      raise Exception.Create('Must select a reason');
    InvalidityDate := Trunc(dlgRevokation.dtpInvalidityDate.DateTime);
    NextCRLUpdate := Trunc(dlgRevokation.dtpNextCRLUpdate.DateTime);
    PrivKey := dmCert.StreamSecII1.FindCreatePrivateKey(CACert^);
    Assert(dmCert.MyCerts.RevokeCertificate(dmCert.SelectedCert.Data,Reason,HoldInstruction,InvalidityDate,NextCRLUpdate,PrivKey));
    if not dmCert.MyCerts.FindCRL(CACert^,[TX509ReasonFlag(Reason)],F) then
      raise Exception.Create('There is no CRL for the selected CA certificate');
    CRL := TCertificateList.Create(nil,nil);
    try
      CRL.AssignStruct(F^);
      URI := CRL.TbsCertList.DistPointURI;
      if URI <> '' then begin
        SaveDialog2.Title := 'Save CRL ' + URI;
        I:= Length(URI);
        while (I > 0) and (URI[I] <> '/') do
          Dec(I);
        SaveDialog1.FileName := Copy(URI,I+1,MaxInt);
      end;
      if SaveDialog2.Execute then
        CRL.SaveToFile(SaveDialog2.FileName);
    finally
      CRL.Free;
    end;
    Display;
  end;
end;

procedure TfrmCertMgrMain.actExitExecute(Sender: TObject);
begin
  Close;
end;

procedure TfrmCertMgrMain.actOpenExecute(Sender: TObject);
var
  Ext: string;
begin
  if OpenDialog1.Execute then begin
    Ext := LowerCase(ExtractFileExt(OpenDialog1.FileName));
    if (Ext = '.cer') or (Ext = '.crt') then begin
      if frmSelectStoreDlg.Execute(False,OpenDialog1.FileName) then
        dmCert.LoadCertificate(frmSelectStoreDlg.rgSelectStore.ItemIndex,
                               OpenDialog1.FileName);
    end else if (Ext = '.p7b') or (Ext = '.p7c') or (Ext = '.spc') then begin
      if frmSelectStoreDlg.Execute(False,OpenDialog1.FileName) then
        dmCert.ImportFromCMS(frmSelectStoreDlg.rgSelectStore.ItemIndex,
                             OpenDialog1.FileName,True);
    end else if Ext = '.pvk' then
      dmCert.ImportFromPVK(OpenDialog1.FileName)
    else if (Ext = '.pem') or (Ext = '.key') then
      dmCert.ImportFromPEM(OpenDialog1.FileName)
    else if Ext = '.pkr' then
      dmCert.LoadPrivateKeyRing(OpenDialog1.FileName)
    else if Ext = '.crl' then
      dmCert.LoadCrl(OpenDialog1.FileName)
    else if Ext = '.scl' then begin
      if frmSelectStoreDlg.Execute(False,OpenDialog1.FileName) then
        dmCert.ImportFromSCL(frmSelectStoreDlg.rgSelectStore.ItemIndex,
                             OpenDialog1.FileName,True);
    end else if (Ext = '.pfx') or (Ext = '.p12') then begin
      dmCert.ImportFromPFX(OpenDialog1.FileName);
    end else
      raise Exception.Create('Not implemented');
    TabControl1Change(Sender);
  end;
end;

procedure TfrmCertMgrMain.actSaveAsExecute(Sender: TObject);
var
  Ext, FileName: string;
  CACert, F: PASN1Struct;
  PrivKey: IMPPrivateKey;
  Reason: CRLReason;
  HoldInstruction: string;
  InvalidityDate, NextCRLUpdate: TDateTime;
  CRL: TCertificateList;
  Res: Boolean;
begin
  if (TabControl1.TabIndex = 0) and Assigned(dmCert.SelectedCert) then
    SaveDialog1.Filter := 'Certificate (*.cer)|*.cer|' +
                          'CMS Certificate List (*.p7b)|*.p7b|' +
                          'Signed Certificate List (*.scl)|*.scl|' +
                          'Private Key Ring (*.pkr)|*.pkr|' +
                          'Revokation List (*crls)|*.crls|' +
                          'Revokation List (*.crl)|*.crl|' +
                          'MS SignCode Certificate List (*.spc)|*.spc|' +
                          'PKCS#12 file (*.pfx)|*.pfx|' +
                          'RSA Private key (*.pvk)|*.pvk'
  else
    SaveDialog1.Filter := 'Certificate (*.cer)|*.cer|' +
                          'CMS Certificate List (*.p7b)|*.p7b|' +
                          'Signed Certificate List (*.scl)|*.scl|' +
                          'Private Key Ring (*.pkr)|*.pkr|' +
                          'Revokation List (*crls)|*.crls|' +
                          'Revokation List (*.crl)|*.crl|' +
                          'MS SignCode Certificate List (*.spc)|*.spc';
  if SaveDialog1.Execute then begin
    FileName := SaveDialog1.FileName;
    Ext := LowerCase(ExtractFileExt(FileName));
    if Ext = '' then begin
      SaveDialog1TypeChange(Sender);
      Ext := '.' + SaveDialog1.DefaultExt;
      FileName := FileName + Ext;
    end;
    if Ext = '.cer' then begin
      dmCert.SelectedCert.SaveToFile(FileName);
    end else if Ext = '.p7b' then begin
      if frmSelectStoreDlg.Execute(True,FileName) then
        dmCert.ExportToCMS(frmSelectStoreDlg.rgSelectStore.ItemIndex,
                           FileName);
    end else if Ext = '.spc' then begin
      dmCert.ExportSelectedToSPC(FileName);
    end else if Ext = '.scl' then begin
      if frmSelectStoreDlg.Execute(True,FileName) then
        dmCert.SaveSCL(frmSelectStoreDlg.rgSelectStore.ItemIndex,
                       FileName);

    end else if Ext = '.pkr' then begin
      dmCert.SavePrivateKeyRing(FileName);
    end else if Ext = '.pvk' then begin
      dmCert.ExportSelectedToPVK(FileName);
    end else if Ext = '.crls' then
      dmCert.SaveCrls(FileName)
    else if Ext = '.crl' then begin
      if dlgRevokation.ShowModal = mrOK then begin
        HoldInstruction := '';
        if dlgRevokation.chbKeyCompromise.Checked then
          Reason := keyCompromise
        else if dlgRevokation.chbCACompromise.Checked then
          Reason := caCompromise
        else if dlgRevokation.chbAffiliationChanged.Checked then
          Reason := affiliationChanged
        else if dlgRevokation.chbSuperseded.Checked then
          Reason := superseded
        else if dlgRevokation.chbCessationOfOperation.Checked then
          Reason := cessationOfOperation
        else if dlgRevokation.chbCertificateHold.Checked then begin
          Reason := certificateHold;
          if dlgRevokation.rbNone.Checked then
            HoldInstruction := id_holdinstruction_none
          else if dlgRevokation.rbCallIssuer.Checked then
            HoldInstruction := id_holdinstruction_callIssuer
          else if dlgRevokation.rbReject.Checked then
            HoldInstruction := id_holdinstruction_reject;
        end else if dlgRevokation.chbRemoveFromCRL.Checked then
          Reason := removeFromCRL
        else
          raise Exception.Create('Must select a reason');
        Res := True;
        repeat
          if dmCert.MyCerts.FindCRL(dmCert.SelectedCert.Data,[TX509ReasonFlag(Reason)],F) then begin
            Res := True;
            CRL := TCertificateList.Create(nil,nil);
            try
              CRL.AssignStruct(F^);
              CRL.SaveToFile(FileName);
            finally
              CRL.Free;
            end;
            Display;
          end else begin
            if not Res then Break;
            InvalidityDate := Trunc(dlgRevokation.dtpInvalidityDate.DateTime);
            NextCRLUpdate := Trunc(dlgRevokation.dtpNextCRLUpdate.DateTime);
            PrivKey := dmCert.StreamSecII1.FindCreatePrivateKey(dmCert.SelectedCert.Data);
            New(CACert);
            try
              CACert^ := dmCert.SelectedCert.Data;
              Assert(dmCert.MyCerts.RevokeCertificate(nil,
                                                      Reason,
                                                      HoldInstruction,
                                                      InvalidityDate,
                                                      NextCRLUpdate,
                                                      PrivKey,
                                                      CACert));
            finally
              Dispose(CACert);
            end;
            Res := False;
          end;
        until Res;
        if not Res then
          MessageDlg('Could not find a CRL for the selected certificate.'#13#10 +
                     'Please verify that it is a CA certificate with the CRL Distribution Points extension set.',
                     mtWarning,[mbOK],0);
      end;
    end else if Ext = '.pfx' then
      dmCert.ExportSelectedToPFX(FileName)
    else
      raise Exception.Create('Not implemented');
  end;
end;

procedure TfrmCertMgrMain.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  try
    dmCert.SaveDefault;
  except
    on E: ENoSignatureCert do begin
      MessageDlg(E.Message,mtError,[mbOK],0);
      actNewCert.Execute;
      dmCert.SaveDefault;
    end;
  end;
end;

procedure TfrmCertMgrMain.actImportExecute(Sender: TObject);
var
  ImportDlg: TfrmImportDlg;
begin
  ImportDlg := TfrmImportDlg.Create(nil);
  try
    case ImportDlg.ShowModal of
      mrOK:
        if frmSelectStoreDlg.ExecuteMIME(ImportDlg.memImport.Text) then
          dmCert.ImportCertificate(frmSelectStoreDlg.rgSelectStore.ItemIndex,
                                   ImportDlg.memImport.Text);
      mrAbort:
        if OpenDialog2.Execute then
          if frmSelectStoreDlg.Execute(False,OpenDialog2.FileName) then
            dmCert.LoadCertificate(frmSelectStoreDlg.rgSelectStore.ItemIndex,
                                   OpenDialog2.FileName);
    end;
  finally
    ImportDlg.Free;
  end;
end;

procedure TfrmCertMgrMain.actVerifyUpdate(Sender: TObject);
begin
  actVerify.Enabled := dmCert.SelectedCert <> nil;
end;

procedure TfrmCertMgrMain.actVerifyExecute(Sender: TObject);
var
  Status: TCertStatusCode;
  Msg: string;
  Warning: Boolean;
begin
  Msg := '';
  Warning := False;
  case TabControl1.TabIndex of
    0: begin
         dmCert.MyCerts.VerifyCertificate(dmCert.SelectedCert.Data,False,Status);
         if dmCert.StreamSecII1.ValidateMyCert(dmCert.SelectedCert.Data) then
           Msg := 'You have a Private Key for this Certificate'#13#10
         else begin
           Msg := 'Warning: Private key missing'#13#10;
           Warning := True;
         end;
       end;
    1: dmCert.RootCerts.VerifyCertificate(dmCert.SelectedCert.Data,False,Status);
    2: dmCert.IssuedCerts.VerifyCertificate(dmCert.SelectedCert.Data,False,Status);
  else
    dmCert.CollectedCerts.VerifyCertificate(dmCert.SelectedCert.Data,False,Status);
  end;
  case Status of
    crcOK:
      Msg := Msg + 'The Certificate is valid';
    crcExpired:
      Msg := Msg + 'Certificate has expired.';
    crcInvalidSignature:
      Msg := Msg + 'Invalid signature.';
    crcCANotTrusted:
      Msg := Msg + 'The certificate is signed by an unknown CA.';
    crcRevoked:
      Msg := Msg + 'Revoked certificate.';
    crcCAExpiredOrRevoked:
      Msg := Msg + 'The CA certificate has expired or been revoked.';
    crcConstraintFault:
      Msg := Msg + 'Constraint fault.';
    crcInvalidKeyUsage:
      Msg := Msg + 'Invalid key usage: The issuer certificate is not a CA certificate.';
    crcUnsupportedExtension:
      Msg := Msg + 'Unsupported extension: Unable to parse at least one critical extension.';
    crcSyntax:
      Msg := Msg + 'Syntax error: Invalid format.';
    crcTooSmallKey:
      Msg := Msg + Format('Too small key: The certified key is smaller than %d bits',[dmCert.CollectedCerts.LeastKeyBitSize]);
    crcDuplicateKeyIdentifier:
      Msg := Msg + 'Duplicate Key Identifier: You already have another certificate with the same public key';
  else
    Msg := Msg + 'Invalid certificate';
  end;
  if Status = crcOK then begin
    if Warning then
      MessageDlg(Msg,mtWarning,[mbOK],0)
    else
      MessageDlg(Msg,mtInformation,[mbOK],0);
  end else
    MessageDlg(Msg,mtError,[mbOK],0);
end;

procedure TfrmCertMgrMain.actExportUpdate(Sender: TObject);
begin
  actExport.Enabled := dmCert.SelectedCert <> nil;
end;

procedure TfrmCertMgrMain.actExportExecute(Sender: TObject);
var
  ExportDlg: TfrmExportDlg;
begin
  ExportDlg := TfrmExportDlg.Create(nil);
  try
    ExportDlg.memImport.Text := StrToMIME64('-----BEGIN CERTIFICATE-----',
                                            '-----END CERTIFICATE-----',
                                            dmCert.SelectedCert.Data.ContentAsOctetString);
    case ExportDlg.ShowModal of
      mrAbort:
        if SaveDialog3.Execute then
          dmCert.SelectedCert.SaveToFile(SaveDialog3.FileName);
    end;
  finally
    ExportDlg.Free;
  end;
end;

procedure TfrmCertMgrMain.actRemoveUpdate(Sender: TObject);
begin
  actRemove.Enabled := dmCert.SelectedCert <> nil;
end;

procedure TfrmCertMgrMain.actRemoveExecute(Sender: TObject);
begin
  if MessageDlg('Are you sure you want to remove the selected certificate from all certificate stores?',
                mtConfirmation,[mbOK,mbCancel],0) = mrOK then begin
    dmCert.RemoveSelected;
  end;
end;

procedure TfrmCertMgrMain.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  CanClose := dmYarrow.CanUnload;
end;

procedure TfrmCertMgrMain.SaveDialog1TypeChange(Sender: TObject);
begin
  case SaveDialog1.FilterIndex of
    1: SaveDialog1.DefaultExt := 'cer';
    2: SaveDialog1.DefaultExt := 'p7b';
    3: SaveDialog1.DefaultExt := 'scl';
    4: SaveDialog1.DefaultExt := 'pkr';
    5: SaveDialog1.DefaultExt := 'crls';
    6: SaveDialog1.DefaultExt := 'crl';
    7: SaveDialog1.DefaultExt := 'spc';
    8: SaveDialog1.DefaultExt := 'pfx';
    9: SaveDialog1.DefaultExt := 'pvk';
  end;
end;

procedure TfrmCertMgrMain.actRegisterExecute(Sender: TObject);
begin               
{$IFNDEF NO_REGISTER}
  frmRegistration.Show;
{$ENDIF  NO_REGISTER}
end;

procedure TfrmCertMgrMain.actRegisterUpdate(Sender: TObject);
begin
{$IFDEF NO_REGISTER}
  actRegister.Enabled := False;
{$ELSE  NO REGISTER}
  actRegister.Enabled := True;
{$ENDIF NO_REGISTER}
end;

procedure TfrmCertMgrMain.EventActivate;
var
  MS: TMemoryStream;
  F: TASN1Struct;
  Idx: Integer;
begin
  MS := TMemoryStream.Create;
  try
    MS.WriteBuffer(Cert^,CertLen);
    MS.Position := 0;
    F := TASN1Struct.Create;
    try
      F.LoadFromStream(MS,fmtDER);
      Idx := dmCert.MyCerts.IndexOfCert(F);
      if Idx >= 0 then begin
        TabControl1.TabIndex := 0;
        ListView1.Items[Idx].Selected := True;
        ListView1.Items[Idx].Focused := True;
      end;
    finally
      F.Free;
    end;
  finally
    MS.Free;
  end;
  if WindowState = wsMinimized then
    WindowState := wsNormal;
  Application.Restore;
  Application.BringToFront;
  if WindowState = wsMinimized then
    WindowState := wsNormal;
  BringToFront;
  Activate;
end;

procedure TfrmCertMgrMain.actNewWindowExecute(Sender: TObject);
begin
  ShellExecute(0,'OPEN',PChar(ParamStr(0)),'','',SW_SHOWNORMAL);
end;

procedure TfrmCertMgrMain.actActivateExecute(Sender: TObject);
begin
  if dmCert.SelectedCert = nil then Exit;
  if TabControl1.TabIndex > 0 then
    dmCert.FindSelected;
end;

procedure TfrmCertMgrMain.FormActivate(Sender: TObject);
begin
  if Assigned(dmCert) then
    dmCert.Timer1.Enabled := True;
end;

procedure TfrmCertMgrMain.FormDeactivate(Sender: TObject);
begin
  if Assigned(dmCert) then
    dmCert.Timer1.Enabled := True;
end;

procedure TfrmCertMgrMain.FormShow(Sender: TObject);
begin
  Application.OnActivate := FormActivate;
  Application.OnDeactivate := FormDeactivate;
  FormActivate(Sender);
end;

end.
