{
  LICENSE

  Copyright (c) 2004, Henrick Wibell Hellstrm, StreamSec
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of StreamSec nor the names of its contributors may be
      used to endorse or promote products derived from this software without
      specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
}
{$I ver.inc}
unit CertModule;

interface

uses
  SysUtils, Classes, Forms, SecUtils, Asn1, StreamSecII, MpX509, Pkix_Cert,
  Pkix_CRMF, SecComp, ResourceFile, TlsInternalServer, ExtCtrls;

type
  TdmCert = class(TDataModule)
    RootCerts: TX509TrustedCertificates;
    StreamSecII1: TStreamSecII;
    MyCerts: TX509TrustedCertificates;
    CollectedCerts: TX509TrustedCertificates;
    IssuedCerts: TX509TrustedCertificates;
    StreamsecRoot: TResourceFile;
    StreamsecRA: TResourceFile;
    StreamSecCA: TResourceFile;
    Timer1: TTimer;
    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);
    procedure StreamSecII1AdminPassword(Sender: TObject;
      Password: ISecretKey);
    procedure StreamSecII1Password(Sender: TObject; Password: ISecretKey);
    procedure RootCertsCertNotAccepted(Sender: TObject; Cert: TASN1Struct;
      Status: TCertStatusCode);
    procedure Timer1Timer(Sender: TObject);
  private
    FRootCerts: TList;
    FMyCerts: TList;
    FCollectedCerts: TList;
    FIssuedCerts: TList;
    FDefaultPath: string;
    FDefaultPathName: string;
    FSelectedCert: TCertificate;
    FPassword: ISecretKey;
    FCertReqs: TCertReqMessages;
    FCliHandle: LongInt;
    function GetStore(Store: Byte): TX509TrustedCertificates;
    procedure LoadCertificateStream(Store: Byte; Stream: TStream);
    procedure LoadDefault;
    procedure SignSaveCerts(const FileName: string; Store: TX509TrustedCertificates);
  protected
{$IFDEF D6UP}
    function HandleCreateException: Boolean; override;
{$ENDIF D6UP}
  public
    function GetCRLs(var CRLs: TASN1Struct): Boolean;
    procedure ImportFromCMS(Store: Byte;
                            const FileName: string;
                            ExplicitTrust: Boolean);
    procedure ImportFromSCL(Store: Byte;
                            const FileName: string;
                            ExplicitTrust: Boolean);
    procedure ExportToCMS(Store: Byte;
                          const FileName: string);
    procedure ExportSelectedToPFX(const FileName: string);
    procedure ExportSelectedToSPC(const FileName: string);  
    procedure ExportSelectedToPVK(const FileName: string);
    procedure ImportFromPFX(const FileName: string);      
    procedure ImportFromPVK(const FileName: string);
    procedure ImportFromPEM(const FileName: string);
    procedure ImportCertificate(Store: Byte; const Value: string);
    procedure LoadCertificate(Store: Byte; const FileName: string);
    procedure LoadCrl(const FileName: string);
    procedure LoadPrivateKeyRing(const FileName: string);
    procedure RemoveSelected;
    procedure SaveDefault;
    procedure SaveCrls(const FileName: string);
    procedure SavePrivateKeyRing(const FileName: string);
    procedure SaveSCL(Store: Byte; const FileName: string);
    procedure SelectCert(Store: Byte; Index: Integer); overload;
    procedure SelectCert(Cert: TASN1Struct); overload;

    procedure FindSelected;

    property CertReqs: TCertReqMessages read FCertReqs;
    property SelectedCert: TCertificate read FSelectedCert;
  end;

  ENoSignatureCert = class(Exception);
  ENoPasswordException = class(Exception);

var
  dmCert: TdmCert;

implementation

uses
{$IFNDEF D6UP}
  Windows,
{$ENDIF}
  SsBase64, Pkix, Controls, Dialogs, PasswordMain, ReadStrm, Pkcs12,
  SsRijndael, MpPK, MpDL, MpIF, MpEC, X509Base, CertViewDlg,
  CertPathDlg, SsCertMgrUtils, NewCertDlg, MpIFConversion, CertMgrMain;

{$R *.dfm}


function CreateClient: LongInt; stdcall; external 'CertMgrCli.dll';
procedure FreeClient(AHandle: LongInt); stdcall; external 'CertMgrCli.dll';

function LogIn(AHandle: LongInt;
               AStoreName, AStorePath: PWideChar): Boolean; stdcall; external 'CertMgrCli.dll';
procedure LogOut(AHandle: LongInt); stdcall; external 'CertMgrCli.dll';

procedure AddCert(AHandle: LongInt;
                  Cert: Pointer; CertLen: LongInt); stdcall; external 'CertMgrCli.dll';
procedure RemoveCert(AHandle: Integer;
                     Cert: Pointer; CertLen: LongInt); stdcall; external 'CertMgrCli.dll';
function FindCert(AHandle: LongInt;
                  Cert: Pointer; CertLen: LongInt): Boolean; stdcall; external 'CertMgrCli.dll';
function ActivateQuery(AHandle: Integer): Integer; stdcall; external 'CertMgrCli.dll';
procedure ListCACerts(AHandle: LongInt;
                      out CACerts: Pointer; out CACertsLen: LongInt); stdcall; external 'CertMgrCli.dll';
procedure SignCertReq(AHandle: LongInt;
                      ACACertStoreName: PWideChar;
                      CertReq: Pointer; CertReqLen: LongInt;
                      Cert: Pointer; var CertLen: LongInt); stdcall; external 'CertMgrCli.dll';
function CheckSignCertReq(AHandle: LongInt;
                          out CertReq: Pointer; out CertReqLen: LongInt): Boolean; stdcall; external 'CertMgrCli.dll';
procedure SignedCertReq(AHandle: LongInt;
                        Cert: Pointer; CertLen: LongInt); stdcall; external 'CertMgrCli.dll';

{ TdmCert }

procedure TdmCert.LoadCertificate(Store: Byte; const FileName: string);
var
  Ext: string;
  FS: TFileStream;
  MS: TMemoryStream;
  SS: TStringStream;
  S: string;
begin
  Ext := LowerCase(ExtractFileExt(FileName));
  if (Ext = '.cer') or (Ext = '.der') then begin
    FS := TFileStream.Create(FileName,fmOpenRead);
    try
      LoadCertificateStream(Store,FS);
    finally
      FS.Free;
    end;
  end else begin
    MS := TMemoryStream.Create;
    try
      MS.LoadFromFile(FileName);
      SetLength(S,MS.Size);
      MS.Read(Pointer(S)^,Length(S));
    finally
      MS.Free;
    end;
    SS := TStringStream.Create(MIME64ToStr('-----BEGIN CERTIFICATE-----',
                                           '-----END CERTIFICATE-----',
                                           S));
    try
      LoadCertificateStream(Store,SS);
    finally
      SS.Free;
    end;
  end;
end;

procedure TdmCert.DataModuleCreate(Sender: TObject);
var
  PathName, Path: WideString;
  Tries: Integer;
begin
  Tries := 0;
  repeat
    try
      FCliHandle := CreateClient;
    except
      Inc(Tries);
      if Tries > 3 then
        raise
      else
        Sleep(2000);
    end;
  until FCliHandle <> 0;
  GCliHandle := FCliHandle;
  IssuedCerts.HoursOffsetFromGMT := MpX509.OffsetFromUTC;
  CollectedCerts.HoursOffsetFromGMT := MpX509.OffsetFromUTC;
  RootCerts.HoursOffsetFromGMT := MpX509.OffsetFromUTC;
  MyCerts.HoursOffsetFromGMT := MpX509.OffsetFromUTC;
  FCollectedCerts := TList.Create;
  FIssuedCerts := TList.Create;
  FMyCerts := TList.Create;
  FRootCerts := TList.Create;
  FCertReqs := TCertReqMessages.Create(nil,nil);
  LoadDefault;
  PathName := FDefaultPathName;
  Path := FDefaultPath;
  if not LogIn(FCliHandle,PWideChar(PathName),PWideChar(Path)) then
    raise ENoPasswordException.Create('Unable to lock path');
end;

procedure TdmCert.DataModuleDestroy(Sender: TObject);

  procedure ClearAndFreeList(List: TList);
  var
    I: Integer;
  begin
    for I := 0 to List.Count - 1 do
      TObject(List[I]).Free;
    List.Free;
  end;

begin
  GCliHandle := 0;
  dmCert := nil;
  Timer1.Enabled := False;
  LogOut(FCliHandle);
  ClearAndFreeList(FCollectedCerts);
  ClearAndFreeList(FIssuedCerts);
  ClearAndFreeList(FMyCerts);
  ClearAndFreeList(FRootCerts);
  FCertReqs.Free;
  GCliHandle := 0;
  FreeClient(FCliHandle);
  FPassword := nil;
end;

procedure TdmCert.ImportFromCMS(Store: Byte;
                                const FileName: string;
                                ExplicitTrust: Boolean);
var
  TC: TX509TrustedCertificates;
  Status: TCertStatusCodes;
begin
  TC := GetStore(Store);

  {IMPORTANT: This code does NOT test if certificates imported into the
              Personal store have corresponding private keys. Please confer
              the LoadCertificate method to see how this is accomplished.}

  if not TC.ImportFromCMSFile(FileName,ExplicitTrust,Status) then
    MessageDlg('Invalid format',mtError,[mbOK],0)
  else if not ((Status <= [crcOK,crcDuplicateKeyIdentifier]) and
               (Status <> [])) then
    MessageDlg('Some certificates were not accepted',mtWarning,[mbOK],0);
  case Store of
    0: FMyCerts.Count := TC.Count;
    1: FRootCerts.Count := TC.Count;
    2: FIssuedCerts.Count := TC.Count;
    3: FCollectedCerts.Count := TC.Count;
  end;
end;

procedure TdmCert.LoadDefault;
var
  FileName: string;
  Password: ISecretKey;
  FS: TFileStream;
  NoPrivateKeyRing: Boolean;  
  Dlg: TfrmNewCertDlg;
  lDlg: INewCertDlg;

  procedure LoadCerts(const FileName: string; Store: TX509TrustedCertificates);
  begin
    if FileExists(FileName) then try
      Store.LoadFromFile(FileName);
      if Store.SignatureCert <> nil then
        if not StreamSecII1.ValidateMyCert(Store.SignatureCert) then
          raise Exception.Create('Unrecognized or invalid signature certificate in ' + FileName);
    except
      on E: Exception do
        MessageDlg(E.Message,mtError,[mbOK],0);
    end;
  end;

begin
  if ParamCount > 1 then begin
    if LowerCase(ExtractFileName(ParamStr(1))) = 'user.pkr' then begin
      FDefaultPath := ExtractFilePath(ParamStr(1));
      FDefaultPathName := TSsCertMgrPaths.GetPathName(FDefaultPath);
    end else if TSsCertMgrPaths.GetPath(ParamStr(1)) <> '' then begin
      FDefaultPathName := ParamStr(1);
      FDefaultPath := TSsCertMgrPaths.GetPath(FDefaultPathName);
    end;
    if FDefaultPath = '' then
      MessageDlg(Format('Unknown parameter: %s',[ParamStr(1)]),mtError,[mbOK],0);
  end;
  if FDefaultPath = '' then
    with TfrmCertPath.Create(Application) do begin
      ShowModal;
      FDefaultPath := CertFilePath;
      FDefaultPathName := CertFilePathName;
      Free;
    end;
  FileName := FDefaultPath + 'User.pkr';
  FDefaultPath := ExtractFilePath(FileName);
  if FileExists(FileName) then begin
    NoPrivateKeyRing := False;
    if PasswordDlg(Format('Private Key Ring (%s)',[FDefaultPathName]),Password) then begin
      FS := TFileStream.Create(FileName,fmOpenRead);
      try
        if StreamSecII1.LoadPrivateKeyRingFromStream(FS,Password) then
          FPassword := Password
        else
          raise ENoPasswordException.Create('Wrong password or corrupt key ring file: Execution brutally aborted');
      finally
        FS.Free;
      end;
    end else
      raise ENoPasswordException.Create('No password: Execution brutally aborted');
  end else
    NoPrivateKeyRing := True;
  LoadCerts(FDefaultPath + 'User.scl',MyCerts);
  FMyCerts.Count := MyCerts.Count;
  LoadCerts(FDefaultPath + 'Root.scl',RootCerts);
  FRootCerts.Count := RootCerts.Count;
  LoadCerts(FDefaultPath + 'Issued.scl',IssuedCerts);
  FIssuedCerts.Count := IssuedCerts.Count;
  LoadCerts(FDefaultPath + 'Collected.scl',CollectedCerts);
  FCollectedCerts.Count := CollectedCerts.Count;

  if FileExists(FDefaultPath + 'outgoing.crm') then
    FCertReqs.LoadFromFile(FDefaultPath + 'outgoing.crm');

  StreamSecRoot.DataStream.Position := 0;
  LoadCertificateStream(1,StreamSecRoot.DataStream);
  StreamSecCA.DataStream.Position := 0;
  LoadCertificateStream(1,StreamSecCA.DataStream);
  StreamSecRA.DataStream.Position := 0;
  LoadCertificateStream(1,StreamSecRA.DataStream);

  if NoPrivateKeyRing then
    case MessageDlg('Could not find a private key ring. You are adviced to create a Personal Certificate'#13#10 +
                    'Press OK to create a Personal Certificate now (recommended)'#13#10 +
                    'Press Cancel to run CertMgr without creating a new certificate at this point'#13#10 +
                    'Press Abort to exit CertMgr',
                    mtWarning,[mbOK,mbCancel,mbAbort],0) of
      mrAbort:
        raise ENoPasswordException.Create('Execution brutally aborted');
      mrOK:
        begin
          Dlg := TfrmNewCertDlg.Create(nil);
          try
            if Supports(Dlg,INewCertDlg,lDlg) then
              with lDlg do begin
                MyCerts := Self.MyCerts;
                RootCerts := Self.RootCerts;
                IssuedCerts := Self.IssuedCerts;
                PrivateKeyRing := Self.StreamSecII1;
                SelectWizard(ncwCA);
              end;
            if Dlg.ShowModal = mrOK then
              SaveDefault;
          finally
            Dlg.Free;
          end;
        end;
    end;
  Application.Title := FDefaultPathName;

  frmCertMgrMain.TabControl1Change(nil);
end;

procedure TdmCert.SaveDefault;
var
  MS: TSecureMemoryStream;
begin
  SignSaveCerts(FDefaultPath + 'User.scl',MyCerts);
  SignSaveCerts(FDefaultPath + 'Root.scl',RootCerts);
  SignSaveCerts(FDefaultPath + 'Issued.scl',IssuedCerts);
  SignSaveCerts(FDefaultPath + 'Collected.scl',CollectedCerts);

  FCertReqs.SaveToFile(FDefaultPath + 'outgoing.crm');

  if (FPassword = nil) or (FPassword.KeyLen = 0) then
    NewPasswordDlg('Private key ring',FPassword);
  if Assigned(FPassword) and (FPassword.KeyLen > 0) then begin
    MS := TSecureMemoryStream.Create;
    try
      StreamSecII1.SavePrivateKeyRingToStream(MS,FPassword,
                                              kdfWeakPassword256,
                                              $10000,
                                              True,
                                              id_aes256_wrap,
                                              TRijndael_ABC);
      MS.SaveToFile(FDefaultPath + 'User.pkr');
      if FDefaultPathName = TSsCertMgrPaths.GetDefaultPathName then
        InputQuery('Register ' + FDefaultPath,'Enter name',FDefaultPathName);
      TSsCertMgrPaths.RegisterPath(FDefaultPath,FDefaultPathName);
    finally
      MS.Free;
    end;
  end;
end;

procedure TdmCert.SelectCert(Store: Byte; Index: Integer);
var
  List: TList;
  TC: TX509TrustedCertificates;
begin
  if (FMyCerts.IndexOf(FSelectedCert) < 0) and
     (FRootCerts.IndexOf(FSelectedCert) < 0) and
     (FIssuedCerts.IndexOf(FSelectedCert) < 0) and
     (FCollectedCerts.IndexOf(FSelectedCert) < 0) then
    FSelectedCert.Free;
  case Store of
    0: begin
         List := FMyCerts;
         TC := MyCerts;
       end;
    1: begin
         List := FRootCerts;
         TC := RootCerts;
       end;
    2: begin
         List := FIssuedCerts;
         TC := IssuedCerts;
       end;
    3: begin
         List := FCollectedCerts;
         TC := CollectedCerts;
       end;
  else
    Exit;
  end;
  if List.Count < TC.Count then
    List.Count := TC.Count;
  FSelectedCert := List[Index];
  if FSelectedCert = nil then begin
    FSelectedCert := TCertificate.Create(nil,nil);
    try
      FSelectedCert.Data.ForgetOptionalDefault := True;
      FSelectedCert.AssignStruct(TC.Certs[Index]);
      List[Index] := FSelectedCert;
    except
      FSelectedCert.Free;
      FSelectedCert := nil;
      raise;
    end;
  end;
end;

function TdmCert.GetStore(Store: Byte): TX509TrustedCertificates;
begin
  case Store of
    0: Result := MyCerts;
    1: Result := RootCerts;
    2: Result := IssuedCerts;
    3: Result := CollectedCerts;
  else
    Result := nil;
  end;
end;

procedure TdmCert.LoadPrivateKeyRing(const FileName: string);
var
  Password: ISecretKey;
  FS: TFileStream;
begin
  if PasswordDlg('Private key ring',Password) then begin
    FS := TFileStream.Create(FileName,fmOpenRead);
    try
      if StreamSecII1.LoadPrivateKeyRingFromStream(FS,Password) then
        FPassword := Password;
    finally
      FS.Free;
    end;
  end;
end;

procedure TdmCert.SelectCert(Cert: TASN1Struct);
begin
  if (FMyCerts.IndexOf(FSelectedCert) < 0) and
     (FRootCerts.IndexOf(FSelectedCert) < 0) and
     (FIssuedCerts.IndexOf(FSelectedCert) < 0) and
     (FCollectedCerts.IndexOf(FSelectedCert) < 0) then
    FSelectedCert.Free;
  FSelectedCert := TCertificate.Create(nil,nil);
  try
    FSelectedCert.Data.ForgetOptionalDefault := True;
    FSelectedCert.AssignStruct(Cert);
  except
    FSelectedCert.Free;
    FSelectedCert := nil;
    raise;
  end;
end;

procedure TdmCert.LoadCrl(const FileName: string);
var
  F: TASN1Struct;
  FS: TFileStream;
  P: PASN1Struct;
  Status: TCertStatusCode;
begin
  F := TASN1Struct.Create;
  try
    FS := TFileStream.Create(FileName,fmOpenRead);
    try
      F.LoadFromStream(FS,fmtDER);
      if RootCerts.FindCRLCACert(F,P) then
        RootCerts.AddCRL(F,Status)
      else if CollectedCerts.FindCRLCACert(F,P) then
        CollectedCerts.AddCRL(F,Status)
      else if MyCerts.FindCRLCACert(F,P) then
        MyCerts.AddCRL(F,Status)
      else if IssuedCerts.FindCRLCACert(F,P) then
        IssuedCerts.AddCRL(F,Status);
    finally
      FS.Free;
    end;
  finally
    F.Free;
  end;
end;

procedure TdmCert.ExportToCMS(Store: Byte; const FileName: string);
begin
  GetStore(Store).ExportAllToCMSFile(FileName);
end;

procedure TdmCert.SavePrivateKeyRing(const FileName: string);
var
  Password: string;
  FS: TFileStream;
begin
  if NewPasswordDlg('Private key ring',Password) then try
    FS := TFileStream.Create(FileName,fmCreate);
    try
      StreamSecII1.SavePrivateKeyRingToStream(FS,Password,kdfWeakPassword256,$10000,True,id_aes256_wrap,TRijndael_ABC);
    finally
      FS.Free;
    end;
  finally
    ProtectClear(Pointer(Password)^,Length(Password));
  end;
end;

procedure TdmCert.SaveCrls(const FileName: string);
var
  FS: TFileStream;
  P: TASN1Struct;
begin
  FS := TFileStream.Create(FileName,fmCreate);
  try
    P := nil;
    if not GetCRLs(P) then
      raise Exception.Create('No CRL issued by the selected certificate found');
    P.SaveToStream(FS,fmtDER)
  finally
    FS.Free;
  end;
end;

function TdmCert.GetCRLs(var CRLs: TASN1Struct): Boolean;
begin
  Result := True;
  CRLs := nil;
  if SelectedCert = nil then
    Result := False
  else if RootCerts.FindCRLs(SelectedCert.Data,CRLs) = 0 then
    if CollectedCerts.FindCRLs(SelectedCert.Data,CRLs) = 0 then
      if MyCerts.FindCRLs(SelectedCert.Data,CRLs) = 0 then
        if IssuedCerts.FindCRLs(SelectedCert.Data,CRLs) = 0 then
          Result := False;
end;

procedure TdmCert.StreamSecII1AdminPassword(Sender: TObject;
  Password: ISecretKey);
begin
  PasswordDlg('Private Key Ring Admin (Write access)',Password);
end;

procedure TdmCert.StreamSecII1Password(Sender: TObject;
  Password: ISecretKey);
begin
  NewPasswordDlg(Format('Private Key Ring (%s)',[FDefaultPathName]),Password);
  FPassword := Password;
end;

procedure TdmCert.ImportCertificate(Store: Byte; const Value: string);
var
  SS: TStringStream;
begin
  SS := TStringStream.Create(MIME64ToStr('-----BEGIN CERTIFICATE-----',
                                         '-----END CERTIFICATE-----',
                                         Value));
  try
    LoadCertificateStream(Store,SS);
  finally
    SS.Free;
  end;
end;

procedure TdmCert.LoadCertificateStream(Store: Byte; Stream: TStream);
var
  TC: TX509trustedCertificates;
  F: TASN1Struct;
  Status: TCertStatusCode;
  ExplicitTrust: Boolean;
  Index: Integer;
  S: OctetString;
begin
  TC := GetStore(Store);
  F := TASN1Struct.Create;
  try
    F.LoadFromStream(Stream,fmtDER);
    ExplicitTrust := Store = 1;
    if Store = 0 then
      if StreamSecII1.FindCreatePrivateKey(F) = nil then
        raise Exception.Create('Unable to find a matching private key');
    repeat
      Index := TC.AddCertificate(F,ExplicitTrust,Status);
      if Status in [crcOK,crcTrusted] then begin
        case Store of
          0: FMyCerts.Count := TC.Count;
          1: FRootCerts.Count := TC.Count;
          2: FIssuedCerts.Count := TC.Count;
          3: FCollectedCerts.Count := TC.Count;
        end;
        if (Store = 0) and (Index = TC.Count - 1) then begin
          S := F.ContentAsOctetString;
          AddCert(FCliHandle,Pointer(S),Length(S));
        end;
        SelectCert(Store,Index);
        Break;
      end else begin
        case Status of
          crcExpired:
            begin
              MessageDlg('Certificate has expired.',mtError,[mbOK],0);
              ExplicitTrust := False;
            end;
          crcInvalidSignature:
            begin
              MessageDlg('Invalid signature.',mtError,[mbOK],0);
              ExplicitTrust := False;
            end;
          crcCANotTrusted:
            ExplicitTrust := MessageDlg('The certificate is signed by an unknown CA.'#13#10 +
                                        'Load anyway?',mtWarning,[mbYes,mbNo],0) =
                             mrYes;
          crcRevoked:
            begin
              MessageDlg('Revoked certificate.',mtError,[mbOK],0);
              ExplicitTrust := False;
            end;
          crcCAExpiredOrRevoked:
            begin
              MessageDlg('The CA certificate has expired or been revoked.',mtError,[mbOK],0);
              ExplicitTrust := False;
            end;
          crcConstraintFault:
            begin
              MessageDlg('Constraint fault.',mtError,[mbOK],0);
              ExplicitTrust := False;
            end;
          crcInvalidKeyUsage:
            begin
              MessageDlg('Invalid key usage: The issuer certificate is not a CA certificate.',mtError,[mbOK],0);
              ExplicitTrust := False;
            end;
          crcUnsupportedExtension:
            begin
              MessageDlg('Unsupported extension: Unable to parse at least one critical extension.',mtError,[mbOK],0);
              ExplicitTrust := False;
            end;
          crcSyntax:
            begin
              MessageDlg('Syntax error: Invalid format.',mtError,[mbOK],0);
              ExplicitTrust := False;
            end;
          crcTooSmallKey:
            begin
              MessageDlg(Format('Too small key: The certified key is smaller than %d bits',[CollectedCerts.LeastKeyBitSize]),
                         mtError,[mbOK],0);
              ExplicitTrust := False;
            end; 
          crcDuplicateKeyIdentifier:
            begin
              MessageDlg('Duplicate Key Identifier: You already have another certificate with the same public key',
                         mtError,[mbOK],0);
              ExplicitTrust := False;
            end;
        end;
      end;
    until not ExplicitTrust;
  finally
    F.Free;
  end;
end;

procedure TdmCert.ImportFromSCL(Store: Byte; const FileName: string;
  ExplicitTrust: Boolean);
var
  TC: TX509TrustedCertificates;
begin
  TC := TX509TrustedCertificates.Create(nil);
  try
    TC.LoadFromFile(FileName);
    TC.ExportAllToCMSFile(FileName + '.p7b');
    ImportFromCMS(Store,FileName + '.p7b',ExplicitTrust);
  finally
    TC.Free;
  end;
end;

procedure TdmCert.SaveSCL(Store: Byte; const FileName: string);
var
  TC: TX509TrustedCertificates;
begin
  TC := GetStore(Store);
  SignSaveCerts(FileName,TC);
end;

procedure TdmCert.SignSaveCerts(const FileName: string;
  Store: TX509TrustedCertificates);
var
  MustFindSignCert: Boolean;
  Cert: PASN1Struct;
  Idx: Integer;
  PrivKey: IMPPrivateKey;
begin
  MustFindSignCert := (Store.SignatureCert = nil) or
                      Store.SignatureCert.IsEmpty or
                      not StreamSecII1.ValidateMyCert(Store.SignatureCert);
  if Store.Modified or MustFindSignCert then begin
    if MustFindSignCert then begin
      Idx := 0;
      Cert := nil;
      if not MyCerts.FindCert([],[rsaEncryption,id_dsa,id_ecPublicKey],[keyCertSign],Idx,Cert) then
        MyCerts.FindCert([],[rsaEncryption,id_dsa,id_ecPublicKey],[digitalSignature],Idx,Cert);
      if Assigned(Cert) then
        Store.SignatureCert := Cert^
      else
        raise ENoSignatureCert.Create('You must create a certificate with a digital signature key in order to save your certificate store');
    end;
    PrivKey := StreamSecII1.FindCreatePrivateKey(Store.SignatureCert);
    if Assigned(PrivKey) then
      Store.SignSaveToFile(FileName,Store.SignatureCert,PrivKey);
  end else 
    Store.SaveToFile(FileName);
end;

procedure TdmCert.RootCertsCertNotAccepted(Sender: TObject;
  Cert: TASN1Struct; Status: TCertStatusCode);
var
  T: TCertificate;
  RC: TX509RevokedCertificate;
begin
  T := TCertificate.Create(nil,nil);
  try
    T.AssignStruct(Cert);
    RC := nil;
    try
      if not dmCert.RootCerts.FindRevocation(T.Data,RC) then
        if not dmCert.MyCerts.FindRevocation(T.Data,RC) then
          if not dmCert.IssuedCerts.FindRevocation(T.Data,RC) then
            dmCert.CollectedCerts.FindRevocation(T.Data,RC);
      if frmCertificateViewer = nil then
        frmCertificateViewer := TfrmCertificateViewer.Create(Application);
      case Status of
        crcExpired:                frmCertificateViewer.Caption := 'Expired Certificate';
        crcInvalidSignature:       frmCertificateViewer.Caption := 'Invalid Signature';
        crcCANotTrusted:           frmCertificateViewer.Caption := 'CA Not Trusted';
        crcRevoked:                frmCertificateViewer.Caption := 'Revoked Certificate';
        crcCAExpiredOrRevoked:     frmCertificateViewer.Caption := 'CA Expired or Revoked';
        crcConstraintFault:        frmCertificateViewer.Caption := 'Constraint Fault';
        crcInvalidKeyUsage:        frmCertificateViewer.Caption := 'Invalid Key Usage';
        crcUnsupportedExtension:   frmCertificateViewer.Caption := 'Unsupported Extension';
        crcSuperseeded:            frmCertificateViewer.Caption := 'Superseeded';
        crcSyntax:                 frmCertificateViewer.Caption := 'Syntax Error';
        crcBaseNotFound:           frmCertificateViewer.Caption := 'Base Not Found';
        crcTrusted:                frmCertificateViewer.Caption := 'Trusted';
        crcTooSmallKey:            frmCertificateViewer.Caption := 'Too Small Key';
        crcPolicyNotAccepted:      frmCertificateViewer.Caption := 'Policy Not Accepted';
        crcDuplicateKeyIdentifier: frmCertificateViewer.Caption := 'Duplicate Key Identifier';
      end;
      frmCertificateViewer.CertificateViewer1.Display(T,RC,nil,nil);
      frmCertificateViewer.ShowModal;
    finally
      RC.Free;
    end;
  finally
    T.Free;
  end;
  frmCertificateViewer.Caption := 'Certificate';
end;

procedure TdmCert.ExportSelectedToPFX(const FileName: string);
var
  Password, AdminPW: ISecretKey;
  RSAPrivateKey: TASN1Struct;
  PrivateKeyInfo: TASN1Struct;
  SafeBag: TASN1Struct;
  SafeContents: TASN1Struct;
  ContentInfo: TASN1Struct;
  AuthenticatedSafe: TASN1Struct;
  PDU: TASN1Struct;
  Priv: TMPIFPrivateKey;
  Cert: TASN1Struct;
  CACert: PASN1Struct;
begin
  if PasswordMain.NewPasswordDlg(FileName,Password,True) then begin
    PDU := nil;
    AuthenticatedSafe := nil;
    ContentInfo := nil;
    SafeContents := nil;
    try
      NewComposeASN1Struct(AuthenticatedSafe,V_ASN1_UNIVERSAL,True,V_ASN1_SEQUENCE);
      NewComposeASN1Struct(SafeContents,V_ASN1_UNIVERSAL,True,V_ASN1_SEQUENCE);
      SafeBag := nil;
      try
        PrivateKeyInfo := nil;
        try
          RSAPrivateKey := nil;
          try
            AdminPW := nil;
            Priv := StreamSecII1.FindCreatePrivateKey(SelectedCert.Data) as TMPIFPrivateKey;
            if not Priv.ExportToRSAPrivateKey(RSAPrivateKey,False) then
              if not (StreamSecII1.AdminLogin(AdminPW) and
                      Priv.ExportToRSAPrivateKey(RSAPrivateKey,True)) then
                raise Exception.Create('Could not export private key');
            Pkcs12.ImposeRSAPrivateKey(RSAPrivateKey,
                                       ExtractKeyUsage(SelectedCert.Data),
                                       PrivateKeyInfo);
          finally
            RSAPrivateKey.Free;
          end;
          Pkcs12.ImposePrivateKeyInfo(PrivateKeyInfo,
                                      Password,
                                      OSToHex(SelectedCert.PublicKeyIdentifier),
                                      #1#0#0#0,
                                      'Microsoft RSA SChannel Cryptographic Provider',
                                      2000,pbe3DES_192,
                                      SafeBag);
        finally
          PrivateKeyInfo.Free;
        end;
        SafeContents.AddField('','',Safebag);
        SafeContents.CalculateLength;
        Pkcs12.ImposeSafeContents(SafeContents,Password,2000,pbeNone,ContentInfo);
        AuthenticatedSafe.AddField('','',ContentInfo);

        NewComposeASN1Struct(SafeContents,V_ASN1_UNIVERSAL,True,V_ASN1_SEQUENCE);
        Pkcs12.ImposeCertificate(SelectedCert.Data,'StrSecII CertMgr ' + DateTimeToStr(Now),#1#0#0#0,SafeBag);
        SafeContents.AddField('','',Safebag);
        Cert := SelectedCert.Data;
        while MyCerts.FindCACert(Cert,CACert) do begin
          Pkcs12.ImposeCertificate(CACert^,'','',SafeBag);
          SafeContents.AddField('','',Safebag);
          Cert := CACert^;
        end;
        SafeContents.CalculateLength;
        Pkcs12.ImposeSafeContents(SafeContents,Password,2000,pbeRC2_40,ContentInfo);
        AuthenticatedSafe.AddField('','',ContentInfo);
      finally
        SafeBag.Free;
        SafeContents.Free;
        ContentInfo.Free;
      end;
      AuthenticatedSafe.CalculateLength;
      Pkcs12.ImposeAuthSafeContent(AuthenticatedSafe,Password,2000,PDU);
      PDU.SaveToFile(FileName,fmtDER);
    finally
      AuthenticatedSafe.Free;
      PDU.Free;
    end;
  end;
end;

procedure TdmCert.ImportFromPFX(const FileName: string);
var
  Password: ISecretKey;
  RSAPrivateKey: TASN1Struct;
  Cert: TASN1Struct;
  PrivateKeyInfo: TASN1Struct;
  SafeContents: TASN1Struct;
  SafeContentsList: TASN1Struct;
  AuthenticatedSafe: TASN1Struct;
  PDU: TASN1Struct;
  OK: Boolean;
  I, J, K: Integer;
  FriendlyName, LocalKeyId, CertLocalKeyId: string;
  Status: TCertStatusCode;
  Priv: TIFPrivateKey;
  S: OctetString;
begin
  if PasswordMain.PasswordDlg(FileName,Password,True) then begin
    SafeContents := nil;
    AuthenticatedSafe := nil;
    PrivateKeyInfo := nil;
    RSAPrivateKey := nil;
    Cert := nil;
    PDU := TASN1Struct.Create;
    try
      PDU.LoadFromFile(FileName,fmtDER);
      OK := Pkcs12.ExtractAuthSafeContent(PDU,Password,AuthenticatedSafe);
      if OK then begin
        SafeContentsList := nil;
        NewComposeASN1Struct(SafeContentsList,V_ASN1_UNIVERSAL,True,V_ASN1_SEQUENCE);
        try
          for I := 0 to AuthenticatedSafe.ItemCount - 1 do begin
            OK := Pkcs12.ExtractSafeContents(AuthenticatedSafe.Items[I]^,Password,SafeContents);
            if OK then
              for J := 0 to SafeContents.ItemCount - 1 do
                SafeContentsList.AddField('','SEQUENCE',SafeContents.Items[J]^);
          end;
          if SafeContentsList.ItemCount > 0 then begin
            for J := 0 to SafeContentsList.ItemCount - 1 do begin
              OK := Pkcs12.ExtractPrivateKeyInfo(SafeContentsList.Items[J]^,Password,PrivateKeyInfo,FriendlyName,LocalKeyId);
              if OK then begin
                for K := 0 to SafeContentsList.ItemCount - 1 do begin
                  OK := Pkcs12.ExtractCertificate(SafeContentsList.Items[K]^,Cert,CertLocalKeyId);
                  if OK then
                    if CertLocalKeyId = LocalKeyId then begin
                      OK := not StreamSecII1.ValidateMyCert(Cert);
                      if OK then
//                        FriendlyName := ExtractSubjectKeyIdentifier(Cert);
                        with TCertificate.Create(nil,nil) do try
                          AssignStruct(Cert);
                          FriendlyName := PublicKeyIdentifier;
                          if FriendlyName = '' then
                            FriendlyName := SubjectPublicKeyInfo.PublicKeyIdentifier;
                        finally
                          Free;
                        end;
                      Break;
                    end;
                  OK := True;
                end;
                if OK and (FriendlyName <> '') then begin
                  OK := Pkcs12.ExtractRSAPrivateKey(PrivateKeyInfo,RSAPrivateKey);
                  if OK then begin
                    FillChar(Priv,SizeOf(Priv),0);
                    try
                      OK := ASN1StructToRSAPrivateKey(RSAPrivateKey,Priv);
                      if OK then
                        if StreamSecII1.AddPrivateRSAKey(Priv,Pointer(FriendlyName),Length(FriendlyName),True) < 0 then
                          if StreamSecII1.AdminLogin then
                            if StreamSecII1.AddPrivateRSAKey(Priv,Pointer(FriendlyName),Length(FriendlyName),True) < 0 then
                              raise Exception.Create('The import failed');
                    finally
                      DisposeIFPrivateKey(Priv);
                    end;
                  end;
                end else if FriendlyName = '' then
                  raise Exception.Create('Import failure: No SubjectKeyIdentifier');
              end;
            end;
            for J := 0 to SafeContentsList.ItemCount - 1 do begin
              OK := Pkcs12.ExtractCertificate(SafeContentsList.Items[J]^,Cert,LocalKeyId);
              if OK then begin
                try
                  OK := StreamSecII1.ValidateMyCert(Cert);
                except
                  OK := False;
                end;
                if OK then begin
                  MyCerts.AddCertificate(Cert,True,Status,True);
                  S := Cert.ContentAsOctetString;
                  AddCert(FCliHandle,Pointer(S),Length(S));
                end else if IsAllegedIssuer(Cert,Cert) then
                  RootCerts.AddCertificate(Cert,True,Status,True)
                else
                  CollectedCerts.AddCertificate(Cert,True,Status,True);
              end;
            end;
          end;
        finally
          SafeContentsList.Free;
        end;
      end else
        raise Exception.Create('Import failure: Bad format, corrupt file or wrong password');
    finally
      PDU.Free;
      AuthenticatedSafe.Free;
      SafeContents.Free;
      PrivateKeyInfo.Free;
      Cert.Free;
      RSAPrivateKey.Free;
    end;
  end;
end;

procedure TdmCert.FindSelected;
var
  S: OctetString;
begin
  S := FSelectedCert.Data.ContentAsOctetString;
  if not FindCert(FCliHandle,Pointer(S),Length(S)) then
    raise Exception.Create('Unable to locate certificate');
end;

procedure TdmCert.RemoveSelected;
var
  S: OctetString;
begin
  if MyCerts.RemoveCertificate(SelectedCert.Data) then begin
    S := SelectedCert.Data.ContentAsOctetString;
    RemoveCert(FCliHandle,Pointer(S),Length(S));
  end;
  RootCerts.RemoveCertificate(SelectedCert.Data);
  IssuedCerts.RemoveCertificate(SelectedCert.Data);
  CollectedCerts.RemoveCertificate(SelectedCert.Data);
end;

procedure TdmCert.Timer1Timer(Sender: TObject);
var
  Idx: Integer;
  MS, Crt: TMemoryStream;
  CrtReq: TReadStream;
  P: Pointer;
  PLen: Integer;
  Dlg: TfrmNewCertDlg;
begin
  Timer1.Enabled := False;
  try
    Timer1.Interval := 100;
    Idx := ActivateQuery(FCliHandle);
    if Idx > -1 then begin
      MS := TMemoryStream.Create;
      try
        MyCerts.Certs[Idx].SaveToStream(MS,fmtDER);
        frmCertMgrMain.EventActivate(MS.Memory,MS.Size);
      finally
        MS.Free;
      end;
    end;
    if CheckSignCertReq(FCliHandle,P,PLen) then begin
      CrtReq := TReadStream.Create(P^,PLen);
      try
        Crt := TMemoryStream.Create;
        try
          Dlg := TfrmNewCertDlg.Create(nil);
          try
            with Dlg do begin
              ImportData := CrtReq;
              ExportData := Crt;
              MyCerts := dmCert.MyCerts;
              RootCerts := dmCert.RootCerts;
              IssuedCerts := dmCert.IssuedCerts;
              PrivateKeyRing := dmCert.StreamSecII1;
              SelectWizard(0);
              ShowModal;
            end;
          finally
            Dlg.FreeOnRelease;
          end;
          SignedCertReq(FCliHandle,Crt.Memory,Crt.Size);
        finally
          Crt.Free;
        end;
        frmCertMgrMain.TabControl1Change(Sender);
      finally
        CrtReq.Free;
        Timer1.Interval := 1000;
      end;
    end;
  finally
    Timer1.Enabled := True;
  end;
end;

procedure TdmCert.ImportFromPEM(const FileName: string);
begin
  StreamSecII1.LoadPrivateKeyFromFile(FileName);
end;

procedure TdmCert.ImportFromPVK(const FileName: string);
begin
  StreamSecII1.LoadPrivateKeyFromFile(FileName);
end;

procedure TdmCert.ExportSelectedToSPC(const FileName: string);
var
  Idx: Integer;
begin
  if SelectedCert = nil then
    raise Exception.Create('No selection');
  Idx := MyCerts.IndexOfCert(SelectedCert.Data);
  if Idx < 0 then
    raise Exception.Create('No personal certificate selected');
  MyCerts.ExportChainToCMSFile(FileName,Idx,False,True);
end;

procedure TdmCert.ExportSelectedToPVK(const FileName: string);
var
  PW: ISecretKey;
begin
  if SelectedCert = nil then
    raise Exception.Create('No selection');
  if MyCerts.IndexOfCert(SelectedCert.Data) < 0 then
    raise Exception.Create('No personal certificate selected');
  if not PasswordDlg('Private Key ' + FileName,PW) then
    if MessageDlg('You did not enter a password'#13#10 +
                  'The private key will be saved without encryption',
                   mtWarning,[mbOK,mbAbort],0) = mrAbort then
      Abort;
  if not StreamSecII1.SavePrivateKeyToFile(SelectedCert.Data,FileName,PW) then
    raise Exception.Create('Unable to save private key');
end;

{$IFDEF D6UP}
function TdmCert.HandleCreateException: Boolean;
begin
  if ExceptObject is ENoPasswordException then begin
    inherited HandleCreateException;
    Result := False;
  end else
    Result := inherited HandleCreateException;
end;         
{$ENDIF D6UP}

initialization
  SsCertMgrUtils.GSignCertReq := SignCertReq;
  SsCertMgrUtils.GCheckSignCertReq := CheckSignCertReq;
  SsCertMgrUtils.GSignedCertReq := SignedCertReq;
end.
