unit CertPathDlg;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ImgList;

type
  TfrmCertPath = class(TForm)
    lvPaths: TListView;
    btnOK: TButton;
    btnEdit: TButton;
    btnAdd: TButton;
    btnCancel: TButton;
    OpenDialog1: TOpenDialog;
    StateImages: TImageList;
    procedure FormShow(Sender: TObject);
    procedure btnEditClick(Sender: TObject);
    procedure btnAddClick(Sender: TObject);
    procedure btnCancelClick(Sender: TObject);
    procedure lvPathsSelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
  private
    function GetCertFilePath: string;
    function GetCertFilePathName: string;
  public
    property CertFilePath: string read GetCertFilePath;
    property CertFilePathName: string read GetCertFilePathName;
  end;

var
  frmCertPath: TfrmCertPath;

implementation

uses
  SsCertMgrUtils;

{$R *.dfm}

procedure TfrmCertPath.FormShow(Sender: TObject);
var
  SL: TStringList;
  I: Integer;
  LI: TListItem;
begin
  lvPaths.Items.Clear;
  SL := TStringList.Create;
  try
    TSsCertMgrPaths.GetNamesAndPaths(SL);
    LI := nil;
    for I := 0 to SL.Count - 1 do begin
      LI := lvPaths.Items.Add;
      LI.Caption := SL.Names[I];
      LI.SubItems.Add(SL.Values[LI.Caption]);
      if TSsCertMgrPaths.PathIsLocked(LI.SubItems[0]) then
        LI.StateIndex := 0;
    end;
    if Assigned(LI) then
      LI.Selected := True;
  finally
    SL.Free;
  end;
end;

function TfrmCertPath.GetCertFilePath: string;
var
  LI: TListItem;
begin
  LI := lvPaths.Selected;
  if LI = nil then
    Result := ExtractFilePath(ExpandUNCFileName('User.pkr'))
  else
    Result := LI.SubItems[0];
end;

procedure TfrmCertPath.btnEditClick(Sender: TObject);
var
  LI: TListItem;
  PathName: string;
begin
  LI := lvPaths.Selected;
  if LI = nil then
    Beep
  else begin
    PathName := LI.Caption;
    if InputQuery('Rename ' + LI.SubItems[0],'Enter name',PathName) then begin
      TSsCertMgrPaths.RegisterPath(LI.SubItems[0],PathName);
      LI.Caption := PathName;
    end;
  end;
end;

procedure TfrmCertPath.btnAddClick(Sender: TObject);
var
  LI: TListItem;
  PathName: string;
  PathCaption: string;
begin
  if OpenDialog1.Execute then begin
    LI := lvPaths.Items.Add;
    LI.SubItems.Add(ExtractFilePath(OpenDialog1.FileName));
    PathName := TSsCertMgrPaths.GetPathName(LI.SubItems[0]);
    if PathName = '' then begin
      PathName := TSsCertMgrPaths.GetDefaultPathName;
      PathCaption := 'Add ' + LI.SubItems[0];
    end else
      PathCaption := 'Rename ' + LI.SubItems[0];
    if InputQuery(PathCaption,'Enter name',PathName) then begin
      TSsCertMgrPaths.RegisterPath(LI.SubItems[0],PathName);
      LI.Caption := PathName;
      LI.Selected := True;
    end else
      LI.Free;
  end;
end;

function TfrmCertPath.GetCertFilePathName: string;
var
  LI: TListItem;
begin
  LI := lvPaths.Selected;
  if LI = nil then
    Result := TSsCertMgrPaths.GetDefaultPathName
  else
    Result := LI.Caption;
end;

procedure TfrmCertPath.btnCancelClick(Sender: TObject);
begin
  lvPaths.Selected := nil;
end;

procedure TfrmCertPath.lvPathsSelectItem(Sender: TObject; Item: TListItem;
  Selected: Boolean);
begin
  if Selected and (Item.StateIndex = 0) then begin
    Item.Selected := False;
    raise Exception.Create('Item is locked');
  end;
end;

end.
