{*******************************************************}
{                                                       }
{     StreamSec Security Library for Borland Delphi     }
{     CertificateSelection Unit                         }
{
  LICENSE

  Copyright (c) 2004, Henrick Wibell Hellstrm, StreamSec
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of StreamSec nor the names of its contributors may be
      used to endorse or promote products derived from this software without
      specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
{*******************************************************}
{$I ver.inc}
unit CertificateSelection;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  Dialogs, ExtCtrls, CertificateViewer,
  Pkix_Cert, MpX509, StdCtrls, CheckLst;

type
  TCertChangeEvent = procedure (Sender: TObject; Cert: TCertificate) of object;

  TCertificateSelection = class(TCustomPanel)
  private
    CertificateViewer1: TCertificateViewer;
    clbCertList: TCheckListBox;
    procedure clbCertListClick(Sender: TObject);
  private
    FGroupIndex: Integer;
    FCertSource: TCustomX509TrustedCertificates;
    FCert: TCertificate;
    FOnCertChange: TCertChangeEvent;
    procedure SetCertSource(const Value: TCustomX509TrustedCertificates);
    procedure SetGroupIndex(const Value: Integer);
    procedure SetOnCertChange(const Value: TCertChangeEvent);
  protected
    procedure CreateControls;
    procedure SetControlProperties;
    procedure SetParent(AParent: TWinControl); override;
    procedure MakeAllVisible;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DisplayAll;
    procedure PostGroupIndex;
    property SelectedCert: TCertificate read FCert;
  published
    property Align;
    property CertSource: TCustomX509TrustedCertificates read FCertSource write SetCertSource;
    property GroupIndex: Integer read FGroupIndex write SetGroupIndex;
    property OnCertChange: TCertChangeEvent read FOnCertChange write SetOnCertChange;
  end;

procedure Register;

implementation

uses
  Asn1, X509Base;

procedure Register;
begin
  RegisterComponents('StreamSec', [TCertificateSelection]);
end;

{ TCertificateSelection }

constructor TCertificateSelection.Create(AOwner: TComponent);
begin
  inherited;
  CreateControls;
end;

destructor TCertificateSelection.Destroy;
begin
  FCert.Free;
  inherited;
end;

procedure TCertificateSelection.DisplayAll;
var
  Item: TCertificateItem;
  I, J, Idx: Integer;
begin
  clbCertList.Clear;
  MakeAllVisible;
  for I := 0 to FCertSource.CertificateCollectionCount - 1 do
    for J := 0 to FCertSource.CertificateCollections[I].Count - 1 do begin
      Item := FCertSource.CertificateCollections[I].Items[J] as TCertificateItem;
      Idx := clbCertList.Items.AddObject(Item.DisplayName,Item);
      clbCertList.Checked[Idx] := Item.GroupIndex = GroupIndex;
    end;
  if clbCertList.Items.Count > 0 then begin
    clbCertList.ItemIndex := 0;
    clbCertListClick(Self);
  end;
end;

procedure TCertificateSelection.MakeAllVisible;
var
  Item: TCertificateItem;
  I, J, Idx: Integer;
begin
  for I := 0 to FCertSource.CertificateCollectionCount - 1 do begin
    FCertSource.CertificateCollections[I].BeginUpdate;
    try
      J := 0;
      while J < FCertSource.CertificateCollections[I].Count do begin
        Item := FCertSource.CertificateCollections[I].Items[J] as TCertificateItem;
        Idx := Item.CertIndex;
        if Idx > J then begin
          Item := FCertSource.CertificateCollections[I].Add as TCertificateItem;
          Item.CertIndex := J;
          Item.Index := J;
          Item.GroupIndex := -1;
        end else if Idx < J then begin
          {$IFDEF D5UP}
          if Idx < 0 then
            FCertSource.CertificateCollections[I].Delete(J)
          else begin
            FCertSource.CertificateCollections[I].Delete(Idx);
            Item.Index := Idx;
          end;
          {$ELSE D5UP}
          if Idx < 0 then
            FCertSource.CertificateCollections[I].Items[J].Free
          else begin
            FCertSource.CertificateCollections[I].Items[Idx].Free;
            Item.Index := Idx;
          end;
          {$ENDIF}
          Dec(J);
        end;
        Inc(J);
      end;
      repeat
        Item := FCertSource.CertificateCollections[I].Add as TCertificateItem;
        if Item.CertIndex < 0 then begin
          Item.Free;
          Item := nil;
        end;
      until Item = nil;
    finally
      FCertSource.CertificateCollections[I].EndUpdate;
    end;
  end;
end;

procedure TCertificateSelection.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if Operation = opRemove then begin
    if AComponent = FCertSource then
      FCertSource := nil;
  end;
end;

procedure TCertificateSelection.PostGroupIndex;
var
  Item: TCertificateItem;
  I, J, Idx: Integer;
begin
  for I := 0 to FCertSource.CertificateCollectionCount - 1 do begin
    FCertSource.CertificateCollections[I].BeginUpdate;
    try
      for J := 0 to FCertSource.CertificateCollections[I].Count - 1 do begin
        Item := FCertSource.CertificateCollections[I].Items[J] as TCertificateItem;
        Idx := clbCertList.Items.IndexOf(Item.DisplayName);
        if Idx >= 0 then begin
          if clbCertList.Checked[Idx] then
            Item.GroupIndex := GroupIndex
          else if Item.GroupIndex = GroupIndex then
            Item.GroupIndex := -1;
        end;
      end;
    finally
      FCertSource.CertificateCollections[I].EndUpdate;
    end;
  end;
end;

procedure TCertificateSelection.SetCertSource(
  const Value: TCustomX509TrustedCertificates);
begin
  FCertSource := Value;
  if Assigned(Value) then
    FreeNotification(Value);
end;

procedure TCertificateSelection.SetGroupIndex(const Value: Integer);
begin
  FGroupIndex := Value;
end;

procedure TCertificateSelection.clbCertListClick(Sender: TObject);
var
  Idx: Integer;
  Item: TCertificateItem;
  S: PASN1Struct;
  RC: TX509RevokedCertificate;
  C: TASN1Struct;
begin
  Idx := clbCertList.ItemIndex;
  if Idx >= 0 then begin
    Item := clbCertList.Items.Objects[Idx] as TCertificateItem;

    FCert.Free;
    FCert := TCertificate.Create(nil,nil);
    FCert.AssignStruct(Item.Cert);

    RC := nil;
    try
      FCertSource.FindRevocation(Item.Cert,RC);
      C := nil;
      FCertSource.FindCRLs(Item.Cert,C);
      S := nil;
      if not FCertSource.FindSMIMECapabilities(Item.Cert,S) then
        S := nil;
      if Assigned(S) then
        CertificateViewer1.Display(FCert,RC,C,S^)
      else
        CertificateViewer1.Display(FCert,RC,C,nil);
    finally
      RC.Free;
    end;
    if Assigned(OnCertChange) then OnCertChange(Self,FCert);
  end;
end;

procedure TCertificateSelection.CreateControls;
begin
  CertificateViewer1 := TCertificateViewer.Create(Self);
  CertificateViewer1.Name := 'CertificateViewer1';
  clbCertList := TCheckListBox.Create(Self);
  clbCertList.Name := 'clbCertList';
end;

procedure TCertificateSelection.SetControlProperties;
begin
  Width := 623;
  Height := 563;
  CertificateViewer1.Parent := Self;
  with CertificateViewer1 do begin
    Height := 287;
    Align := alBottom;
    TabOrder := 0;
  end;
  clbCertList.Parent := Self;
  with clbCertList do begin
    Left := 0;
    Top := 0;
    Align := alClient;
    OnClick := clbCertListClick;
  end;
end;

procedure TCertificateSelection.SetParent(AParent: TWinControl);
begin
  inherited;
  if Assigned(AParent) then SetControlProperties;
end;

procedure TCertificateSelection.SetOnCertChange(
  const Value: TCertChangeEvent);
begin
  FOnCertChange := Value;
end;

end.
