{
  LICENSE

  Copyright (c) 2004, Henrick Wibell Hellstrm, StreamSec
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of StreamSec nor the names of its contributors may be
      used to endorse or promote products derived from this software without
      specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
}
unit QCertificateViewer;

interface

uses
  SysUtils, Classes, QGraphics, QControls, QForms,
  QStdCtrls, QExtCtrls, QComCtrls, Asn1, Pkix_Cert, Pkix_CRL, X509Base;

type
  TCertificateViewer = class(TCustomPanel)
  published
    pcCertificateViewer: TPageControl;
    tabSubject: TTabSheet;
    lvSubject: TListView;
    tabIssuer: TTabSheet;
    lvIssuer: TListView;
    tabFields: TTabSheet;
    Label1: TLabel;
    lblVersion: TLabel;
    Label3: TLabel;
    lblSerial: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    lblNotBefore: TLabel;
    Label8: TLabel;
    lblNotAfter: TLabel;
    Label10: TLabel;
    lblSignature: TLabel;
    Label12: TLabel;
    lblSUID: TLabel;
    Label14: TLabel;
    lblIUID: TLabel;
    Label15: TLabel;
    memSignature: TMemo;
    tabPublicKey: TTabSheet;
    memSubjectPublicKey: TMemo;
    tabExtensions: TTabSheet;
    lbExtensions: TListBox;
    memExtension: TMemo;
    tabRevokation: TTabSheet;
    memRevokation: TMemo;
    tabCRLs: TTabSheet;
    memCRLs: TMemo;
    tabSMIMECapabilities: TTabSheet;
    memSMIMECapabilities: TMemo;
    procedure lbExtensionsClick(Sender: TObject);
  private
    FSelectedExt: TExtnValueEnum;
  protected
    procedure CreateControls;
    procedure SetControlProperties;
    procedure SetParent(const AParent: TWidgetControl); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Display(Cert: TCertificate;
                      RC: TX509RevokedCertificate;
                      CRLs, SMIMECapabilities: TASN1Struct);
  published
    property Align;
    property Anchors;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnEnter;
    property OnExit;
    property OnResize;
  end;

procedure Register;

implementation

uses
  Pkix, SsArc2, MpCms;

procedure Register;
begin
  RegisterComponents('StreamSec', [TCertificateViewer]);
end;

{ TCertificateViewer }

constructor TCertificateViewer.Create(AOwner: TComponent);
begin
  inherited;
  CreateControls;
end;

procedure TCertificateViewer.CreateControls;
var
  O: TComponent;
begin
  O := Self;

  pcCertificateViewer := TPageControl.Create(O);
  pcCertificateViewer.Name  := 'pcCertificateViewer';

  tabSubject := TTabSheet.Create(O);
  tabSubject.Name := 'tabSubject';
  lvSubject := TListView.Create(O);
  lvSubject.Name := 'lvSubject';

  tabIssuer := TTabSheet.Create(O);
  tabIssuer.Name := 'tabIssuer';
  lvIssuer := TListView.Create(O);
  lvIssuer.Name := 'lvIssuer';

  tabFields := TTabSheet.Create(O);
  tabFields.Name := 'tabFields';
  Label1 := TLabel.Create(O);
  Label1.Name := 'Label1';
  lblVersion := TLabel.Create(O);
  lblVersion.Name := 'lblVersion';
  Label3 := TLabel.Create(O);
  Label3.Name := 'Label3';
  lblSerial := TLabel.Create(O);
  lblSerial.Name := 'lblSerial';
  Label5 := TLabel.Create(O);
  Label5.Name := 'Label5';
  Label6 := TLabel.Create(O);
  Label6.Name := 'Label6';
  lblNotBefore := TLabel.Create(O);
  lblNotBefore.Name := 'lblNotBefore';
  Label8 := TLabel.Create(O);
  Label8.Name := 'Label8';
  lblNotAfter := TLabel.Create(O);
  lblNotAfter.Name := 'lblNotAfter';
  Label10 := TLabel.Create(O);
  Label10.Name := 'Label10';
  lblSignature := TLabel.Create(O);
  lblSignature.Name := 'lblSignature';
  Label12 := TLabel.Create(O);
  Label12.Name := 'Label12';
  lblSUID := TLabel.Create(O);
  lblSUID.Name := 'lblSUID';
  Label14 := TLabel.Create(O);
  Label14.Name := 'Label14';
  lblIUID := TLabel.Create(O);
  lblIUID.Name := 'lblIUID';
  Label15 := TLabel.Create(O);
  Label15.Name := 'Label15';
  memSignature := TMemo.Create(O);
  memSignature.Name := 'memSignature';

  tabPublicKey := TTabSheet.Create(O);
  tabPublicKey.Name := 'tabPublicKey';
  memSubjectPublicKey := TMemo.Create(O);
  memSubjectPublicKey.Name := 'memSubjectPublicKey';

  tabExtensions := TTabSheet.Create(O);
  tabExtensions.Name := 'tabExtensions';
  lbExtensions := TListBox.Create(O);
  lbExtensions.Name := 'lbExtensions';
  memExtension := TMemo.Create(O);
  memExtension.Name := 'memExtension';

  tabRevokation := TTabSheet.Create(O);
  tabRevokation.Name := 'tabRevokation';
  memRevokation := TMemo.Create(O);
  memRevokation.Name := 'memRevokation';

  tabCRLs := TTabSheet.Create(O);
  tabCRLs.Name := 'tabCRLs';
  memCRLs := TMemo.Create(O);
  memCRLs.Name := 'memCRLs';

  tabSMIMECapabilities := TTabSheet.Create(O);
  tabSMIMECapabilities.Name := 'tabSMIMECapabilities';
  memSMIMECapabilities := TMemo.Create(O);
  memSMIMECapabilities.Name := 'memSMIMECapabilities';
end;

destructor TCertificateViewer.Destroy;
begin
  try
    inherited;
  except
    raise;
  end;
end;

procedure TCertificateViewer.Display(Cert: TCertificate;
  RC: TX509RevokedCertificate; CRLs, SMIMECapabilities: TASN1Struct);
var
  I, J: Integer;
  RDName: TName;
  Item: TListItem;
  SPKI: TSubjectPublicKeyInfo;
  Ext: TExtension;
  CRL: TCertificateList;
  CRLExt: TCrlExtension;
  Cap: TSMIMECapabilities;
begin
  lvSubject.Items.Clear;
  RDName := Cert.TbsCertificate.Subject;
  for I := 0 to RDName.AsRdnSequence.ItemCount - 1 do begin
    for J := 0 to RDName.AsRdnSequence.Items[I].ItemCount - 1 do begin
      Item := lvSubject.Items.Add;
      Item.Caption := GetObjectName(RDName.AsRdnSequence.Items[I].Items[J].AttrType);
      Item.SubItems.Add(RDName.AsRdnSequence.Items[I].Items[J].AttrValue.Value);
    end;
  end;

  lvIssuer.Items.Clear;
  RDName := Cert.TbsCertificate.Issuer;
  for I := 0 to RDName.AsRdnSequence.ItemCount - 1 do begin
    for J := 0 to RDName.AsRdnSequence.Items[I].ItemCount - 1 do begin
      Item := lvIssuer.Items.Add;
      Item.Caption := GetObjectName(RDName.AsRdnSequence.Items[I].Items[J].AttrType);
      Item.SubItems.Add(RDName.AsRdnSequence.Items[I].Items[J].AttrValue.Value);
    end;
  end;

  lblVersion.Caption := Format('Version %d',[Cert.Version]);
  lblSerial.Caption := Cert.SerialNumber.AsHexString + ' (hex)';
  lblNotBefore.Caption := DateTimeToStr(Cert.Validity.NotBefore.AsUtcTime.AsDateTime);
  lblNotAfter.Caption := DateTimeToStr(Cert.Validity.NotAfter.AsUtcTime.AsDateTime);
  lblSignature.Caption := GetObjectName(Cert.TbsCertificate.Signature.Algorithm);
  lblSUID.Caption := Cert.TbsCertificate.SubjectUniqueID.AsHexString;
  lblIUID.Caption := Cert.TbsCertificate.IssuerUniqueID.AsHexString;

  memSignature.Text := Cert.Signature.AsHexString;

  memSubjectPublicKey.Clear;
  SPKI := Cert.TbsCertificate.SubjectPublicKeyInfo;
  memSubjectPublicKey.Lines.Add('Algorithm: ' + GetObjectName(SPKI.Algorithm.Algorithm));
  case SPKI.Algorithm.Parameters.Choice of
    aideRSAEncryption:
      begin
        memSubjectPublicKey.Lines.Add(Format('Key size: %d bits',
          [SPKI.SubjectPublicKey.AsRsaEncryption.Modulus.Length*8-8]));
        memSubjectPublicKey.Lines.Add('Modulus (n): ' +
          SPKI.SubjectPublicKey.AsRsaEncryption.Modulus.AsHexString);
        memSubjectPublicKey.Lines.Add('Exponent (e): ' +
          SPKI.SubjectPublicKey.AsRsaEncryption.PublicExponent.AsHexString);
      end;
    aideIdDsa:
      begin
        memSubjectPublicKey.Lines.Add(Format('Key size: %d / %d bits',
          [SPKI.Algorithm.Parameters.AsDsa.P.Length*8-8,
           SPKI.Algorithm.Parameters.AsDsa.Q.Length*8-8]));
        memSubjectPublicKey.Lines.Add('Modulus (p): ' +
          SPKI.Algorithm.Parameters.AsDsa.P.AsHexString);
        memSubjectPublicKey.Lines.Add('Modulus (g): ' +
          SPKI.Algorithm.Parameters.AsDsa.G.AsHexString);
        memSubjectPublicKey.Lines.Add('Modulus (q): ' +
          SPKI.Algorithm.Parameters.AsDsa.Q.AsHexString);
        memSubjectPublicKey.Lines.Add('Public key (y): ' +
          SPKI.SubjectPublicKey.AsDsa.AsHexString);
      end;
    aideDhPublicNumber:
      begin
        memSubjectPublicKey.Lines.Add(Format('Key size: %d / %d bits',
          [SPKI.Algorithm.Parameters.AsDhPublicNumber.P.Length*8-8,
           SPKI.Algorithm.Parameters.AsDhPublicNumber.Q.Length*8-8]));
        memSubjectPublicKey.Lines.Add('Modulus (p): ' +
          SPKI.Algorithm.Parameters.AsDhPublicNumber.P.AsHexString);
        memSubjectPublicKey.Lines.Add('Modulus (g): ' +
          SPKI.Algorithm.Parameters.AsDhPublicNumber.G.AsHexString);
        memSubjectPublicKey.Lines.Add('Modulus (q): ' +
          SPKI.Algorithm.Parameters.AsDhPublicNumber.Q.AsHexString);
        memSubjectPublicKey.Lines.Add('Public key (y): ' +
          SPKI.SubjectPublicKey.AsDhPublicNumber.AsHexString);
      end;
    aideIdEcPublicKey:
      begin
        memSubjectPublicKey.Lines.Add('Curve: ' +
          GetObjectName(SPKI.Algorithm.Parameters.AsEcPublicKey));
        memSubjectPublicKey.Lines.Add('Public key (W): ' +
          SPKI.SubjectPublicKey.AsEcPublicKey.AsHexString);
      end;
  end;

  if Cert.Version < 3 then
    tabExtensions.TabVisible := False
  else begin
    lbExtensions.Clear;
    for I := 0 to Cert.TbsCertificate.Extensions.Count - 1 do
      lbExtensions.Items.AddObject(GetObjectName(Cert.TbsCertificate.Extensions.Items[I].ExtnID),
        Cert.TbsCertificate.Extensions.Items[I]);

    tabExtensions.TabVisible := True;
    Ext := Cert.TbsCertificate.Extensions.UniqueItem[FSelectedExt];
    if Assigned(Ext) then begin
      lbExtensions.ItemIndex := lbExtensions.Items.IndexOf(GetObjectName(Ext.ExtnID));
      lbExtensionsClick(nil);
    end else
      memExtension.Clear;
  end;

  memRevokation.Clear;
  if Assigned(RC) then begin
    tabRevokation.TabVisible := True;
    memRevokation.Lines.Add('Revokation date: ' + DateTimeToStr(RC.revocationDate));
    case RC.ReasonCode of
      keyCompromise:       memRevokation.Lines.Add('Reason: keyCompromise');
      cACompromise:        memRevokation.Lines.Add('Reason: caCompromise');
      affiliationChanged:  memRevokation.Lines.Add('Reason: affiliationChanged');
      superseded:          memRevokation.Lines.Add('Reason: superseded');
      cessationOfOperation:memRevokation.Lines.Add('Reason: cessationOfOperation');
      certificateHold:     memRevokation.Lines.Add('Reason: certificateHold');
      removeFromCRL:       memRevokation.Lines.Add('Reason: removeFromCRL');
    else
      memRevokation.Lines.Add('Invalid Reason: unrecognized code in CRL');
    end;
    memRevokation.Lines.Add('Invalidity date: ' + DateTimeToStr(RC.InvalidityDate));
  end else
    tabRevokation.TabVisible := False;

  memCRLs.Clear;
  if Assigned(CRLs) and (CRLs.ItemCount > 0) then begin
    tabCRLs.TabVisible := True;
    for I := 0 to CRLs.ItemCount - 1 do begin
      CRL := TCertificateList.Create(nil,nil);
      try
        CRL.AssignStruct(CRLs.Items[I]^);
        memCRLs.Lines.Add('This Update: ' + CRL.TbsCertList.ThisUpdate.Value);
        memCRLs.Lines.Add('Next Update: ' + CRL.TbsCertList.NextUpdate.Value);

        CrlExt := CRL.TbsCertList.CrlExtensions.UniqueItem[cveIdCeCRLNumber];
        if Assigned(CrlExt) then
          memCRLs.Lines.Add('CRL Number: ' + CrlExt.ExtnValue.AsCe_CRLNumber.Value);

        CrlExt := CRL.TbsCertList.CrlExtensions.UniqueItem[cveIdCeDeltaCRLIndicator];
        if Assigned(CrlExt) then
          memCRLs.Lines.Add('Delta CRL Indicator: ' + CrlExt.ExtnValue.AsCe_DeltaCRLIndicator.Value);

        CrlExt := CRL.TbsCertList.CrlExtensions.UniqueItem[cveIdCeIssuingDistributionPoint];
        if Assigned(CrlExt) then begin
          memCRLs.Lines.Add('Issuing distribution point name: ' + CrlExt.ExtnValue.AsCe_IssuingDistributionPoint.DistributionPoint.AsFullName.Text);
          memCRLs.Lines.Add('Reasons:');
          if CrlExt.ExtnValue.AsCe_IssuingDistributionPoint.OnlySomeReasons.Bits[Ord(crlrKeyCompromise)] then
            memCRLs.Lines.Add('  keyCompromise');
          if CrlExt.ExtnValue.AsCe_IssuingDistributionPoint.OnlySomeReasons.Bits[Ord(crlrCACompromise)] then
            memCRLs.Lines.Add('  cACompromise');
          if CrlExt.ExtnValue.AsCe_IssuingDistributionPoint.OnlySomeReasons.Bits[Ord(crlrAffiliationChanged)] then
            memCRLs.Lines.Add('  affiliationChanged');
          if CrlExt.ExtnValue.AsCe_IssuingDistributionPoint.OnlySomeReasons.Bits[Ord(crlrSuperseded)] then
            memCRLs.Lines.Add('  superseded');
          if CrlExt.ExtnValue.AsCe_IssuingDistributionPoint.OnlySomeReasons.Bits[Ord(crlrCessationOfOperation)] then
            memCRLs.Lines.Add('  cessationOfOperation');
          if CrlExt.ExtnValue.AsCe_IssuingDistributionPoint.OnlySomeReasons.Bits[Ord(crlrCertificateHold)] then
            memCRLs.Lines.Add('  certificateHold');
          if CrlExt.ExtnValue.AsCe_IssuingDistributionPoint.OnlySomeReasons.Bits[Ord(crlrRemoveFromCRL)] then
            memCRLs.Lines.Add('  removeFromCRL');
        end;
      finally
        CRL.Free;
      end;
    end;
  end else
    tabCRLs.TabVisible := False;

  memSMIMECapabilities.Clear;
  if Assigned(SMIMECapabilities) and (SMIMECapabilities.ItemCount > 0) then begin
    tabSMIMECapabilities.TabVisible := True;
    Cap := TSMIMECapabilities.Create(nil,TSMIMECapabilityItem);
    try
      Cap.AssignStruct(SMIMECapabilities,False);
      for I := 0 to Cap.Count - 1 do
        memSMIMECapabilities.Lines.Add(Cap.Items[I].DisplayName);
    finally
      Cap.Free;
    end;
  end else
    tabSMIMECapabilities.TabVisible := False;
end;

procedure TCertificateViewer.lbExtensionsClick(Sender: TObject);
var
  I, J, K, L: Integer;
  Ext: TExtension;
  S: string;
begin
  memExtension.Clear;
  I := lbExtensions.ItemIndex;
  if I >= 0 then begin
    Ext := TExtension(lbExtensions.Items.Objects[I]);
    if Ext.Critical then
      memExtension.Lines.Add('Critical Extension: True')
    else
      memExtension.Lines.Add('Critical Extension: False');
    memExtension.Lines.Add('');
    FSelectedExt := Ext.ExtnValue.Choice;
    case Ext.ExtnValue.Choice of
      eveIdCeAuthorityKeyIdentifier:
        begin
          memExtension.Lines.Add('Key identifier: ' +
            Ext.ExtnValue.AsCe_AuthorityKeyIdentifier.KeyIdentifier.AsHexString);
          memExtension.Lines.Add('Authority Cert Issuer: '#13#10 +
            Ext.ExtnValue.AsCe_AuthorityKeyIdentifier.AuthorityCertIssuer.Text);
          memExtension.Lines.Add('Authority Cert Serial: ' +
            Ext.ExtnValue.AsCe_AuthorityKeyIdentifier.AuthorityCertSerialNumber.AsHexString);
        end;
      eveIdCeSubjectKeyIdentifier:
        begin
          memExtension.Lines.Add('Key identifier: ' +
            Ext.ExtnValue.AsCe_SubjectKeyIdentifier.AsHexString);
        end;
      eveIdCeKeyUsage:
        begin
          if Ext.ExtnValue.AsCe_KeyUsage.Bits[digitalSignatureVal] then
            memExtension.Lines.Add('digitalSignature');
          if Ext.ExtnValue.AsCe_KeyUsage.Bits[nonRepudiationVal] then
            memExtension.Lines.Add('nonRepudiation');
          if Ext.ExtnValue.AsCe_KeyUsage.Bits[keyEnciphermentVal] then
            memExtension.Lines.Add('keyEncipherment');
          if Ext.ExtnValue.AsCe_KeyUsage.Bits[dataEnciphermentVal] then
            memExtension.Lines.Add('dataEncipherment');
          if Ext.ExtnValue.AsCe_KeyUsage.Bits[keyAgreementVal] then
            memExtension.Lines.Add('keyAgreement');
          if Ext.ExtnValue.AsCe_KeyUsage.Bits[keyCertSignVal] then
            memExtension.Lines.Add('keyCertSign');
          if Ext.ExtnValue.AsCe_KeyUsage.Bits[cRLSignVal] then
            memExtension.Lines.Add('cRLSign');
          if Ext.ExtnValue.AsCe_KeyUsage.Bits[encipherOnlyVal] then
            memExtension.Lines.Add('encipherOnly');
          if Ext.ExtnValue.AsCe_KeyUsage.Bits[decipherOnlyVal] then
            memExtension.Lines.Add('decipherOnly');
        end;
      eveIdCeExtKeyUsage:
        begin
          memExtension.Lines.Add('Extended key usage:');
          for J := 0 to Ext.ExtnValue.AsCe_ExtKeyUsage.Count - 1 do
            memExtension.Lines.Add(GetObjectName(Ext.ExtnValue.AsCe_ExtKeyUsage[J]));
        end;
      eveIdCePrivateKeyUsagePeriod:
        begin
          memExtension.Lines.Add('Private key usage period:');
          memExtension.Lines.Add('  Not before: ' +
            DateTimeToStr(Ext.ExtnValue.AsCe_PrivateKeyUsagePeriod.NotBefore.AsDateTime));
          memExtension.Lines.Add('  Not after: ' +
            DateTimeToStr(Ext.ExtnValue.AsCe_PrivateKeyUsagePeriod.NotAfter.AsDateTime));
        end;
      eveIdCeCertificatePolicies:
        begin
          memExtension.Lines.Add('Certificate policies:');
          for J := 0 to Ext.ExtnValue.AsCe_CertificatePolicies.Count - 1 do begin
            memExtension.Lines.Add(
              GetObjectName(Ext.ExtnValue.AsCe_CertificatePolicies[J].PolicyIdentifier));
            for K := 0 to Ext.ExtnValue.AsCe_CertificatePolicies[J].PolicyQualifiers.Count - 1 do
              if Ext.ExtnValue.AsCe_CertificatePolicies[J].PolicyQualifiers.Items[K].Qualifier.Choice = qeIdQtCps then
                memExtension.Lines.Add('  cps: ' +
                  Ext.ExtnValue.AsCe_CertificatePolicies[J].PolicyQualifiers.Items[K].Qualifier.AsQt_Cps)
              else begin
                memExtension.Lines.Add('  User notice: ' +
                  Ext.ExtnValue.AsCe_CertificatePolicies[J].PolicyQualifiers.Items[K].Qualifier.AsQt_Unotice.ExplicitText.Value);
                S := Ext.ExtnValue.AsCe_CertificatePolicies[J].PolicyQualifiers.Items[K].Qualifier.AsQt_Unotice.NoticeRef.Organization.Value;
                if S <> '' then begin
                  for L := 0 to Ext.ExtnValue.AsCe_CertificatePolicies[J].PolicyQualifiers.Items[K].Qualifier.AsQt_Unotice.NoticeRef.NoticeNumbers.Count - 1 do
                    S := S + ' ' + Ext.ExtnValue.AsCe_CertificatePolicies[J].PolicyQualifiers.Items[K].Qualifier.AsQt_Unotice.NoticeRef.NoticeNumbers[L].Value;
                  memExtension.Lines.Add('    Notice reference: ' + S);
                end;
              end;
          end;
        end;
      eveIdCePolicyMappings:
        begin
          memExtension.Lines.Add('Policy mappings:');
          for J := 0 to Ext.ExtnValue.AsCe_PolicyMappings.Count - 1 do
            memExtension.Lines.Add('  Issuer domain policy ' +
              GetObjectName(Ext.ExtnValue.AsCe_PolicyMappings[J].IssuerDomainPolicy) +
              ' = Subject domain policy ' +
              GetObjectName(Ext.ExtnValue.AsCe_PolicyMappings[J].SubjectDomainPolicy));
        end;
      eveIdCeSubjectAltName:
        memExtension.Lines.Add('Subject alternative name:'#13#10 +
          Ext.ExtnValue.AsCe_SubjectAltName.Text);
      eveIdCeIssuerAltName:
        memExtension.Lines.Add('Issuer alternative name:'#13#10 +
          Ext.ExtnValue.AsCe_IssuerAltName.Text);
      eveIdCeBasicConstraints:
        begin
          memExtension.Lines.Add('Basic constraints:');
          if Ext.ExtnValue.AsCe_BasicConstraints.CA then begin
            memExtension.Lines.Add('  This certificate may serve as a CA certificate');
            if Ext.ExtnValue.AsCe_BasicConstraints.PathLenConstraint.Length = 0 then
              memExtension.Lines.Add('  There may be any number of CA certificates in any chain depending on this certificate')
            else begin
              memExtension.Lines.Add(
                Format('  There may be at most %d CA certificates in any chain depending on this certificate',
                       [Ext.ExtnValue.AsCe_BasicConstraints.PathLenConstraint.AsInteger]));
              if Ext.ExtnValue.AsCe_BasicConstraints.PathLenConstraint.AsInteger = 0 then
                memExtension.Lines.Add('  The CA who this certificate belongs to may only sign end entiy certificates');
            end;
          end else
            memExtension.Lines.Add('  This is an end entity certificate');
        end;
      eveIdCeNameConstraints:
        memExtension.Lines.Add('Name constraints:'#13#10 +
          Ext.ExtnValue.AsCe_NameConstraints.Text);
      eveIdCePolicyConstraints:
        begin
          memExtension.Lines.Add('Policy constraints:');
          if Ext.ExtnValue.AsCe_PolicyConstraints.RequireExplicitPolicy.Length > 0 then
            memExtension.Lines.Add(
              Format('  An explicit policy must be stated after at least %d certificates down the chain',
                     [Ext.ExtnValue.AsCe_PolicyConstraints.RequireExplicitPolicy.AsInteger]));
          if Ext.ExtnValue.AsCe_PolicyConstraints.InhibitPolicyMapping.Length > 0 then
            memExtension.Lines.Add(
              Format('  Policy mappings are allowed in at most %d certificates down the chain',
                     [Ext.ExtnValue.AsCe_PolicyConstraints.InhibitPolicyMapping.AsInteger]));
        end;
      eveIdCeCRLDistributionPoints:
        begin
          memExtension.Lines.Add('CRL distribution points:');
          for J := 0 to Ext.ExtnValue.AsCe_CRLDistributionPoints.Count - 1 do begin
            memExtension.Lines.Add('  Distribution point: ' +
              Ext.ExtnValue.AsCe_CRLDistributionPoints.Items[J].DistributionPoint.Text);
            memExtension.Lines.Add('  Reasons (if not all):');
            if Ext.ExtnValue.AsCe_CRLDistributionPoints.Items[J].Reasons.Bits[Ord(crlrKeyCompromise)] then
              memExtension.Lines.Add('    Key compromise');
            if Ext.ExtnValue.AsCe_CRLDistributionPoints.Items[J].Reasons.Bits[Ord(crlrCACompromise)] then
              memExtension.Lines.Add('    CA Compromise');
            if Ext.ExtnValue.AsCe_CRLDistributionPoints.Items[J].Reasons.Bits[Ord(crlrAffiliationChanged)] then
              memExtension.Lines.Add('    Affiliation Changed');
            if Ext.ExtnValue.AsCe_CRLDistributionPoints.Items[J].Reasons.Bits[Ord(crlrSuperseded)] then
              memExtension.Lines.Add('    Superseded');
            if Ext.ExtnValue.AsCe_CRLDistributionPoints.Items[J].Reasons.Bits[Ord(crlrCessationOfOperation)] then
              memExtension.Lines.Add('    Cessation of Operation');
            if Ext.ExtnValue.AsCe_CRLDistributionPoints.Items[J].Reasons.Bits[Ord(crlrCertificateHold)] then
              memExtension.Lines.Add('    Certificate Hold');
            if Ext.ExtnValue.AsCe_CRLDistributionPoints.Items[J].Reasons.Bits[Ord(crlrRemoveFromCRL)] then
              memExtension.Lines.Add('    Remove from CRL');
            memExtension.Lines.Add('  CRL Issuer (if different from certificate issuer):' +
              Ext.ExtnValue.AsCe_CRLDistributionPoints[J].CRLIssuer.Text);
          end;
        end;
      eveIdCeSubjectDirectoryAttributes:
        begin
          memExtension.Lines.Add('Subject directory attributes:');
          for J := 0 to Ext.ExtnValue.AsCe_SubjectDirectoryAttributes.Count - 1 do begin
            memExtension.Lines.Add('  ' +
              GetObjectName(Ext.ExtnValue.AsCe_SubjectDirectoryAttributes[J].AttributeType) + ': ');
            for K := 0 to Ext.ExtnValue.AsCe_SubjectDirectoryAttributes[J].AttributeValue.Count - 1 do
              memExtension.Lines.Add('    ' +
                Ext.ExtnValue.AsCe_SubjectDirectoryAttributes[J].AttributeValue[0].Value);
          end;
        end;
      eveIdPeAuthorityInfoAccess:
        begin
          memExtension.Lines.Add('Authority info access:');
          for J := 0 to Ext.ExtnValue.AsPe_AuthorityInfoAccess.Count - 1 do begin
            memExtension.Lines.Add('  ' +
              GetObjectName(Ext.ExtnValue.AsPe_AuthorityInfoAccess[I].AccessMethod) + ':');
            memExtension.Lines.Add('    ' +
              Ext.ExtnValue.AsPe_AuthorityInfoAccess[I].AccessLocation.Text)
          end;
        end;
      eveNetscapeCertType:
        begin
          memExtension.Lines.Add('Netscape Certificate Type:');
          if Ext.ExtnValue.AsNetscapeCertType.Bits[0] then
            memExtension.Lines.Add('SSL client');              
          if Ext.ExtnValue.AsNetscapeCertType.Bits[1] then
            memExtension.Lines.Add('SSL server');
          if Ext.ExtnValue.AsNetscapeCertType.Bits[2] then
            memExtension.Lines.Add('S/MIME');
          if Ext.ExtnValue.AsNetscapeCertType.Bits[3] then
            memExtension.Lines.Add('Object Signing');
          if Ext.ExtnValue.AsNetscapeCertType.Bits[5] then
            memExtension.Lines.Add('SSL CA');
          if Ext.ExtnValue.AsNetscapeCertType.Bits[6] then
            memExtension.Lines.Add('S/MIME CA');
          if Ext.ExtnValue.AsNetscapeCertType.Bits[7] then
            memExtension.Lines.Add('Object Signing CA');
        end;
    end;
    memExtension.Lines.Add('');
    memExtension.Lines.Add('');
    memExtension.Lines.Add('(extValue (hex): ' + Ext.ExtnValue.Value + ')');
  end;
end;

procedure TCertificateViewer.SetControlProperties;
begin
  Width := 617;
  Height := 287;
  BevelInner := bvNone;
  BevelOuter := bvNone;
  BevelWidth := 1;

  pcCertificateViewer.Parent := Self;
  pcCertificateViewer.Align := alClient;

  tabSubject.PageControl := pcCertificateViewer;
  tabSubject.Caption := 'Subject';
  with lvSubject do begin
    Parent := tabSubject;
    Align := alClient;
    with Columns.Add do begin
      Caption := 'Type';
      Width := 100;
    end;
    with Columns.Add do begin
      Caption := 'Value';
      Width := 500;
    end;
    ReadOnly := True;
    ViewStyle := vsReport;
  end;

  tabIssuer.PageControl := pcCertificateViewer;
  tabIssuer.Caption := 'Issuer';
  with lvIssuer do begin
    Parent := tabIssuer;
    Align := alClient;
    with Columns.Add do begin
      Caption := 'Type';
      Width := 100;
    end;
    with Columns.Add do begin
      Caption := 'Value';
      Width := 500;
    end;
    ReadOnly := True;
    ViewStyle := vsReport;
  end;

  tabFields.PageControl := pcCertificateViewer;
  tabFields.Caption := 'Fields';
  with Label1 do begin
    Parent := tabFields;
    Left := 8;
    Top := 8;
    Width := 35;
    Height := 13;
    Caption := 'Version';
  end;
  with lblVersion do begin
    Parent := tabFields;
    Left := 112;
    Top := 8;
    Width := 45;
    Height := 13;
    Caption := '';
  end;
  with Label3 do begin
    Parent := tabFields;
    Left := 8;
    Top := 24;
    Width := 64;
    Height := 13;
    Caption := 'Serial number';
  end;
  with lblSerial do begin
    Parent := tabFields;
    Left := 112;
    Top := 24;
    Width := 36;
    Height := 13;
    Caption := '';
  end;
  with Label5 do begin
    Parent := tabFields;
    Left := 8;
    Top := 40;
    Width := 33;
    Height := 13;
    Caption := 'Validity';
  end;
  with Label6 do begin
    Parent := tabFields;
    Left := 16;
    Top := 56;
    Width := 50;
    Height := 13;
    Caption := 'Not before';
  end;
  with lblNotBefore do begin
    Parent := tabFields;
    Left := 112;
    Top := 56;
    Width := 58;
    Height := 13;
    Caption := '';
  end;
  with Label8 do begin
    Parent := tabFields;
    Left := 16;
    Top := 72;
    Width := 41;
    Height := 13;
    Caption := 'Not after';
  end;
  with lblNotAfter do begin
    Parent := tabFields;
    Left := 112;
    Top := 72;
    Width := 49;
    Height := 13;
    Caption := '';
  end;
  with Label10 do begin
    Parent := tabFields;
    Left := 8;
    Top := 88;
    Width := 90;
    Height := 13;
    Caption := 'Signature algorithm';
  end;
  with lblSignature do begin
    Parent := tabFields;
    Left := 112;
    Top := 88;
    Width := 55;
    Height := 13;
    Caption := '';
  end;
  with Label12 do begin
    Parent := tabFields;
    Left := 8;
    Top := 104;
    Width := 58;
    Height := 13;
    Caption := 'Subject UID';
  end;
  with lblSUID do begin
    Parent := tabFields;
    Left := 112;
    Top := 104;
    Width := 36;
    Height := 13;
    Caption := '';
  end;
  with Label14 do begin
    Parent := tabFields;
    Left := 8;
    Top := 120;
    Width := 50;
    Height := 13;
    Caption := 'Issuer UID';
  end;
  with lblIUID do begin
    Parent := tabFields;
    Left := 112;
    Top := 120;
    Width := 32;
    Height := 13;
    Caption := '';
  end;
  with Label15 do begin
    Parent := tabFields;
    Left := 8;
    Top := 136;
    Width := 45;
    Height := 13;
    Caption := 'Signature';
  end;
  with memSignature do begin
    Parent := tabFields;
    Left := 112;
    Top := 136;
    Width := 489;
    Height := 113;
    ReadOnly := True;
    ScrollBars := ssVertical;
  end;

  tabPublicKey.PageControl := pcCertificateViewer;
  tabPublicKey.Caption := 'Public Key';
  with memSubjectPublicKey do begin
    Parent := tabPublicKey;
    Align := alClient;
    ReadOnly := True;
  end;

  tabExtensions.PageControl := pcCertificateViewer;
  tabExtensions.TabVisible := False;
  tabExtensions.Caption := 'Extensions';
  with lbExtensions do begin
    Parent := tabExtensions;
    Width := 169;
    Align := alLeft;
    OnClick := lbExtensionsClick;
  end;
  with memExtension do begin
    Parent := tabExtensions;
    ReadOnly := True;
    Align := alClient;
  end;

  tabRevokation.PageControl := pcCertificateViewer;
  tabRevokation.TabVisible := False;
  tabRevokation.Caption := 'Revokation Status';
  with memRevokation do begin
    Parent := tabRevokation;
    Align := alClient;
    ReadOnly := True;
  end;

  tabCRLs.PageControl := pcCertificateViewer;
  tabCRLs.TabVisible := False;
  tabCRLs.Caption := 'CRLs';
  with memCRLs do begin
    Parent := tabCRLs;
    Align := alClient;
    ReadOnly := True;
  end;                         

  tabSMIMECapabilities.PageControl := pcCertificateViewer;
  tabSMIMECapabilities.TabVisible := False;
  tabSMIMECapabilities.Caption := 'S/MIME';
  with memSMIMECapabilities do begin
    Parent := tabSMIMECapabilities;
    Align := alClient;
    ReadOnly := True;
  end;
end;

procedure TCertificateViewer.SetParent(const AParent: TWidgetControl);
begin
  inherited SetParent(AParent);
  if Assigned(AParent) then SetControlProperties;
end;

end.
