unit SelectStoreDlg;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, 
  Buttons, ExtCtrls;

type
  TfrmSelectStoreDlg = class(TForm)
    OKBtn: TButton;
    CancelBtn: TButton;
    rgSelectStore: TRadioGroup;
    Label1: TLabel;
    Label2: TLabel;
    btnView: TButton;
    procedure btnViewClick(Sender: TObject);
  private
    FFileName: string;
    FMimeStr: string;
  public
    function Execute(Save: Boolean; const FileName: string): Boolean;
    function ExecuteMIME(const MimeStr: string): Boolean;
  end;

var
  frmSelectStoreDlg: TfrmSelectStoreDlg;

implementation

uses
  SsBase64, Asn1, X509Base, Pkix_Cert, CertModule, CertViewDlg;

{$R *.dfm}

{ TfrmSelectStoreDlg }

function TfrmSelectStoreDlg.Execute(Save: Boolean;
  const FileName: string): Boolean;
begin
  btnView.Visible := False;
  FFileName := FileName;
  FMimeStr := '';
  if not Save then begin
    if (FileName = '') or (ExtractFileExt(FileName) = '.cer') then begin
      Label1.Caption := 'Select the store were the certificate should be installed:';
      btnView.Visible := (FileName <> '') and FileExists(FileName);
    end else
      Label1.Caption := 'Select the store were the certificates should be installed:'
  end else
    Label1.Caption := 'Select the store to be exported:';
  Result := ShowModal = mrOK;
end;

procedure TfrmSelectStoreDlg.btnViewClick(Sender: TObject);
var
  T: TCertificate;
  SS: TStringStream;
  RC: TX509RevokedCertificate;
  C: TASN1Struct;
  S: PASN1Struct;
begin
  T := TCertificate.Create(nil,nil);
  try
    if FMimeStr = '' then
      T.LoadFromFile(FFileName)
    else begin
      SS := TStringStream.Create(MIME64ToStr('-----BEGIN CERTIFICATE-----',
                                             '-----END CERTIFICATE-----',
                                             FMIMEStr));
      try
        T.LoadFromStream(SS);
      finally
        SS.Free;
      end;
    end;
    RC := nil;
    try
      if not dmCert.RootCerts.FindRevocation(T.Data,RC) then
        if not dmCert.MyCerts.FindRevocation(T.Data,RC) then
          if not dmCert.IssuedCerts.FindRevocation(T.Data,RC) then
            dmCert.CollectedCerts.FindRevocation(T.Data,RC);
      C := nil;
      dmCert.GetCRLs(C);
      S := nil;
      dmCert.CollectedCerts.FindSMIMECapabilities(T.Data,S);
      if Assigned(S) then
        frmCertificateViewer.CertificateViewer1.Display(T,RC,C,S^)
      else
        frmCertificateViewer.CertificateViewer1.Display(T,RC,C,nil);
      frmCertificateViewer.ShowModal;
    finally
      RC.Free;
    end;
  finally
    T.Free;
  end;
end;

function TfrmSelectStoreDlg.ExecuteMIME(const MimeStr: string): Boolean;
begin
  btnView.Visible := False;
  FMimeStr := MimeStr;
  Label1.Caption := 'Select the store were the certificate should be installed:';
  btnView.Visible := True;
  Result := ShowModal = mrOK;
end;

end.
