unit YarrowXModule;

interface

uses
  Windows, Messages, SysUtils, Classes, SyncObjs,
  MpYarrow;

type
  TdmYarrow = class(TDataModule)
    MPYarrow1: TMPYarrow;
    procedure DataModuleCreate(Sender: TObject);
    procedure MPYarrow1Reseeded(Sender: TObject; Slow: Boolean);
    procedure MPYarrow1Reseeding(Sender: TObject; Slow: Boolean);
    procedure DataModuleDestroy(Sender: TObject);
  private
    FLock: TCriticalSection;
    FSlowReseeds: Integer;
    FFastReseeds: Integer;
    FReseedThreads: Integer;
    FOnReseed: TNotifyEvent;
    FOnReseeding: TNotifyEvent;
    procedure SetOnReseed(const Value: TNotifyEvent);
    procedure SetOnReseeding(const Value: TNotifyEvent);
    { Private declarations }
  public
    procedure Accumulate(Seed: TCustomMemoryStream);
    function CanUnload: Boolean;
    procedure ForceSlowReseed(const Buf; Count: Integer);
    procedure Generate(Seed: TCustomMemoryStream);
    property FastReseeds: Integer read FFastReseeds;
    property SlowReseeds: Integer read FSlowReseeds;
    property ReseedThreads: Integer read FReseedThreads;
    property OnReseed: TNotifyEvent read FOnReseed write SetOnReseed;
    property OnReseeding: TNotifyEvent read FOnReseeding write SetOnReseeding;
  end;

var
  dmYarrow: TdmYarrow;

implementation

uses
  SecUtils;

{$R *.DFM}

procedure TdmYarrow.DataModuleCreate(Sender: TObject);
begin
  if Assigned(dmYarrow) and (dmYarrow <> Self) then
    raise Exception.Create('There can be only one instance of TdmYarrow per application');
  MPYarrow1.SourceCount := 2;
  MPYarrow1.ForceSlowReseed(nil^,0,1000);
  FLock := TCriticalSection.Create;
end;

procedure TdmYarrow.SetOnReseed(const Value: TNotifyEvent);
begin
  FOnReseed := Value;
end;

procedure TdmYarrow.MPYarrow1Reseeded(Sender: TObject; Slow: Boolean);
begin
  if FReseedThreads > 0 then
    InterlockedDecrement(FReseedThreads);
  if Slow then
    InterlockedIncrement(FSlowReseeds)
  else
    InterlockedIncrement(FFastReseeds);
  if Assigned(FOnReseed) then
    FOnReseed(Self);
end;

procedure TdmYarrow.ForceSlowReseed(const Buf; Count: Integer);
begin
  MPYarrow1.Accumulate(Buf,Count,0,Count shr 1);
  MPYarrow1.ForceSlowReseed(Buf,Count,50);
end;

procedure TdmYarrow.MPYarrow1Reseeding(Sender: TObject; Slow: Boolean);
begin
  InterlockedIncrement(FReseedThreads);
  if Assigned(FonReseeding) then
    FOnReseeding(Self);
end;

procedure TdmYarrow.SetOnReseeding(const Value: TNotifyEvent);
begin
  FOnReseeding := Value;
end;

procedure TdmYarrow.DataModuleDestroy(Sender: TObject);
begin
  dmYarrow := nil;
  FLock.Free;
end;

function TdmYarrow.CanUnload: Boolean;
begin
  Result := FReseedThreads = 0;
end;

procedure TdmYarrow.Accumulate(Seed: TCustomMemoryStream);
begin
  FLock.Acquire;
  try
    MPYarrow1.Accumulate(Seed.Memory^,Seed.Size,1,0);
  finally
    FLock.Release;
  end;
end;

procedure TdmYarrow.Generate(Seed: TCustomMemoryStream);
begin
  FLock.Acquire;
  try
    MPYarrow1.Generate(Seed.Memory^,Seed.Size);
  finally
    FLock.Release;
  end;
end;

end.
