unit CAClientMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls,
  SecUtils, SecComp, StreamSecII, MpX509, TlsInternalServer;

type
  TForm2 = class(TForm)
    Label5: TLabel;
    edtClientRegToken: TEdit;
    Label3: TLabel;
    edtClientCommonName: TEdit;
    Label4: TLabel;
    edtClientOrganizationalUnit: TEdit;
    btnCreateReq: TButton;
    btnImport: TButton;
    TLSServer: TSimpleTLSInternalServer;
    PrivateKeyRing: TSsPrivateKeyRingComponent;
    SaveDialog1: TSaveDialog;
    OpenDialog1: TOpenDialog;
    procedure PrivateKeyRingAdminPassword(Sender: TObject;
      Password: ISecretKey);
    procedure PrivateKeyRingPassword(Sender: TObject;
      Password: ISecretKey);
    procedure btnCreateReqClick(Sender: TObject);
    procedure TLSServerBeforeCreateNewCertDlg(Sender: TObject;
      ACreateType: TNewCertCreateEnum; var ANewCertDlg: INewCertDlg);
    procedure btnImportClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form2: TForm2;

implementation

uses
  SsBase64, Asn1, ImportDlg, ExportDlg,
  // The following unit was created by the NewCertDlg wizard:
  ClientCertReqCreator;

{$R *.dfm}

procedure TForm2.PrivateKeyRingAdminPassword(Sender: TObject;
  Password: ISecretKey);
begin
  Password.SetKeyStr('abc');
end;

procedure TForm2.PrivateKeyRingPassword(Sender: TObject;
  Password: ISecretKey);
begin
  Password.SetKeyStr('def');
end;

procedure TForm2.btnCreateReqClick(Sender: TObject);
var
  MS: TMemoryStream;
  ExportDlg: TfrmExportDlg;
  Cert: string;
begin
  MS := TMemoryStream.Create;
  try
    if not TLSServer.NewClientCertReqDlg(MS) then
      raise Exception.Create('Unable to create certification request');
    MS.Position := 0;
    SetLength(Cert,MS.Size);
    MS.Read(Pointer(Cert)^,Length(Cert));
    ExportDlg := TfrmExportDlg.Create(nil);
    try
      ExportDlg.memImport.Text := StrToMIME64('-----BEGIN CERTIFICATE REQUEST-----',
                                              '-----END CERTIFICATE REQUEST-----',
                                              Cert);
      case ExportDlg.ShowModal of
        mrAbort:
          if SaveDialog1.Execute then
            MS.SaveToFile(SaveDialog1.FileName);
      end;
    finally
      ExportDlg.Free;
    end;
  finally
    MS.Free;
  end;
end;

procedure TForm2.TLSServerBeforeCreateNewCertDlg(Sender: TObject;
  ACreateType: TNewCertCreateEnum; var ANewCertDlg: INewCertDlg);
var
  CertReq: IClientCertReqCreator;
begin
  case ACreateType of
    nccCertReq:
      begin
        CertReq := TClientCertReqCreator.Create;
        CertReq.CommonName := edtClientCommonName.Text;
        CertReq.OrganizationalUnitName := edtClientOrganizationalUnit.Text;
        CertReq.RegToken := edtClientRegToken.Text;
        ANewCertDlg := CertReq;
      end;
  end;
end;

procedure TForm2.btnImportClick(Sender: TObject);
var
  ImportDlg: TfrmImportDlg;
  ImportData: TStream;
  Crt: TAsn1Struct;
  Cert: string;
  Status: TCertStatusCode;
begin
  ImportData := nil;
  ImportDlg := TfrmImportDlg.Create(nil);
  try
    case ImportDlg.ShowModal of
      mrOK:
        begin
          Cert := MIME64ToStr('-----BEGIN CERTIFICATE-----',
                              '-----END CERTIFICATE-----',
                              ImportDlg.memImport.Text);
          if Cert = '' then
            raise Exception.Create('No Certificate was selected');
          ImportData := TStringStream.Create(Cert);
        end;
      mrCancel:
        raise Exception.Create('No Certificate was selected');
      mrAbort:
        if OpenDialog1.Execute then
          ImportData := TFileStream.Create(OpenDialog1.FileName,fmOpenRead)
        else
          raise Exception.Create('No Certificate was selected');
    end;
    Crt := TASN1Struct.Create;
    try
      Crt.LoadFromStream(ImportData,fmtDER);
      TLSServer.AddCertificate(Crt,False,Status);
      if Status = crcOK then
        TLSServer.ExportToPFX('Client.pfx',TSecretKey.CreateBMPStr('somepassword'));
    finally
      Crt.Free;
    end;
  finally
    ImportDlg.Free;
    ImportData.Free;
  end;
end;

procedure TForm2.FormCreate(Sender: TObject);
begin
  Sleep(1000);
  Application.ProcessMessages;
  TLSServer.AccumulateEntropy;
  if FileExists('Root.cer') then
    TLSServer.LoadRootCertsFromFile('Root.cer');
  if FileExists('Client.pfx') then try
    TLSServer.ImportFromPFX('Client.pfx',TSecretKey.CreateBMPStr('somepassword'));
    btnCreateReq.Enabled := False;
  except
  end;
end;

end.
