unit CAMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls,
  SecUtils,
  SsUserList, MpX509, TlsInternalServer, SsCertMgr,
  Pkix_Cert, SecComp, StreamSecII,
  CertificateViewer, CertificateSelection;

type
  TForm1 = class(TForm)
    btnCreateRoot: TButton;
    X509CertificateAuthority: TX509CertificateAuthority;
    PrivateKeyRing: TSsPrivateKeyRingComponent;
    Label1: TLabel;
    edtRootCommonName: TEdit;
    Label2: TLabel;
    edtRootOrganization: TEdit;
    Label3: TLabel;
    edtServerCommonName: TEdit;
    Label4: TLabel;
    edtServerOrganization: TEdit;
    btnCreateServer: TButton;
    Label5: TLabel;
    edtServerDNSName: TEdit;
    CertificateSelection1: TCertificateSelection;
    btnSignRequest: TButton;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    UserList: TSsEventUserList;
    procedure PrivateKeyRingAdminPassword(Sender: TObject;
      Password: ISecretKey);
    procedure PrivateKeyRingPassword(Sender: TObject;
      Password: ISecretKey);
    procedure X509CertificateAuthorityBeforeCreateNewCertDlg(
      Sender: TObject; ACreateType: TNewCertCreateEnum;
      var ANewCertDlg: INewCertDlg);
    procedure btnCreateRootClick(Sender: TObject);
    procedure btnCreateServerClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnSignRequestClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure UserListVerifyUser(ASubjectName: TRdnsequence;
      ASubjectAltName: TGeneralNames; const ARegToken: WideString;
      var AVerified: Boolean);
  private
    FRootCert: TCertificate;
    FServerCert: TCertificate;
    FClientCert: TCertificate;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses
  SsBase64, ImportDlg, ExportDlg,
  // These units were created by the NewCertDlg wizard. Select File|New and
  // the StreamSec tab.
  RootCertificateCreator, ServerCertificateCreator, ClientCertificateCreator;

{$R *.dfm}

procedure TForm1.PrivateKeyRingAdminPassword(Sender: TObject;
  Password: ISecretKey);
begin
  Password.SetKeyStr('abc');
end;

procedure TForm1.PrivateKeyRingPassword(Sender: TObject;
  Password: ISecretKey);
begin
  Password.SetKeyStr('def');
end;

procedure TForm1.X509CertificateAuthorityBeforeCreateNewCertDlg(
  Sender: TObject; ACreateType: TNewCertCreateEnum;
  var ANewCertDlg: INewCertDlg);
var
  RootDlg: IRootCertificateCreator;
  CertDlg: IServerCertificateCreator;
  CliCertDlg: IClientCertificateCreator;
begin
  case ACreateType of
    nccRoot:
      begin
        RootDlg := TRootCertificateCreator.Create;
        RootDlg.CommonName := edtRootCommonName.Text;
        RootDlg.OrganizationName := edtRootOrganization.Text;
        ANewCertDlg := RootDlg;
      end;
    nccCert:
      begin
        CertDlg := TServerCertificateCreator.Create;
        CertDlg.DNSName := edtServerDNSName.Text;
        CertDlg.CommonName := edtServerCommonName.Text;
        CertDlg.OrganizationName := edtServerOrganization.Text;
        ANewCertDlg := CertDlg;
      end;
    nccCertFromReq:
      begin
        CliCertDlg := TClientCertificateCreator.Create;
        CliCertDlg.OrganizationName := edtServerOrganization.Text;
        ANewCertDlg := CliCertDlg;
      end;
  end;
end;

procedure TForm1.btnCreateRootClick(Sender: TObject);
begin
  if not X509CertificateAuthority.NewCACertDlg(nil) then
    raise Exception.Create('Unable to create root certificate');

  btnCreateRoot.Enabled := False;
  CertificateSelection1.CertSource := X509CertificateAuthority;
  CertificateSelection1.DisplayAll;
end;

procedure TForm1.btnCreateServerClick(Sender: TObject);
begin
  if not X509CertificateAuthority.NewServerCertDlg(nil) then
    raise Exception.Create('Unable to create server certificate');

  btnCreateServer.Enabled := False;
  CertificateSelection1.CertSource := X509CertificateAuthority;
  CertificateSelection1.DisplayAll;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FRootCert.Free;
  FServerCert.Free;
  FClientCert.Free;
end;

procedure TForm1.btnSignRequestClick(Sender: TObject);
var
  ImportData: TStream;
  ExportData: TMemoryStream;
  ImportDlg: TfrmImportDlg;
  ExportDlg: TfrmExportDlg;
  CertReq, Cert: string;
begin
  ImportData := nil;
  ImportDlg := TfrmImportDlg.Create(nil);
  try
    case ImportDlg.ShowModal of
      mrOK:
        begin
          CertReq := MIME64ToStr('-----BEGIN CERTIFICATE REQUEST-----',
                                 '-----END CERTIFICATE REQUEST-----',
                                 ImportDlg.memImport.Text);
          if CertReq = '' then
            CertReq := MIME64ToStr('-----BEGIN NEW CERTIFICATE REQUEST-----',
                                   '-----END NEW CERTIFICATE REQUEST-----',
                                   ImportDlg.memImport.Text);
          if CertReq = '' then
            raise Exception.Create('No Certificate Request was selected');
          ImportData := TStringStream.Create(CertReq);
        end;
      mrCancel:
        raise Exception.Create('No Certificate Request was selected');
      mrAbort:
        if OpenDialog1.Execute then
          ImportData := TFileStream.Create(OpenDialog1.FileName,fmOpenRead)
        else
          raise Exception.Create('No Certificate Request was selected');
    end;
    ExportData := TMemoryStream.Create;
    try
      if not X509CertificateAuthority.NewCertFromReqDlg(ImportData,ExportData) then
        raise Exception.Create('Unable to create certificate');
      SetLength(Cert,ExportData.Size);
      ExportData.Position := 0;
      ExportData.Read(Pointer(Cert)^,Length(Cert));
      ExportDlg := TfrmExportDlg.Create(nil);
      try
        ExportDlg.memImport.Text := StrToMIME64('-----BEGIN CERTIFICATE-----',
                                                '-----END CERTIFICATE-----',
                                                Cert);
        case ExportDlg.ShowModal of
          mrAbort:
            if SaveDialog1.Execute then
              ExportData.SaveToFile(SaveDialog1.FileName);
        end;
      finally
        ExportDlg.Free;
      end;
    finally
      ExportData.Free;
    end;

    btnSignRequest.Enabled := False;
    CertificateSelection1.CertSource := X509CertificateAuthority;
    CertificateSelection1.DisplayAll;
  finally
    ImportData.Free;
    ImportDlg.Free;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  X509CertificateAuthority.AccumulateEntropy;
  if FileExists('Root.pkr') then try
    X509CertificateAuthority.LoadPrivateKeyRingFromFile('Root.pkr',nil);
    if FileExists('Root.cer') then
      X509CertificateAuthority.LoadRootCertsFromFile('Root.cer');
    if FileExists('Root.scl') then
      if X509CertificateAuthority.LoadMyCertsFromFile('Root.scl') then
        btnCreateRoot.Enabled := False;
    if FileExists('Server.pfx') then try
      X509CertificateAuthority.ImportFromPFX('Server.pfx',TSecretKey.CreateBMPStr('somepassword'));
      btnCreateServer.Enabled := False;
    except
    end;
  except
  end;
  CertificateSelection1.DisplayAll;
end;

procedure TForm1.UserListVerifyUser(ASubjectName: TRdnsequence;
  ASubjectAltName: TGeneralNames; const ARegToken: WideString;
  var AVerified: Boolean);
begin
  AVerified := False;
  if (ASubjectName.CommonName = 'localhost') and
     (ASubjectName.OrganizationalUnitName = 'Test Department') then
    AVerified := ARegToken = 'someregtoken';
end;

end.
