unit ClientCertReqCreator;

interface

uses
  SysUtils,
  SecUtils, Pkix, Asn1, Pkix_Cert, Pkcs_10, X509Base, MpX509,
  StreamSecII;

type
  IClientCertReqCreator = interface(INewCertDlg)
  ['{C472B631-31B8-49C9-9638-0C3B2190DFC6}']
    function GetCommonName: WideString;
    function GetOrganizationalUnitName: WideString;   
    function GetChallengePassword: WideString;
    function GetRegToken: WideString;

    procedure SetCommonName(const Value: WideString);
    procedure SetOrganizationalUnitName(const Value: WideString);
    procedure SetChallengePassword(const Value: WideString);
    procedure SetRegToken(const Value: WideString);

    property CommonName: WideString read GetCommonName write SetCommonName;
    property OrganizationalUnitName: WideString read GetOrganizationalUnitName write SetOrganizationalUnitName;
    property RegToken: WideString read GetRegToken write SetRegToken;
    property ChallengePassword: WideString read GetChallengePassword write SetChallengePassword;
  end;

  TClientCertReqCreator = class(TNewCertDlgObject,IClientCertReqCreator)
  private
    FCommonName: WideString;
    FOrganizationalUnitName: WideString;
  protected
    function GetCommonName: WideString;
    function GetOrganizationalUnitName: WideString;

    procedure SetCommonName(const Value: WideString);
    procedure SetOrganizationalUnitName(const Value: WideString);

    procedure ComposeCertificate; override;
    procedure AfterExecute(ATLSServer: ITLSServer); override;
  end;

implementation

{ TClientCertReqCreator }

procedure TClientCertReqCreator.AfterExecute(ATLSServer: ITLSServer);
begin
  // This is where you should save the key or request to a file
  ATLSServer.SavePrivateKeyRing('Client.pkr');
  // inherited will save P10 to ExportData
  inherited;
end;

procedure TClientCertReqCreator.ComposeCertificate;
var
  Ext: TExtension;
begin
  if CreateOpt <> nccCertReq then
    raise Exception.Create('Create Option not supported');
  // Subject
  P10.Subject.LineSeparator := '/';
  if FCommonName <> '' then
    P10.Subject.CommonName := FCommonName
  else
    P10.Subject.CommonName := 'localhost';
  if FOrganizationalUnitName <> '' then
    P10.Subject.OrganizationalUnitName := FOrganizationalUnitName
  else
    P10.Subject.OrganizationalUnitName := 'Research and Development/TEST TEST TEST';
  P10.Subject.DnQualifier := 'Client';

  // Subject Alt Name
  Ext := P10.CertificationRequestInfo.Extensions.AddUniqueItem(eveIdCeSubjectAltName);
  with Ext.ExtnValue.AsCe_SubjectAltName.Add do begin
    Choice := gneDNSName;
    AsDNSName := 'localhost';
  end;
  Ext := P10.CertificationRequestInfo.Extensions.AddUniqueItem(eveIdCeSubjectAltName);
  with Ext.ExtnValue.AsCe_SubjectAltName.Add do begin
    Choice := gneIPAddress;
    ParseIPAddress('127.0.0.1',AsIPAddress);
  end;
  Ext := P10.CertificationRequestInfo.Extensions.UniqueItem[eveIdCeSubjectAltName];
  if Assigned(Ext) then
    Ext.Critical := True;

  //Private key
  PrivateKeyAlg := rsaEncryption;
  if not (Wizard in [ncwAdvanced,ncwCA]) then
    ExportableRSA := True;
  KeySize := 1024;
  SignatureDigestAlg := haSHA1;

  //RegToken and ChallengePassword
  P10.CertificationRequestInfo.Attributes.RegToken := RegToken;
  P10.CertificationRequestInfo.Attributes.ChallengePassword := ChallengePassword;

  inherited;
end;

function TClientCertReqCreator.GetCommonName: WideString;
begin
  Result := FCommonName;
end;

function TClientCertReqCreator.GetOrganizationalUnitName: WideString;
begin
  Result := FOrganizationalUnitName;
end;

procedure TClientCertReqCreator.SetCommonName(const Value: WideString);
begin
  FCommonName := Value;
end;

procedure TClientCertReqCreator.SetOrganizationalUnitName(const Value: WideString);
begin
  FOrganizationalUnitName := Value;
end;



end.