unit ClientCertificateCreator;

interface

uses
  SysUtils,
  SecUtils, Pkix, Asn1, Pkix_Cert, Pkcs_10, X509Base, MpX509,
  StreamSecII;

type
  IClientCertificateCreator = interface(INewCertDlg)
  ['{1A84F9DC-5317-41C0-96CA-0D72A1DD0C97}']
    function GetOrganizationName: WideString;

    procedure SetOrganizationName(const Value: WideString);

    property OrganizationName: WideString read GetOrganizationName write SetOrganizationName;

  end;

  TClientCertificateCreator = class(TNewCertDlgObject,IClientCertificateCreator)
  private
    FOrganizationName: WideString;
  protected
    function GetOrganizationName: WideString;

    procedure SetOrganizationName(const Value: WideString);

    procedure ComposeCertificate; override;
    procedure AfterExecute(ATLSServer: ITLSServer); override;
  end;

implementation

{ TClientCertificateCreator }

procedure TClientCertificateCreator.AfterExecute(ATLSServer: ITLSServer);
begin
  // This is where you should save the new certificate
  // inherited will save Cert to ExportData
  inherited;
end;

procedure TClientCertificateCreator.ComposeCertificate;
var
  Idx: Integer;
  Ext: TExtension;
begin
  if CreateOpt <> nccCertFromReq then
    raise Exception.Create('Create Option not supported');

  //Select issuer
  Idx := MyCerts.Count - 1;
  while Idx >= 0 do begin
    CACert := MyCerts.Certs[Idx];
    if CheckValidity(CACert,OffsetFromUTC) and (keyCertSign in ExtractKeyUsage(CACert)) then 
      Break;
    Dec(Idx);
  end;

  //Verify RegToken (Example)
  if not Verified then
    raise Exception.Create('Unable to verify the request');
  //Subject Name from request
  Cert.Subject.Assign(NewCert.Subject);

  //Subject Alt Name from request
  Ext := NewCert.TbsCertificate.Extensions.UniqueItem[eveIdCeSubjectAltName];
  if Assigned(Ext) then
    Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeSubjectAltName).Assign(Ext);

  // The following assignments are CA overrides of the values entered by the client
  // Subject
  Cert.Subject.LineSeparator := '/';
  Cert.Subject.OrganizationName := FOrganizationName;

  //Validity
  Cert.Validity.NotBefore.Choice := teUTCTime;
  Cert.Validity.NotBefore.AsUtcTime.AsDateTime := Trunc(Now - OffsetFromUTC/24);
  Cert.Validity.NotAfter.Choice := teUTCTime;
  Cert.Validity.NotAfter.AsUtcTime.AsDateTime := Trunc(Now - OffsetFromUTC/24 + 380);

  //Key usage (for client certificate; change if other)
  if SubjectPublicKeyInfo.Algorithm.Algorithm = rsaEncryption then
    Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature,keyEncipherment]
  else if SubjectPublicKeyInfo.Algorithm.Algorithm = id_dsa then
    Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature]
  else if SubjectPublicKeyInfo.Algorithm.Algorithm = id_ecPublicKey then
    Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature];

  //Extended Key Usage (for client certificate; change if other)
  Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeExtKeyUsage);
  Ext.ExtnValue.AsCe_ExtKeyUsage.AddValue(id_kp_clientAuth);

  //Basic constraints (for client certificate; change if other)
  Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeBasicConstraints);
  Ext.ExtnValue.AsCe_BasicConstraints.CA := False;

  //Additional extensions can be set here, such as CRL Distribution Point,
  //Certificate Policies, Name Constraints, etc

  //The Private key is unknown at this point. The public key
  //will be copied from the NewCert object to the Cert object by
  //the TLSServer component

  inherited;
end;

function TClientCertificateCreator.GetOrganizationName: WideString;
begin
  Result := FOrganizationName;
end;

procedure TClientCertificateCreator.SetOrganizationName(const Value: WideString);
begin
  FOrganizationName := Value;
end;



end.