unit RootCertificateCreator;

interface

uses
  SysUtils,
  SecUtils, Pkix, Asn1, Pkix_Cert, Pkcs_10, X509Base, MpX509,
  StreamSecII;

type
  IRootCertificateCreator = interface(INewCertDlg)
  ['{E0A5B3DD-0355-49C2-AB72-642F45052518}']
    function GetCommonName: WideString;
    function GetSurname: WideString;
    function GetGivenName: WideString;
    function GetInitials: WideString;
    function GetGenerationQualifier: WideString;
    function GetCountry: WideString;
    function GetLocalityName: WideString;
    function GetStateOrProvinceName: WideString;
    function GetOrganizationName: WideString;
    function GetOrganizationalUnitName: WideString;
    function GetDnQualifier: WideString;

    procedure SetCommonName(const Value: WideString);
    procedure SetSurname(const Value: WideString);
    procedure SetGivenName(const Value: WideString);
    procedure SetInitials(const Value: WideString);
    procedure SetGenerationQualifier(const Value: WideString);
    procedure SetCountry(const Value: WideString);
    procedure SetLocalityName(const Value: WideString);
    procedure SetStateOrProvinceName(const Value: WideString);
    procedure SetOrganizationName(const Value: WideString);
    procedure SetOrganizationalUnitName(const Value: WideString);
    procedure SetDnQualifier(const Value: WideString);

    property CommonName: WideString read GetCommonName write SetCommonName;
    property Surname: WideString read GetSurname write SetSurname;
    property GivenName: WideString read GetGivenName write SetGivenName;
    property Initials: WideString read GetInitials write SetInitials;
    property GenerationQualifier: WideString read GetGenerationQualifier write SetGenerationQualifier;
    property Country: WideString read GetCountry write SetCountry;
    property LocalityName: WideString read GetLocalityName write SetLocalityName;
    property StateOrProvinceName: WideString read GetStateOrProvinceName write SetStateOrProvinceName;
    property OrganizationName: WideString read GetOrganizationName write SetOrganizationName;
    property OrganizationalUnitName: WideString read GetOrganizationalUnitName write SetOrganizationalUnitName;
    property DnQualifier: WideString read GetDnQualifier write SetDnQualifier;

  end;

  TRootCertificateCreator = class(TNewCertDlgObject,IRootCertificateCreator)
  private
    FCommonName: WideString;
    FSurname: WideString;
    FGivenName: WideString;
    FInitials: WideString;
    FGenerationQualifier: WideString;
    FCountry: WideString;
    FLocalityName: WideString;
    FStateOrProvinceName: WideString;
    FOrganizationName: WideString;
    FOrganizationalUnitName: WideString;
    FDnQualifier: WideString;
  protected
    function GetCommonName: WideString;
    function GetSurname: WideString;
    function GetGivenName: WideString;
    function GetInitials: WideString;
    function GetGenerationQualifier: WideString;
    function GetCountry: WideString;
    function GetLocalityName: WideString;
    function GetStateOrProvinceName: WideString;
    function GetOrganizationName: WideString;
    function GetOrganizationalUnitName: WideString;
    function GetDnQualifier: WideString;

    procedure SetCommonName(const Value: WideString);
    procedure SetSurname(const Value: WideString);
    procedure SetGivenName(const Value: WideString);
    procedure SetInitials(const Value: WideString);
    procedure SetGenerationQualifier(const Value: WideString);
    procedure SetCountry(const Value: WideString);
    procedure SetLocalityName(const Value: WideString);
    procedure SetStateOrProvinceName(const Value: WideString);
    procedure SetOrganizationName(const Value: WideString);
    procedure SetOrganizationalUnitName(const Value: WideString);
    procedure SetDnQualifier(const Value: WideString);

    procedure ComposeCertificate; override;
    procedure AfterExecute(ATLSServer: ITLSServer); override;
  end;

implementation

{ TRootCertificateCreator }

procedure TRootCertificateCreator.AfterExecute(ATLSServer: ITLSServer);
begin
  // This is where you should save the new certificate
  ATLSServer.SavePrivateKeyRing('Root.pkr');
  ATLSServer.SaveMyCertsToSCLFile('Root.scl');
  Cert.SaveToFile('Root.cer');
  // inherited will save Cert to ExportData
  inherited;
end;

procedure TRootCertificateCreator.ComposeCertificate;
var
  Ext: TExtension;
begin
  if not (CreateOpt in [nccRoot,nccCert]) then
    raise Exception.Create('Create Option not supported');
  if not (Wizard in [ncwAdvanced,ncwCA]) then
    raise Exception.Create('Create Option not supported');

  // Subject
  Cert.Subject.LineSeparator := '/';
  Cert.Subject.CommonName := FCommonName;
  Cert.Subject.Surname := FSurname;
  Cert.Subject.GivenName := FGivenName;
  Cert.Subject.Initials := FInitials;
  Cert.Subject.GenerationQualifier := FGenerationQualifier;
  Cert.Subject.Country := FCountry;
  Cert.Subject.LocalityName := FLocalityName;
  Cert.Subject.StateOrProvinceName := FStateOrProvinceName;
  Cert.Subject.OrganizationName := FOrganizationName;
  Cert.Subject.OrganizationalUnitName := FOrganizationalUnitName;
  Cert.Subject.DnQualifier := FDnQualifier;

  //Validity
  Cert.Validity.NotBefore.Choice := teUTCTime;
  Cert.Validity.NotBefore.AsUtcTime.AsDateTime := Trunc(Now - OffsetFromUTC/24);
  Cert.Validity.NotAfter.Choice := teUTCTime;
  Cert.Validity.NotAfter.AsUtcTime.AsDateTime := Trunc(Now - OffsetFromUTC/24 + 3800);

  //Key usage
  Cert.TbsCertificate.Extensions.KeyUsage := [keyCertSign,cRLSign];

  //Basic constraints
  Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeBasicConstraints);
  Ext.ExtnValue.AsCe_BasicConstraints.CA := True;
  // Change PathLen to allow intermediary CAs in the chain:
  Ext.ExtnValue.AsCe_BasicConstraints.PathLenConstraint.AsInteger := 0;

  //Private key
  PrivateKeyAlg := rsaEncryption;
  KeySize := 1024;
  SignatureDigestAlg := haSHA1;

  //Self signed certificate
  CACert := Cert.Data;

  inherited;
end;

function TRootCertificateCreator.GetCommonName: WideString;
begin
  Result := FCommonName;
end;

function TRootCertificateCreator.GetCountry: WideString;
begin
  Result := FCountry;
end;

function TRootCertificateCreator.GetDnQualifier: WideString;
begin
  Result := FDnQualifier;
end;

function TRootCertificateCreator.GetGenerationQualifier: WideString;
begin
  Result := FGenerationQualifier;
end;

function TRootCertificateCreator.GetGivenName: WideString;
begin
  Result := FGivenName;
end;

function TRootCertificateCreator.GetInitials: WideString;
begin
  Result := FInitials;
end;

function TRootCertificateCreator.GetLocalityName: WideString;
begin
  Result := FLocalityName;
end;

function TRootCertificateCreator.GetOrganizationalUnitName: WideString;
begin
  Result := FOrganizationalUnitName;
end;

function TRootCertificateCreator.GetOrganizationName: WideString;
begin
  Result := FOrganizationName;
end;

function TRootCertificateCreator.GetStateOrProvinceName: WideString;
begin
  Result := FStateOrProvinceName;
end;

function TRootCertificateCreator.GetSurname: WideString;
begin
  Result := FSurname;
end;

procedure TRootCertificateCreator.SetCommonName(const Value: WideString);
begin
  FCommonName := Value;
end;

procedure TRootCertificateCreator.SetCountry(const Value: WideString);
begin
  FCountry := Value;
end;

procedure TRootCertificateCreator.SetDnQualifier(const Value: WideString);
begin
  FDnQualifier := Value;
end;

procedure TRootCertificateCreator.SetGenerationQualifier(const Value: WideString);
begin
  FGenerationQualifier := Value;
end;

procedure TRootCertificateCreator.SetGivenName(const Value: WideString);
begin
  FGivenName := Value;
end;

procedure TRootCertificateCreator.SetInitials(const Value: WideString);
begin
  FInitials := Value;
end;

procedure TRootCertificateCreator.SetLocalityName(const Value: WideString);
begin
  FLocalityName := Value;
end;

procedure TRootCertificateCreator.SetOrganizationalUnitName(const Value: WideString);
begin
  FOrganizationalUnitName := Value;
end;

procedure TRootCertificateCreator.SetOrganizationName(const Value: WideString);
begin
  FOrganizationName := Value;
end;

procedure TRootCertificateCreator.SetStateOrProvinceName(const Value: WideString);
begin
  FStateOrProvinceName := Value;
end;

procedure TRootCertificateCreator.SetSurname(const Value: WideString);
begin
  FSurname := Value;
end;



end.