unit ServerCertificateCreator;

interface

uses
  SysUtils,
  SecUtils, Pkix, Asn1, Pkix_Cert, Pkcs_10, X509Base, MpX509,
  StreamSecII;

type
  IServerCertificateCreator = interface(INewCertDlg)
  ['{F38F1F37-5109-48ED-86BE-2958FBBD7E5C}']
    function GetCommonName: WideString;
    function GetSurname: WideString;
    function GetGivenName: WideString;
    function GetInitials: WideString;
    function GetGenerationQualifier: WideString;
    function GetCountry: WideString;
    function GetLocalityName: WideString;
    function GetStateOrProvinceName: WideString;
    function GetOrganizationName: WideString;
    function GetOrganizationalUnitName: WideString;
    function GetDnQualifier: WideString;
    function GetRfc822Name: WideString;
    function GetUniformResourceIdentifier: WideString;
    function GetDNSName: WideString;
    function GetIPAddress: WideString;

    procedure SetCommonName(const Value: WideString);
    procedure SetSurname(const Value: WideString);
    procedure SetGivenName(const Value: WideString);
    procedure SetInitials(const Value: WideString);
    procedure SetGenerationQualifier(const Value: WideString);
    procedure SetCountry(const Value: WideString);
    procedure SetLocalityName(const Value: WideString);
    procedure SetStateOrProvinceName(const Value: WideString);
    procedure SetOrganizationName(const Value: WideString);
    procedure SetOrganizationalUnitName(const Value: WideString);
    procedure SetDnQualifier(const Value: WideString);
    procedure SetRfc822Name(const Value: WideString);
    procedure SetUniformResourceIdentifier(const Value: WideString);
    procedure SetDNSName(const Value: WideString);
    procedure SetIPAddress(const Value: WideString);

    property CommonName: WideString read GetCommonName write SetCommonName;
    property Surname: WideString read GetSurname write SetSurname;
    property GivenName: WideString read GetGivenName write SetGivenName;
    property Initials: WideString read GetInitials write SetInitials;
    property GenerationQualifier: WideString read GetGenerationQualifier write SetGenerationQualifier;
    property Country: WideString read GetCountry write SetCountry;
    property LocalityName: WideString read GetLocalityName write SetLocalityName;
    property StateOrProvinceName: WideString read GetStateOrProvinceName write SetStateOrProvinceName;
    property OrganizationName: WideString read GetOrganizationName write SetOrganizationName;
    property OrganizationalUnitName: WideString read GetOrganizationalUnitName write SetOrganizationalUnitName;
    property DnQualifier: WideString read GetDnQualifier write SetDnQualifier;
    property Rfc822Name: WideString read GetRfc822Name write SetRfc822Name;
    property UniformResourceIdentifier: WideString read GetUniformResourceIdentifier write SetUniformResourceIdentifier;
    property DNSName: WideString read GetDNSName write SetDNSName;
    property IPAddress: WideString read GetIPAddress write SetIPAddress;

  end;

  TServerCertificateCreator = class(TNewCertDlgObject,IServerCertificateCreator)
  private
    FCommonName: WideString;
    FSurname: WideString;
    FGivenName: WideString;
    FInitials: WideString;
    FGenerationQualifier: WideString;
    FCountry: WideString;
    FLocalityName: WideString;
    FStateOrProvinceName: WideString;
    FOrganizationName: WideString;
    FOrganizationalUnitName: WideString;
    FDnQualifier: WideString;
    FRfc822Name: WideString;
    FUniformResourceIdentifier: WideString;
    FDNSName: WideString;
    FIPAddress: WideString;
  protected
    function GetCommonName: WideString;
    function GetSurname: WideString;
    function GetGivenName: WideString;
    function GetInitials: WideString;
    function GetGenerationQualifier: WideString;
    function GetCountry: WideString;
    function GetLocalityName: WideString;
    function GetStateOrProvinceName: WideString;
    function GetOrganizationName: WideString;
    function GetOrganizationalUnitName: WideString;
    function GetDnQualifier: WideString;
    function GetRfc822Name: WideString;
    function GetUniformResourceIdentifier: WideString;
    function GetDNSName: WideString;
    function GetIPAddress: WideString;

    procedure SetCommonName(const Value: WideString);
    procedure SetSurname(const Value: WideString);
    procedure SetGivenName(const Value: WideString);
    procedure SetInitials(const Value: WideString);
    procedure SetGenerationQualifier(const Value: WideString);
    procedure SetCountry(const Value: WideString);
    procedure SetLocalityName(const Value: WideString);
    procedure SetStateOrProvinceName(const Value: WideString);
    procedure SetOrganizationName(const Value: WideString);
    procedure SetOrganizationalUnitName(const Value: WideString);
    procedure SetDnQualifier(const Value: WideString);
    procedure SetRfc822Name(const Value: WideString);
    procedure SetUniformResourceIdentifier(const Value: WideString);
    procedure SetDNSName(const Value: WideString);
    procedure SetIPAddress(const Value: WideString);

    procedure ComposeCertificate; override;
    procedure AfterExecute(ATLSServer: ITLSServer); override;
  end;

implementation

{ TServerCertificateCreator }

procedure TServerCertificateCreator.AfterExecute(ATLSServer: ITLSServer);
var
  WS: WideString;
  PW: ISecretKey;
begin
  WS := 'somepassword';
  PW := TSecretKey.CreateBMPStr(PWideChar(WS));
  ATLSServer.ExportToPFX('Server.pfx',PW);
  // inherited will save Cert to ExportData
  inherited;
end;

procedure TServerCertificateCreator.ComposeCertificate;
var
  Ext: TExtension;
  Idx: Integer;
begin
  if CreateOpt <> nccCert then
    raise Exception.Create('Create Option not supported');

  // Subject
  Cert.Subject.LineSeparator := '/';
  Cert.Subject.CommonName := FCommonName;
  Cert.Subject.Surname := FSurname;
  Cert.Subject.GivenName := FGivenName;
  Cert.Subject.Initials := FInitials;
  Cert.Subject.GenerationQualifier := FGenerationQualifier;
  Cert.Subject.Country := FCountry;
  Cert.Subject.LocalityName := FLocalityName;
  Cert.Subject.StateOrProvinceName := FStateOrProvinceName;
  Cert.Subject.OrganizationName := FOrganizationName;
  Cert.Subject.OrganizationalUnitName := FOrganizationalUnitName;
  Cert.Subject.DnQualifier := FDnQualifier;

  // Subject Alt Name
  if FRfc822Name <> '' then begin
    Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeSubjectAltName);
    with Ext.ExtnValue.AsCe_SubjectAltName.Add do begin
      Choice := gneRfc822Name;
      AsRfc822Name := FRfc822Name;
    end;
  end;
  if FUniformResourceIdentifier <> '' then begin
    Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeSubjectAltName);
    with Ext.ExtnValue.AsCe_SubjectAltName.Add do begin
      Choice := gneUniformResourceIdentifier;
      AsUniformResourceIdentifier := FUniformResourceIdentifier;
    end;
  end;
  if FDNSName <> '' then begin
    Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeSubjectAltName);
    with Ext.ExtnValue.AsCe_SubjectAltName.Add do begin
      Choice := gneDNSName;
      AsDNSName := FDNSName;
    end;
  end;
  if FIPAddress <> '' then begin
    Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeSubjectAltName);
    with Ext.ExtnValue.AsCe_SubjectAltName.Add do begin
      Choice := gneIPAddress;
      ParseIPAddress(FIPAddress,AsIPAddress);
    end;
  end;

  //Validity
  Cert.Validity.NotBefore.Choice := teUTCTime;
  Cert.Validity.NotBefore.AsUtcTime.AsDateTime := Trunc(Now - OffsetFromUTC/24);
  Cert.Validity.NotAfter.Choice := teUTCTime;
  Cert.Validity.NotAfter.AsUtcTime.AsDateTime := Trunc(Now - OffsetFromUTC/24 + 380);

  //Key usage
  case Wizard of
    ncwCA:          Cert.TbsCertificate.Extensions.KeyUsage := [keyCertSign,cRLSign];
    ncwServer:      Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature,keyEncipherment];
    ncwClient:      Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature,keyEncipherment];
    ncwCodeSigning: Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature,nonRepudiation];
    ncwEmail:       Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature,keyEncipherment,dataEncipherment];
    ncwPeerToPeer:  Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature,keyEncipherment];
  end;

  //Extended Key Usage
  if not (Wizard in [ncwAdvanced,ncwCA]) then begin
    Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeExtKeyUsage);
    case Wizard of
      ncwServer:      Ext.ExtnValue.AsCe_ExtKeyUsage.AddValue(id_kp_serverAuth);
      ncwClient:      Ext.ExtnValue.AsCe_ExtKeyUsage.AddValue(id_kp_clientAuth);
      ncwCodeSigning: Ext.ExtnValue.AsCe_ExtKeyUsage.AddValue(id_kp_codeSigning);
      ncwEMail:       Ext.ExtnValue.AsCe_ExtKeyUsage.AddValue(id_kp_emailProtection);
      ncwPeerToPeer:
        begin
          Ext.ExtnValue.AsCe_ExtKeyUsage.AddValue(id_kp_serverAuth);
          Ext.ExtnValue.AsCe_ExtKeyUsage.AddValue(id_kp_clientAuth);
        end;
    end;
  end;

  //Basic constraints
  if Wizard = ncwCA then begin
    Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeBasicConstraints);
    Ext.ExtnValue.AsCe_BasicConstraints.CA := True;
    // Change PathLen to allow intermediary CAs in the chain:
    Ext.ExtnValue.AsCe_BasicConstraints.PathLenConstraint.AsInteger := 0;
  end else if Wizard <> ncwAdvanced then begin
    Ext := Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeBasicConstraints);
    Ext.ExtnValue.AsCe_BasicConstraints.CA := False;
  end;

  //Private key
  PrivateKeyAlg := rsaEncryption;
  if not (Wizard in [ncwAdvanced,ncwCA]) then
    ExportableRSA := True;
  KeySize := 1024;
  SignatureDigestAlg := haSHA1;

  //Select issuer
  Idx := MyCerts.Count - 1;
  while Idx >= 0 do begin
    CACert := MyCerts.Certs[Idx];
    if CheckValidity(CACert,OffsetFromUTC) and (keyCertSign in ExtractKeyUsage(CACert)) then 
      Break;
    Dec(Idx);
  end;

  inherited;
end;

function TServerCertificateCreator.GetCommonName: WideString;
begin
  Result := FCommonName;
end;

function TServerCertificateCreator.GetCountry: WideString;
begin
  Result := FCountry;
end;

function TServerCertificateCreator.GetDnQualifier: WideString;
begin
  Result := FDnQualifier;
end;

function TServerCertificateCreator.GetDNSName: WideString;
begin
  Result := FDNSName;
end;

function TServerCertificateCreator.GetGenerationQualifier: WideString;
begin
  Result := FGenerationQualifier;
end;

function TServerCertificateCreator.GetGivenName: WideString;
begin
  Result := FGivenName;
end;

function TServerCertificateCreator.GetInitials: WideString;
begin
  Result := FInitials;
end;

function TServerCertificateCreator.GetIPAddress: WideString;
begin
  Result := FIPAddress;
end;

function TServerCertificateCreator.GetLocalityName: WideString;
begin
  Result := FLocalityName;
end;

function TServerCertificateCreator.GetOrganizationalUnitName: WideString;
begin
  Result := FOrganizationalUnitName;
end;

function TServerCertificateCreator.GetOrganizationName: WideString;
begin
  Result := FOrganizationName;
end;

function TServerCertificateCreator.GetRfc822Name: WideString;
begin
  Result := FRfc822Name;
end;

function TServerCertificateCreator.GetStateOrProvinceName: WideString;
begin
  Result := FStateOrProvinceName;
end;

function TServerCertificateCreator.GetSurname: WideString;
begin
  Result := FSurname;
end;

function TServerCertificateCreator.GetUniformResourceIdentifier: WideString;
begin
  Result := FUniformResourceIdentifier;
end;

procedure TServerCertificateCreator.SetCommonName(const Value: WideString);
begin
  FCommonName := Value;
end;

procedure TServerCertificateCreator.SetCountry(const Value: WideString);
begin
  FCountry := Value;
end;

procedure TServerCertificateCreator.SetDnQualifier(const Value: WideString);
begin
  FDnQualifier := Value;
end;

procedure TServerCertificateCreator.SetDNSName(const Value: WideString);
begin
  FDNSName := Value;
end;

procedure TServerCertificateCreator.SetGenerationQualifier(const Value: WideString);
begin
  FGenerationQualifier := Value;
end;

procedure TServerCertificateCreator.SetGivenName(const Value: WideString);
begin
  FGivenName := Value;
end;

procedure TServerCertificateCreator.SetInitials(const Value: WideString);
begin
  FInitials := Value;
end;

procedure TServerCertificateCreator.SetIPAddress(const Value: WideString);
begin
  FIPAddress := Value;
end;

procedure TServerCertificateCreator.SetLocalityName(const Value: WideString);
begin
  FLocalityName := Value;
end;

procedure TServerCertificateCreator.SetOrganizationalUnitName(const Value: WideString);
begin
  FOrganizationalUnitName := Value;
end;

procedure TServerCertificateCreator.SetOrganizationName(const Value: WideString);
begin
  FOrganizationName := Value;
end;

procedure TServerCertificateCreator.SetRfc822Name(const Value: WideString);
begin
  FRfc822Name := Value;
end;

procedure TServerCertificateCreator.SetStateOrProvinceName(const Value: WideString);
begin
  FStateOrProvinceName := Value;
end;

procedure TServerCertificateCreator.SetSurname(const Value: WideString);
begin
  FSurname := Value;
end;

procedure TServerCertificateCreator.SetUniformResourceIdentifier(const Value: WideString);
begin
  FUniformResourceIdentifier := Value;
end;



end.