{$ASSERTIONS ON}
unit DigSignMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls;

type
  TfrmDigSign = class(TForm)
    memMessage: TMemo;
    Label1: TLabel;
    memSignature: TMemo;
    Label2: TLabel;
    rgPKAlg: TRadioGroup;
    rgHashAlg: TRadioGroup;
    btnCreateKey: TButton;
    btnSign: TButton;
    btnVerify: TButton;
    memStatus: TMemo;
    btnNewKey: TButton;
    procedure btnCreateKeyClick(Sender: TObject);
    procedure btnSignClick(Sender: TObject);
    procedure btnVerifyClick(Sender: TObject);
    procedure btnNewKeyClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmDigSign: TfrmDigSign;

implementation

uses
  SecUtils, SsBase64, Pkix, DigSignModule, MpPK;

{$R *.dfm}

procedure TfrmDigSign.btnCreateKeyClick(Sender: TObject);
var
  T: LongWord;
begin
  Screen.Cursor := crAppStart;
  try
    T := GetTickCount;
    case rgPKAlg.ItemIndex of
      0:
        case rgHashAlg.ItemIndex of
          0: dmDigSign.CreateIFKeyPair(1024,False,haSHA1);
          1: dmDigSign.CreateIFKeyPair(1024,False,haRipeMD160);
          2: dmDigSign.CreateIFKeyPair(2048,False,haSHA256);
          3: dmDigSign.CreateIFKeyPair(3072,False,haSHA384);
          4: dmDigSign.CreateIFKeyPair(4096,False,haSHA512);
        end;
      1:
        case rgHashAlg.ItemIndex of
          0: dmDigSign.CreateIFKeyPair(1024,True,haSHA1);
          1: dmDigSign.CreateIFKeyPair(1024,True,haRipeMD160);
          2: dmDigSign.CreateIFKeyPair(2048,True,haSHA256);
          3: dmDigSign.CreateIFKeyPair(3072,True,haSHA384);
          4: dmDigSign.CreateIFKeyPair(4096,True,haSHA512);
        end;
      2:
        case rgHashAlg.ItemIndex of
          0: dmDigSign.CreateDLKeyPair(1024,False,haSHA1);
          1: dmDigSign.CreateDLKeyPair(1024,False,haRipeMD160);
          2: dmDigSign.CreateDLKeyPair(2048,False,haSHA256);
          3: dmDigSign.CreateDLKeyPair(3072,False,haSHA384);
          4: dmDigSign.CreateDLKeyPair(4096,False,haSHA512);
        end;
      3:
        case rgHashAlg.ItemIndex of
          0: dmDigSign.CreateDLKeyPair(1024,True,haSHA1);
          1: dmDigSign.CreateDLKeyPair(1024,True,haRipeMD160);
          2: dmDigSign.CreateDLKeyPair(2048,True,haSHA256);
          3: dmDigSign.CreateDLKeyPair(3072,True,haSHA384);
          4: dmDigSign.CreateDLKeyPair(4096,True,haSHA512);
        end;
      4:
        case rgHashAlg.ItemIndex of
          0: dmDigSign.CreateECKeyPair(prime192v1,False,haSHA1);
          1: dmDigSign.CreateECKeyPair(prime192v1,False,haRipeMD160);
          2: dmDigSign.CreateECKeyPair(prime256v1,False,haSHA256);
          3: dmDigSign.CreateECKeyPair(prime384,False,haSHA384);
          4: dmDigSign.CreateECKeyPair(prime521,False,haSHA512);
        end;
      5:
        case rgHashAlg.ItemIndex of
          0: dmDigSign.CreateECKeyPair(prime192v1,True,haSHA1);
          1: dmDigSign.CreateECKeyPair(prime192v1,True,haRipeMD160);
          2: dmDigSign.CreateECKeyPair(prime256v1,True,haSHA256);
          3: dmDigSign.CreateECKeyPair(prime384,True,haSHA384);
          4: dmDigSign.CreateECKeyPair(prime521,True,haSHA512);
        end;
    end;
    T := GetTickCount - T;
    if rgPKAlg.ItemIndex in [0,1] then
      memStatus.Lines.Add(Format('IF key: %d ms',[T]))
    else if rgPKAlg.ItemIndex in [2,3] then
      memStatus.Lines.Add(Format('DL key: %d ms',[T]))
    else if rgPKAlg.ItemIndex in [4,5] then
      memStatus.Lines.Add(Format('EC key: %d ms',[T]));
    rgPKAlg.Enabled := False;
    rgHashAlg.Enabled := False;
    btnSign.Enabled := True;
    btnNewKey.Enabled := True;
  finally
    Screen.Cursor := crDefault;
  end;
end;

procedure TfrmDigSign.btnSignClick(Sender: TObject);
var
  T: LongWord;
  Sign: string;
begin
  Screen.Cursor := crAppStart;
  try
    T := GetTickCount;
    case rgHashAlg.ItemIndex of
      0: Sign := dmDigSign.SignMsg(memMessage.Text,haSHA1);
      1: Sign := dmDigSign.SignMsg(memMessage.Text,haRipeMD160);
      2: Sign := dmDigSign.SignMsg(memMessage.Text,haSHA256);
      3: Sign := dmDigSign.SignMsg(memMessage.Text,haSHA384);
      4: Sign := dmDigSign.SignMsg(memMessage.Text,haSHA512);
    end;
    T := GetTickCount - T;
    memSignature.Text := StrToMIME64(Sign);
    if rgPKAlg.ItemIndex in [0,1] then
      memStatus.Lines.Add(Format('IF Signature: %d ms',[T]))
    else if rgPKAlg.ItemIndex in [2,3] then
      memStatus.Lines.Add(Format('DL Signature: %d ms',[T]))
    else if rgPKAlg.ItemIndex in [4,5] then
      memStatus.Lines.Add(Format('EC Signature: %d ms',[T]));
    btnVerify.Enabled := True;
  finally
    Screen.Cursor := crDefault;
  end;
end;

procedure TfrmDigSign.btnVerifyClick(Sender: TObject);
var
  T: LongWord;
  Sign: string;
  Res: Boolean;
begin
  Screen.Cursor := crAppStart;
  try
    Sign := MIME64ToStr(memSignature.Text);
    T := GetTickCount;
    case rgHashAlg.ItemIndex of
      0: Res := dmDigSign.VerifySign(memMessage.Text,Sign,haSHA1);
      1: Res := dmDigSign.VerifySign(memMessage.Text,Sign,haRipeMD160);
      2: Res := dmDigSign.VerifySign(memMessage.Text,Sign,haSHA256);
      3: Res := dmDigSign.VerifySign(memMessage.Text,Sign,haSHA384);
      4: Res := dmDigSign.VerifySign(memMessage.Text,Sign,haSHA512);
    else
      Res := False;
    end;
    T := GetTickCount - T;
    if rgPKAlg.ItemIndex in [0,1] then
      memStatus.Lines.Add(Format('IF Signature Verification: %d ms',[T]))
    else if rgPKAlg.ItemIndex in [2,3] then
      memStatus.Lines.Add(Format('DL Signature Verification: %d ms',[T]))
    else if rgPKAlg.ItemIndex in [4,5] then
      memStatus.Lines.Add(Format('EC Signature Verification: %d ms',[T]));
    if Res then
      memStatus.Lines.Add('  Signature verified OK')
    else
      memStatus.Lines.Add('  Failure verifying the signature');
  finally
    Screen.Cursor := crDefault;
  end;
end;

procedure TfrmDigSign.btnNewKeyClick(Sender: TObject);
begin
  rgPKAlg.Enabled := True;
  rgHashAlg.Enabled := True;
  btnCreateKey.Enabled := True;
  btnSign.Enabled := False;
  btnVerify.Enabled := False;
  btnNewKey.Enabled := False;
end;

end.
