unit MultiPartSignedMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, MpCms, MpX509, TlsInternalServer, IdMessage,
  IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient,
  IdMessageClient, IdSMTP;

type
  TIdInnerMessageClient = class(TIdMessageClient)
  public
    procedure SendMsgAsInner(AMsg: TIdMessage);
  end;

  TfrmMsgEditor = class(TForm)
    edtSubject: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    edtTo: TEdit;
    Label3: TLabel;
    edtSMTP: TEdit;
    Label4: TLabel;
    memMessage: TMemo;
    lbAttachments: TListBox;
    btnAddAtt: TButton;
    Label5: TLabel;
    btnSend: TButton;
    btnLoad: TButton;
    KeysAndCerts: TSimpleTLSInternalServer;
    SignedData: TMPSignedData;
    OpenDialog1: TOpenDialog;
    SMTP: TIdSMTP;
    MsgSend: TIdMessage;
    MsgSendSigned: TIdMessage;
    Label6: TLabel;
    edtFrom: TEdit;
    btnClear: TButton;
    btnShow: TButton;
    OpenDialog2: TOpenDialog;
    procedure btnLoadClick(Sender: TObject);
    procedure btnSendClick(Sender: TObject);
    procedure btnClearClick(Sender: TObject);
    procedure btnShowClick(Sender: TObject);
    procedure btnAddAttClick(Sender: TObject);
  private
    procedure SaveInnerMIMEContents(AStream: TStream);
  public
    procedure ComposeMessage;
  end;

var
  frmMsgEditor: TfrmMsgEditor;

implementation

uses
  Pkix, Asn1, X509Base, SsBase64, MultipartSignedMsgSource,
  IdIOHandlerStream, IdMessageCoder, IdGlobal;

{$R *.dfm}

procedure TfrmMsgEditor.btnLoadClick(Sender: TObject);
begin
  if OpenDialog1.Execute then begin
    if OpenDialog1.FilterIndex = 1 then
      KeysAndCerts.ImportFromPFX(OpenDialog1.FileName,nil)
    else if OpenDialog1.FilterIndex = 2 then
      KeysAndCerts.LoadPrivateKeyRingFromFile(OpenDialog1.FileName,nil)
    else
      KeysAndCerts.LoadMyCertsFromFile(OpenDialog1.FileName);
  end;
end;

procedure TfrmMsgEditor.btnSendClick(Sender: TObject);
begin
  ComposeMessage;
  SMTP.Host := edtSMTP.Text;
  SMTP.Connect;
  try
    SMTP.Send(MsgSendSigned);
  finally
    SMTP.Disconnect;
  end;
end;

procedure TfrmMsgEditor.btnClearClick(Sender: TObject);
begin
  MsgSend.Clear;
  memMessage.Clear;
  lbAttachments.Clear;
end;

procedure TfrmMsgEditor.ComposeMessage;
var
  Sign, Data: TStringStream;
  Idx: Integer;
  Crt: PASN1Struct;
  Boundary: string;
begin   
  Idx := 0;
  repeat
    if not SignedData.MyCertificates.FindCert([],
                                              [],
                                              [digitalSignature],
                                              [],
                                              Idx,
                                              Crt) then
      raise Exception.Create('Unable to locate digital signature certificate');
  until CompareText(X509Base.ExtractEmailAddress(Crt^),edtFrom.Text) = 0;
  MsgSendSigned.Clear;
  MsgSendSigned.Subject := edtSubject.Text;
  MsgSendSigned.From.Text := edtFrom.Text;
  MsgSendSigned.Recipients.Add.Text := edtTo.Text;
  Data := TStringStream.Create('');
  try
    SaveInnerMIMEContents(Data);
    Sign := TStringStream.Create('');
    try
      SignedData.SignData(Crt^,Data,Sign);
      Boundary := '----=_NextPart_' + IntToStr(Random($7FFFFFFF));
      MsgSendSigned.Headers.Values['From'] := edtFrom.Text;
      MsgSendSigned.Headers.Values['Subject'] := edtSubject.Text;
      MsgSendSigned.Headers.Values['To'] := edtTo.Text;
      MsgSendSigned.AddHeader('MIME-Version: 1.0');
      MsgSendSigned.AddHeader('Content-Type: multipart/signed;');
      MsgSendSigned.AddHeader(#9'boundary="' + Boundary + '";');
      MsgSendSigned.AddHeader(#9'protocol="application/x-pkcs7-signature";');
      MsgSendSigned.AddHeader(#9'micalg=SHA1');
      MsgSendSigned.Headers.Values['Date'] := DateTimeToInternetStr(Now);
      MsgSendSigned.Headers.Values['X-Library'] := gsIdProductName + ' ' + gsIdVersion ; {do not localize}
      MsgSendSigned.GenerateHeader;
      MsgSendSigned.Body.Add('This is a multi-part message in MIME format.');
      MsgSendSigned.Body.Add('');
      MsgSendSigned.Body.Add('--' + Boundary);
      MsgSendSigned.Body.Add(Data.DataString);
      MsgSendSigned.Body.Add('--' + Boundary);
      MsgSendSigned.Body.Add('Content-Type: application/x-pkcs7-signature;');
      MsgSendSigned.Body.Add(#9'name="smime.p7s"');
      MsgSendSigned.Body.Add('Content-Transfer-Encoding: base64');
      MsgSendSigned.Body.Add('Content-Disposition: attachment;');
      MsgSendSigned.Body.Add(#9'filename="smime.p7s"');
      MsgSendSigned.Body.Add(StrToMime64('','',Sign.DataString));
      MsgSendSigned.Body.Add('--' + Boundary + '--');
      MsgSendSigned.NoEncode := True;
    finally
      Sign.Free;
    end;
  finally
    Data.Free;
  end;
end;

procedure TfrmMsgEditor.btnShowClick(Sender: TObject);
var
  Msg: TStringStream;
begin
  ComposeMessage;
  Msg := TStringStream.Create('');
  try
    MsgSendSigned.SaveToStream(Msg);
    frmMsgSource.memMsgSource.Text := Msg.DataString;
    frmMsgSource.Show;
  finally
    Msg.Free;
  end;
end;

procedure TfrmMsgEditor.SaveInnerMIMEContents(AStream: TStream);
var
  vMsgClient: TIdInnerMessageClient;
  vIOHandler: TIdIOHandlerStream;
  vVoidStream: TMemoryStream;
begin           
  MsgSend.Body.Assign(memMessage.Lines);

  vVoidStream := TMemoryStream.Create;
  vIOHandler := TIdIOHandlerStream.Create(nil);
  vMsgClient := TIdInnerMessageClient.Create(nil);

  vIOHandler.StreamType := stWrite;
  vIOHandler.WriteStream := AStream;
  vIOHandler.ReadStream := vVoidStream;
  vMsgClient.IOHandler := vIOHandler;

  try
    vMsgClient.OpenWriteBuffer(32768);
    vMsgClient.SendMsgAsInner(MsgSend);
    vMsgClient.CloseWriteBuffer;
  finally
    vIOHandler.WriteStream := nil;
    vIOHandler.ReadStream := nil;
    vVoidStream.Free;

    FreeAndNil(vIOHandler);
    FreeAndNil(vMsgClient);
  end;
end;

{ TIdInnerMessageClient }

procedure TIdInnerMessageClient.SendMsgAsInner(AMsg: TIdMessage);
begin
  if AMsg.MessageParts.Count = 0 then begin
    if AMsg.CharSet = '' then
      WriteLn('Content-Type: text/plain')
    else begin
      WriteLn('Content-Type: text/plain;');
      WriteLn(#9'charset="' + AMsg.CharSet + '"');
    end;
    WriteLn('Content-Transfer-Encoding: 7bit');
  end else begin          
    AMsg.MessageParts.CountParts;
    TIdMessageEncoderInfo(AMsg.MessageParts.MessageEncoderInfo).InitializeHeaders(AMsg);
    WriteLn('Content-Type: ' + AMsg.ContentType);
  end;
  WriteLn('');
  SendBody(AMsg);
end;

procedure TfrmMsgEditor.btnAddAttClick(Sender: TObject);
begin
  if OpenDialog2.Execute then begin
    TIdAttachment.Create(MsgSend.MessageParts,OpenDialog2.FileName);
    lbAttachments.Items.Add(ExtractFileName(OpenDialog2.FileName));
  end;
end;

initialization
  Randomize;
end.
