unit RecipientMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls,
  SecUtils, Asn1, MpIF, SecComp;

type
  TForm2 = class(TForm)
    Label2: TLabel;
    memEncMsg: TMemo;
    Label1: TLabel;
    memPlain: TMemo;
    btnOpen: TButton;
    btnDecode: TButton;
    Label3: TLabel;
    memStatus: TMemo;
    OpenDialog1: TOpenDialog;
    SsAES1: TSsAES;
    SsHash1: TSsHash;
    btnNewKey: TButton;
    procedure btnOpenClick(Sender: TObject);
    procedure btnDecodeClick(Sender: TObject);
    procedure btnNewKeyClick(Sender: TObject);
  private
    function DecryptKey(Msg: TASN1Struct): ISecretKey;
    function DecryptMessage: string;
    procedure GenerateKey;
    procedure InitASN1(F: TASN1Struct);
    procedure LoadRecipientKey(out AKey: TIFPrivateKey);
    procedure LoadSenderKey(out AKey: TIFPublicKey);
    function VerifySignedMessage(Msg: string): string;
  public
    { Public declarations }
  end;

var
  Form2: TForm2;

implementation

uses
  PasswordMain, X509Base;

{$R *.dfm}

procedure TForm2.btnDecodeClick(Sender: TObject);
begin
  memStatus.Clear;
  memPlain.Text := DecryptMessage;
end;

procedure TForm2.btnNewKeyClick(Sender: TObject);
begin
  GenerateKey;
end;

procedure TForm2.btnOpenClick(Sender: TObject);
var
  F: TASN1Struct;
  SS: TStringStream;
begin
  if OpenDialog1.Execute then begin
    if OpenDialog1.FilterIndex = 1 then begin
      F := TASN1Struct.Create;
      try
        InitASN1(F);
        F.LoadFromFile(OpenDialog1.FileName,fmtDER);
        SS := TStringStream.Create('');
        try
          F.SaveToStream(SS,fmtDOM);
          memEncMsg.Text := SS.DataString;
        finally
          SS.Free;
        end;
      finally
        F.Free;
      end;
    end else begin
      memEncMsg.Lines.LoadFromFile(OpenDialog1.FileName);
    end;
  end;
end;

function TForm2.DecryptKey(Msg: TASN1Struct): ISecretKey;
var
  P: PASN1Struct;
  I: Integer;
  Priv: TIFPrivateKey;
begin
  Assert(Msg.Constructed and (Msg.ItemCount > 0),'Invalid format');
  I := 0;
  repeat
    P := Msg.Items[I];
    if (P.Cls <> V_ASN1_CONTEXT_SPECIFIC) or (P.Tag <> 0) or P.Constructed then
      P := nil;
    Inc(I);
  until (I = Msg.ItemCount) or Assigned(P);
  if Assigned(P) and (P.Length > 0) then begin
    LoadRecipientKey(Priv);
    try
      if not MpIF.IFESDecryption(Priv,P.Content^,P.Length,haSHA1,eeEME1,Result) then
        raise Exception.Create('Could not decrypt key');
    finally
      DisposeIFPrivateKey(Priv);
    end;
  end else
    PasswordMain.PasswordDlg('Encrypted Message',Result);
end;

function TForm2.DecryptMessage: string;
var
  SS: TStringStream;
  F: TASN1Struct;
  PW: ISecretKey;
  Param, Msg, EMsg, Mac: PASN1Struct;

  function FindField(F: TASN1Struct; Tag: Cardinal): PASN1Struct;
  var
    I: Integer;
  begin
    Assert(F.Constructed and (F.ItemCount > 0),'Invalid format');
    I := 0;
    repeat
      Result := F.Items[I];
      if (Result.Cls <> V_ASN1_CONTEXT_SPECIFIC) or (Result.Tag <> Tag) or
         Result.Constructed then
        Result := nil;
      Inc(I);
    until (I = F.ItemCount) or Assigned(Result);
    if Assigned(Result) and (Result.Length = 0) then
      Result := nil;
  end;

begin
  F := TASN1Struct.Create;
  try
    SS := TStringStream.Create(memEncMsg.Text);
    try
      F.LoadFromStream(SS,fmtDOM);
    finally
      SS.Free;
    end;
    Param := FindField(F,1);
    EMsg := FindField(F,2);
    Msg := FindField(F,3);
    Mac := FindField(F,4);
    if Assigned(Msg) then begin
      memStatus.Lines.Add('Warning: Message was not encrypted');
      if Assigned(Mac) then begin
        PW := DecryptKey(F);
        SsHash1.SetKeyIntf(PW);
        SsHash1.HashStr(Msg.ContentAsOctetString);
        SsHash1.Done(nil);
        if SsHash1.Digest = Mac.ContentAsOctetString then
          memStatus.Lines.Add('Mac verified OK')
        else
          raise Exception.Create('Mac verification failure');
      end;
      Result := VerifySignedMessage(Msg.ContentAsOctetString);
    end else if Param = nil then
      raise Exception.Create('Illegal format')
    else begin
      PW := DecryptKey(F);
      SsAES1.Param := Param.ContentAsOctetString;
      if FindField(F,0) = nil then
        SsAES1.PBKDF.Iter := $1000
      else
        SsAES1.PBKDF.Iter := 1;
      SsAES1.PBKDF.HashClass := TSHA1;
      SS := TStringStream.Create('');
      try
        SsAES1.DestinationStream := SS;
        SsAES1.SetKeyPassword(PW,32);
        SsAES1.DecryptStrToDst(EMsg.ContentAsOctetString);
        if SsAES1.DoneToDst then
          memStatus.Lines.Add('Message decrypted OK')
        else
          raise Exception.Create('Illegal padding');
        if Assigned(Mac) then begin
          if SsHash1.Digest = Mac.ContentAsOctetString then
            memStatus.Lines.Add('Mac verified OK')
          else
            raise Exception.Create('Mac verification failure');
        end;
        Result := VerifySignedMessage(SS.DataString);
      finally
        SS.Free;
      end;
    end;
  finally
    F.Free;
  end;
end;

procedure TForm2.GenerateKey;
var
  FileName: string;
  FS: TFileStream;
  PW: ISecretKey;
  Priv: TIFPrivateKey;
  Publ: TIFPublicKey;
  F: TASN1Struct;
begin
  FillChar(Priv,SizeOf(Priv),0);
  try
    FileName := ExtractFilePath(Application.ExeName) + 'Recipient';
    if not FileExists(FileName + '.prv') then begin
      FillChar(Publ,SizeOf(Publ),0);
      try
        IFRSAKeys(1024,1,False,Priv,Publ);
        F := TASN1Struct.Create;
        F.TypeName := 'RSAPublicKey';
        F.Persistent := True;
        F.Tag := V_ASN1_SEQUENCE;
        F.Constructed := True;
        F.AddField('modulus','INTEGER',nil).EditContent(Publ.N,True);
        F.AddField('publicExponent','INTEGER',nil).EditContent(Publ.E,True);
        F.SaveToFile(FileName + '.pub',fmtDOM);
        FS := TFileStream.Create(FileName + '.prv',fmCreate);
        try
          PasswordMain.NewPasswordDlg('Recipient Private Key',PW);
          X509Base.RSAPrivateKeyToStream(FS,PW.Key^,PW.KeyLen,Priv);
        finally
          FS.Free;
        end;
      finally
        DisposeIFPublicKey(Publ);
      end;
    end;
  finally
    DisposeIFPrivateKey(Priv);
  end;
end;

procedure TForm2.InitASN1(F: TASN1Struct);
var
  P: PASN1Struct;
begin
  F.TypeName := 'EncodedMessage';
  F.Persistent := True;
  F.Tag := V_ASN1_SEQUENCE;
  F.Constructed := True;
  F.AddField('encryptedKey','[0] IMPLICIT OCTET STRING',nil).Optional := True;
  F.AddField('params','[1] IMPLICIT OCTET STRING',nil).Optional := True;
  P := F.AddField('contentData','ContentChoices',nil);
  P.Persistent := True;
  P.AddChoice('encryptedContent','[2] IMPLICIT OCTET STRING');
  P.AddChoice('content','[3] IMPLICIT OCTET STRING');
  F.AddField('mac','[4] IMPLICIT OCTET STRING',nil).Optional := True;
end;

procedure TForm2.LoadRecipientKey(out AKey: TIFPrivateKey);
var
  FileName: string;
  FS: TFileStream;
  PW: ISecretKey;
begin
  FillChar(AKey,SizeOf(AKey),0);
  try
    FileName := ExtractFilePath(Application.ExeName) + 'Recipient';
    if FileExists(FileName + '.prv') then begin
      FS := TFileStream.Create(FileName + '.prv',fmOpenRead);
      try
        PasswordMain.PasswordDlg('Recipient Private Key',PW);
        if not X509Base.RSAPrivateKeyFromStream(FS,PW.Key^,PW.KeyLen,AKey) then begin
          DisposeIFPrivateKey(AKey);
          raise Exception.Create(FileName + '.prv could not be opened');
        end;
      finally
        FS.Free;
      end;
    end else
      raise Exception.Create('Unable to load private key');
  except
    DisposeIFPrivateKey(AKey);
    raise;
  end;
end;

procedure TForm2.LoadSenderKey(out AKey: TIFPublicKey);
var
  F: TASN1Struct;
  N, E: PASN1Struct;
begin
  FillChar(AKey,SizeOf(AKey),0);
  F := TASN1Struct.Create;
  try
    if F.LoadFromFile(ExtractFilePath(Application.ExeName) + 'Sender.pub',fmtDOM) = 0 then
      raise Exception.Create('Sender.pub not found. Please run the Sender demo and create a key pair');
    N := F.FindField('modulus');
    E := F.FindField('publicExponent');
    if Assigned(N) and Assigned(E) then begin
      AKey.Scheme := ifRSA1;
      try
        N.ContentAsUMPInt(AKey.N);
        E.ContentAsUMPInt(AKey.E);
      except
        DisposeIFPublicKey(AKey);
      end;
    end else
      raise Exception.Create('Sender.pub did not contain a correctly encoded RSA public key');
  finally
    F.Free;
  end;
end;

function TForm2.VerifySignedMessage(Msg: string): string;
var
  F: TASN1Struct;
  SS: TStringStream;
  PT: string;
  Publ: TIFPublicKey;
begin
  F := TASN1Struct.Create;
  try
    SS := TStringStream.Create(Msg);
    try
      F.LoadFromStream(SS,fmtDER);
      if F.Constructed and (F.ItemCount > 0) then begin
        PT := F.Items[0].ContentAsOctetString;
        if F.ItemCount > 1 then begin
          LoadSenderKey(Publ);
          try
            if IFSSASignatureVerification(Publ,
                                          Pointer(PT)^,Length(PT),
                                          F.Items[1].Content^,F.Items[1].Length,
                                          haSHA1,haSHA1,
                                          seEMSA4) then
              memStatus.Lines.Add('Signature verified OK')
            else
              raise Exception.Create('Signature verification failure');
          finally
            DisposeIFPublicKey(Publ);
          end;
        end;
      end else
        raise Exception.Create('Illegal format');
    finally
      SS.Free;
    end;
  finally
    F.Free;
  end;
  Result := PT;
end;

end.
