unit SenderMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls,
  SecUtils, Asn1, MpIF, SecComp;

type
  TForm1 = class(TForm)
    Label1: TLabel;
    memPlain: TMemo;
    Label2: TLabel;
    memEncMsg: TMemo;
    rgKeyMode: TRadioGroup;
    chbHMac: TCheckBox;
    chbSign: TCheckBox;
    btnEncode: TButton;
    chbEncrypt: TCheckBox;
    SsAES1: TSsAES;
    SsHash1: TSsHash;
    btnSave: TButton;
    SaveDialog1: TSaveDialog;
    procedure btnEncodeClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
  private
    procedure LoadRecipientKey(out AKey: TIFPublicKey);
    procedure LoadSenderKey(out AKey: TIFPrivateKey);
    function SignMessage: string;
    function EncryptMessage: string;
    function EncryptKey(Msg: TASN1Struct): ISecretKey;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses
  PasswordMain, MpYarrow, X509Base;

{$R *.dfm}

{ TForm1 }

function TForm1.EncryptKey(Msg: TASN1Struct): ISecretKey;
var
  EK: string;
  Publ: TIFPublicKey;
  EKLen: Integer;
  P: PASN1Struct;
begin
  LoadRecipientKey(Publ);
  try
    MpYarrow.RawRandomIntf(Result,128);
    EKLen := 128;
    SetLength(EK,EKLen);
    IFESEncryption(Publ,Result.Key^,Result.KeyLen,haSHA1,eeEME1,Pointer(EK)^,EKLen);
    SetLength(EK,EKLen);
  finally
    DisposeIFPublicKey(Publ);
  end;
  P := Msg.AddField('encryptedKey','[0] IMPLICIT OCTET STRING',nil);
  P.SetContent(Pointer(EK)^,EKLen);
end;

function TForm1.EncryptMessage: string;
var
  Msg, Param, Mac: string;
  F: TASN1Struct;
  PW: ISecretKey;
  MS: TMemoryStream;
  SS: TStringStream;
begin
  Msg := SignMessage;
  F := TASN1Struct.Create;
  try
    F.TypeName := 'EncodedMessage';
    F.Persistent := True;
    F.Tag := V_ASN1_SEQUENCE;
    F.Constructed := True;
    if chbEncrypt.Checked or chbHMac.Checked then begin
      SsAES1.Param := '';
      SsAES1.PBKDF.HashClass := TSHA1;
      if rgKeyMode.ItemIndex = 0 then begin
        PasswordMain.NewPasswordDlg('Encrypted Message',PW);
        SsAES1.PBKDF.Iter := $1000;
      end else begin
        PW := EncryptKey(F);
        SsAES1.PBKDF.Iter := 1;
      end;
      if chbEncrypt.Checked then begin
        MS := TMemoryStream.Create;
        try
          SsAES1.SetKeyPassword(PW,32);
          Param := SsAES1.Param;
          F.AddField('params','[1] IMPLICIT OCTET STRING',nil).SetContent(Pointer(Param)^,Length(Param));
          SsAES1.DestinationStream := MS;
          SsAES1.EncryptStrToDst(Msg);
          SsAES1.DoneToDst;
          F.AddField('encryptedContent','[2] IMPLICIT OCTET STRING',nil).SetContent(MS.Memory^,MS.Size);
        finally
          MS.Free;
        end;
      end else begin
        SsHash1.SetKeyIntf(PW);
        SsHash1.HashStr(Msg);
        F.AddField('content','[3] IMPLICIT OCTET STRING',nil).SetContent(Pointer(Msg)^,Length(Msg));
        SsHash1.Done(nil);
      end;
      if chbHMac.Checked then begin
        Mac := SsHash1.Digest;
        F.AddField('mac','[4] IMPLICIT OCTET STRING',nil).SetContent(Pointer(Mac)^,Length(Mac));
      end;
    end else begin
      F.AddField('content','[3] IMPLICIT OCTET STRING',nil).SetContent(Pointer(Msg)^,Length(Msg));
    end;
    SS := TStringStream.Create('');
    try
      F.SaveToStream(SS,fmtDOM);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  finally
    F.Free;
  end;
end;

procedure TForm1.LoadRecipientKey(out AKey: TIFPublicKey);
var
  F: TASN1Struct;
  N, E: PASN1Struct;
begin
  FillChar(AKey,SizeOf(AKey),0);
  F := TASN1Struct.Create;
  try
    if F.LoadFromFile(ExtractFilePath(Application.ExeName) + 'Recipient.pub',fmtDOM) = 0 then
      raise Exception.Create('Recipient.pub not found. Please run the Recipient demo and create a key pair');
    N := F.FindField('modulus');
    E := F.FindField('publicExponent');
    if Assigned(N) and Assigned(E) then begin
      AKey.Scheme := ifRSA1;
      try
        N.ContentAsUMPInt(AKey.N);
        E.ContentAsUMPInt(AKey.E);
      except
        DisposeIFPublicKey(AKey);
      end;
    end else
      raise Exception.Create('Recipient.pub did not contain a correctly encoded RSA public key');
  finally
    F.Free;
  end;
end;

procedure TForm1.LoadSenderKey(out AKey: TIFPrivateKey);
var
  FileName: string;
  FS: TFileStream;
  PW: ISecretKey;
  Publ: TIFPublicKey;
  F: TASN1Struct;
begin
  FillChar(AKey,SizeOf(AKey),0);
  try
    FileName := ExtractFilePath(Application.ExeName) + 'Sender';
    if FileExists(FileName + '.prv') then begin
      FS := TFileStream.Create(FileName + '.prv',fmOpenRead);
      try
        PasswordMain.PasswordDlg('Sender private key',PW);
        if not X509Base.RSAPrivateKeyFromStream(FS,PW.Key^,PW.KeyLen,AKey) then begin
          DisposeIFPrivateKey(AKey);
          raise Exception.Create(FileName + '.prv could not be opened');
        end;
      finally
        FS.Free;
      end;
    end else begin
      FillChar(Publ,SizeOf(Publ),0);
      try
        IFRSAKeys(1024,1,False,AKey,Publ);
        F := TASN1Struct.Create;
        F.TypeName := 'RSAPublicKey';
        F.Persistent := True;
        F.Tag := V_ASN1_SEQUENCE;
        F.Constructed := True;
        F.AddField('modulus','INTEGER',nil).EditContent(Publ.N,True);
        F.AddField('publicExponent','INTEGER',nil).EditContent(Publ.E,True);
        F.SaveToFile(FileName + '.pub',fmtDOM);
        FS := TFileStream.Create(FileName + '.prv',fmCreate);
        try
          PasswordMain.NewPasswordDlg('Sender private key',PW);
          X509Base.RSAPrivateKeyToStream(FS,PW.Key^,PW.KeyLen,AKey);
        finally
          FS.Free;
        end;
      finally
        DisposeIFPublicKey(Publ);
      end;
    end;
  except
    DisposeIFPrivateKey(AKey);
    raise;
  end;
end;

function TForm1.SignMessage: string;
var
  F: TASN1Struct;
  M, S: string;
  Priv: TIFPrivateKey;
begin
  F := TASN1Struct.Create;
  try
    F.TypeName := 'SignedMessage';
    F.Persistent := True;
    F.Tag := V_ASN1_SEQUENCE;
    F.Constructed := True;
    M := memPlain.Text;
    F.AddField('message','OCTET STRING',nil).SetContent(Pointer(M)^,Length(M));
    if chbSign.Checked then begin
      LoadSenderKey(Priv);
      try
        SetLength(S,256);
        SetLength(S,IFSSASignatureGeneration(Priv,
                                             Pointer(M)^,Length(M),
                                             haSHA1,haSHA1,seEMSA4,
                                             Pointer(S)^,Length(S)));
      finally
        DisposeIFPrivateKey(Priv);
      end;
      F.AddField('signature','OCTET STRING',nil).SetContent(Pointer(S)^,Length(S));
    end;
    F.CalculateLength;
    Result := F.ContentAsOctetString;
  finally
    F.Free;
  end;
end;

procedure TForm1.btnEncodeClick(Sender: TObject);
begin
  memEncMsg.Text := EncryptMessage;
end;

procedure TForm1.btnSaveClick(Sender: TObject);
var
  F: TASN1Struct;
  SS: TStringStream;
  FileName: string;
begin
  F := TASN1Struct.Create;
  try
    SS := TStringStream.Create(memEncMsg.Text);
    try
      F.LoadFromStream(SS,fmtDOM);
    finally
      SS.Free;
    end;
    if SaveDialog1.Execute then begin
      FileName := SaveDialog1.FileName;
      if SaveDialog1.FilterIndex = 1 then begin
        if ExtractFileExt(FileName) = '' then
          FileName := FileName + '.der';
        F.SaveToFile(FileName,fmtDER)
      end else begin
        if ExtractFileExt(FileName) = '' then
          FileName := FileName + '.dom';
        F.SaveToFile(FileName,fmtDOM);
      end;
    end;
  finally
    F.Free;
  end;
end;

end.
