unit HttpDemoMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  HTTPSend, StdCtrls, ExtCtrls, Asn1, MpX509, TlsClass, TlsInternalServer,
  ResourceFile;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    Panel4: TPanel;
    Memo1: TMemo;
    Memo2: TMemo;
    Label1: TLabel;
    Edit1: TEdit;
    Button1: TButton;
    Label2: TLabel;
    Label3: TLabel;
    Edit2: TEdit;
    Button2: TButton;
    Label4: TLabel;
    Label5: TLabel;
    Edit3: TEdit;
    Label6: TLabel;
    Edit4: TEdit;
    Label7: TLabel;
    Label8: TLabel;
    Edit5: TEdit;
    Button3: TButton;
    Label9: TLabel;
    Edit6: TEdit;
    Edit7: TEdit;
    Label10: TLabel;
    TLSInternalServer1: TTLSInternalServer;
    ResourceFile1: TResourceFile;
    MicrosoftCert: TResourceFile;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SimpleTLSInternalServer1CertNotAccepted(Sender: TObject;
      Cert: TASN1Struct; Status: TCertStatusCode);
    procedure TLSInternalServer1TLSIncomingAlert(Sender: TObject;
      Client: TCustomTLS_ContentLayer; var Fatal: Boolean;
      AlertCode: Integer);
    procedure TLSInternalServer1TLSOutgoingAlert(Sender: TObject;
      Client: TCustomTLS_ContentLayer; var Fatal: Boolean;
      AlertCode: Integer);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses
  TlsSynaSock;

{$R *.DFM}

function ProxyHttpPostURL(const URL, URLData: string; const Data: TStream): Boolean;
var
  HTTP: THTTPSend;
begin
  HTTP := THTTPSend.Create;
  try
//    HTTP.ProxyHost := Form1.Edit8.Text;
//    HTTP.ProxyPort := Form1.Edit9.Text;
    HTTP.Document.Write(Pointer(URLData)^, Length(URLData));
    HTTP.MimeType := 'application/x-www-form-urlencoded';
    Result := HTTP.HTTPMethod('POST', URL);
    Data.CopyFrom(HTTP.Document, 0);
  finally
    HTTP.Free;
  end;
end;

function ProxyHttpPostFile(const URL, FieldName, FileName: string;
  const Data: TStream; const ResultData: TStrings): Boolean;
const
  CRLF = #$0D + #$0A;
var
  HTTP: THTTPSend;
  Bound, s: string;
begin
  Bound := IntToHex(Random(MaxInt), 8) + '_Synapse_boundary';
  HTTP := THTTPSend.Create;
  try
//    HTTP.ProxyHost := Form1.Edit8.Text;
//    HTTP.ProxyPort := Form1.Edit9.Text;
    s := '--' + Bound + CRLF;
    s := s + 'content-disposition: form-data; name="' + FieldName + '";';
    s := s + ' filename="' + FileName +'"' + CRLF;
    s := s + 'Content-Type: Application/octet-string' + CRLF + CRLF;
    HTTP.Document.Write(Pointer(s)^, Length(s));
    HTTP.Document.CopyFrom(Data, 0);
    s := CRLF + '--' + Bound + '--' + CRLF;
    HTTP.Document.Write(Pointer(s)^, Length(s));
    HTTP.MimeType := 'multipart/form-data, boundary=' + Bound;
    Result := HTTP.HTTPMethod('POST', URL);
    ResultData.LoadFromStream(HTTP.Document);
  finally
    HTTP.Free;
  end;
end;


procedure TForm1.Button1Click(Sender: TObject);
var
  HTTP: THTTPSend;
begin
  HTTP := THTTPSend.Create;
  try
// Note: If you get a compiler error here you have to add STREAMSEC to the
// Conditional Defines in the Project Options dialog box. Select Build before
// you Run the project to force the compiler to recompile all Synapse units.
    HTTP.TLSServer := TLSInternalServer1;
//    HTTP.ProxyHost := Edit8.Text;
//    HTTP.ProxyPort := Edit9.Text;
    HTTP.HTTPMethod('GET', Edit1.text);
    Memo1.Lines.Assign(HTTP.Headers);
    HTTP.Document.Position := 0;
    Memo2.Lines.LoadFromStream(HTTP.Document);
  finally
    HTTP.Free;
  end;
end;

procedure TForm1.Button2Click(Sender: TObject);
var
  st: TMemoryStream;
begin
  st:=TMemoryStream.Create;
  try
    GlobalTLSInternalServer := TLSInternalServer1;
    ProxyHTTPpostURL(Edit2.Text, Edit3.Text + '=' + Edit4.Text, st);
    st.Seek(0,soFromBeginning);
    Memo2.Lines.LoadFromStream(st);
  finally
    st.Free;
  end;
end;

procedure TForm1.Button3Click(Sender: TObject);
var
  st: TFileStream;
begin
  st := TFileStream.Create(Edit7.Text, fmOpenRead	or fmShareDenyWrite);
  try
    GlobalTLSInternalServer := TLSInternalServer1;
    ProxyHTTPPostFile(Edit5.Text, Edit6.Text, ExtractFilename(Edit7.Text), st, TStringList(memo2.Lines));
  finally
    st.Free;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  TlsSynaSock.GlobalTLSInternalServer := TLSInternalServer1;
end;

procedure TForm1.SimpleTLSInternalServer1CertNotAccepted(Sender: TObject;
  Cert: TASN1Struct; Status: TCertStatusCode);
begin
  MessageDlg('A certificate was not accepted',mtWarning,[mbOK],0);
end;

procedure TForm1.TLSInternalServer1TLSIncomingAlert(Sender: TObject;
  Client: TCustomTLS_ContentLayer; var Fatal: Boolean; AlertCode: Integer);
var
  Msg: string;
begin
  if AlertCode = 0 then
    Msg := 'Incoming close notification'#13#10
  else if Fatal then
    Msg := 'Incoming fatal alert ' + IntToStr(AlertCode) + #13#10
  else
    Msg := 'Incoming warning ' + IntToStr(AlertCode) + #13#10;
  case AlertCode and $FF of
    0:  Msg := Msg + 'close_notify: The remote host has terminated the connection.';
    10: Msg := Msg + 'unexpected_message:'#13#10 +
       'An inappropriate message was received. This alert is always fatal'#13#10 +
       'and should never be observed in communication between proper'#13#10 +
       'implementations. Please notify StreamSec if you get this message.';
    20: Msg := Msg + 'bad_record_mac:'#13#10 +
       'This alert is returned if a record is received with an incorrect'#13#10 +
       'MAC. This alert also SHOULD be returned if a TLSCiphertext'#13#10 +
       'decrypted in an invalid way: either it wasn''t an even multiple of'#13#10 +
       'the block length, or its padding values, when checked, weren''t'#13#10 +
       'correct. This message is always fatal.';
    21: Msg := Msg + 'decryption_failed:'#13#10 +
       'This alert MAY be returned if a TLSCiphertext decrypted in an'#13#10 +
       'invalid way: either it wasn''t an even multiple of the block'#13#10 +
       'length, or its padding values, when checked, weren''t correct.'#13#10 +
       'This message is always fatal.'#13#10 +
       #13#10 +
       'NB: Differentiating between bad_record_mac and decryption_failed'#13#10 +
       'alerts may permit certain attacks against CBC mode as used in'#13#10 +
       'TLS. It is preferable to uniformly use the bad_record_mac alert'#13#10 +
       'to hide the specific type of the error.';
    22: Msg := Msg + 'record_overflow:'#13#10 +
       'A TLSCiphertext record was received which had a length more than'#13#10 +
       '2^14+2048 bytes, or a record decrypted to a TLSCompressed record'#13#10 +
       'with more than 2^14+1024 bytes. This message is always fatal.';
    30: Msg := Msg + 'decompression_failure:'#13#10 +
       'The decompression function received improper input (e.g. data'#13#10 +
       'that would expand to excessive length). This message is always'#13#10 +
       'fatal.';
    40: Msg := Msg + 'handshake_failure:'#13#10 +
       'Reception of a handshake_failure alert message indicates that the'#13#10 +
       'sender was unable to negotiate an acceptable set of security'#13#10 +
       'parameters given the options available. This is a fatal error.';
    41: Msg := Msg + 'no_certificate_RESERVED:'#13#10 +
       'This alert was used in SSLv3 but not in TLS. It should not be'#13#10 +
       'sent by compliant implementations.';
    42: Msg := Msg + 'bad_certificate:'#13#10 +
       'A certificate was corrupt, contained signatures that did not'#13#10 +
       'verify correctly, etc. This alert code is e.g. generated by the'#13#10 +
       'client if the server certificate is not allowed to be used for'#13#10 +
       'server authentication';
    43: Msg := Msg + 'unsupported_certificate:'#13#10 +
       'A certificate was of an unsupported type.';
    44: Msg := Msg + 'certificate_revoked:'#13#10 +
       'A certificate was revoked by its signer.';
    45: Msg := Msg + 'certificate_expired:'#13#10 +
       'A certificate has expired or is not currently valid.';
    46: Msg := Msg + 'certificate_unknown:'#13#10 +
       'Some other (unspecified) issue arose in processing the'#13#10 +
       'certificate, rendering it unacceptable.';
    47: Msg := Msg + 'illegal_parameter:'#13#10 +
       'A field in the handshake was out of range or inconsistent with'#13#10 +
       'other fields. This is always fatal.';
    48: Msg := Msg + 'unknown_ca:'#13#10 +
       'A valid certificate chain or partial chain was received, but the'#13#10 +
       'certificate was not accepted because the CA certificate could not'#13#10 +
       'be located or couldn''t be matched with a known, trusted CA.  This'#13#10 +
       'message is always fatal.';
    49: Msg := Msg + 'access_denied:'#13#10 +
       'A valid certificate was received, but when access control was'#13#10 +
       'applied, the sender decided not to proceed with negotiation.'#13#10 +
       'This message is always fatal.';
    50: Msg := Msg + 'decode_error:'#13#10 +
       'A message could not be decoded because some field was out of the'#13#10 +
       'specified range or the length of the message was incorrect. This'#13#10 +
       'message is always fatal.';
    51: Msg := Msg + 'decrypt_error:'#13#10 +
       'A handshake cryptographic operation failed, including being'#13#10 +
       'unable to correctly verify a signature, decrypt a key exchange,'#13#10 +
       'or validate a finished message.';
    60: Msg := Msg + 'export_restriction:'#13#10 +
       'A negotiation not in compliance with export restrictions was'#13#10 +
       'detected; for example, attempting to transfer a 1024 bit'#13#10 +
       'ephemeral RSA key for the RSA_EXPORT handshake method. This'#13#10 +
       'message is always fatal.';
    70: Msg := Msg + 'protocol_version:'#13#10 +
       'The protocol version the client has attempted to negotiate is'#13#10 +
       'recognized, but not supported. (For example, old protocol'#13#10 +
       'versions might be avoided for security reasons). This message is'#13#10 +
       'always fatal.';
    71: Msg := Msg + 'insufficient_security:'#13#10 +
       'Returned instead of handshake_failure when a negotiation has'#13#10 +
       'failed specifically because the server requires ciphers more'#13#10 +
       'secure than those supported by the client. This message is always'#13#10 +
       'fatal.';
    80: Msg := Msg + 'internal_error:'#13#10 +
       'An internal error unrelated to the peer or the correctness of the'#13#10 +
       'protocol makes it impossible to continue (such as a memory'#13#10 +
       'allocation failure). This message is always fatal.';
    90: Msg := Msg + 'user_canceled:'#13#10 +
       'This handshake is being canceled for some reason unrelated to a'#13#10 +
       'protocol failure. If the user cancels an operation after the'#13#10 +
       'handshake is complete, just closing the connection by sending a'#13#10 +
       'close_notify is more appropriate. This alert should be followed'#13#10 +
       'by a close_notify. This message is generally a warning.';
    100: Msg := Msg + 'no_renegotiation:'#13#10 +
       'Sent by the client in response to a hello request or by the'#13#10 +
       'server in response to a client hello after initial handshaking.'#13#10 +
       'Either of these would normally lead to renegotiation; when that'#13#10 +
       'is not appropriate, the recipient should respond with this alert'#13#10 +
       'at that point, the original requester can decide whether to'#13#10 +
       'proceed with the connection. One case where this would be'#13#10 +
       'appropriate would be where a server has spawned a process to'#13#10 +
       'satisfy a request; the process might receive security parameters'#13#10 +
       '(key length, authentication, etc.) at startup and it might be'#13#10 +
       'difficult to communicate changes to these parameters after that'#13#10 +
       'point. This message is always a warning.';
  else
    Msg := Msg + 'Unknown alert code: No description available.';
  end;
  if AlertCode <> 0 then
    MessageDlg(Msg,mtWarning,[mbAbort],0)
end;

procedure TForm1.TLSInternalServer1TLSOutgoingAlert(Sender: TObject;
  Client: TCustomTLS_ContentLayer; var Fatal: Boolean; AlertCode: Integer);
var
  Msg: string;
begin
  if AlertCode = 0 then
    Msg := 'Outgoing close notification'#13#10
  else if Fatal then
    Msg := 'Outgoing fatal alert ' + IntToStr(AlertCode) + #13#10
  else
    Msg := 'Outgoing warning ' + IntToStr(AlertCode) + #13#10;
  case AlertCode and $FF of
    0:  Msg := Msg + 'close_notify: The remote host has terminated the connection.';
    10: Msg := Msg + 'unexpected_message:'#13#10 +
       'An inappropriate message was received. This alert is always fatal'#13#10 +
       'and should never be observed in communication between proper'#13#10 +
       'implementations. Please notify StreamSec if you get this message.';
    20: Msg := Msg + 'bad_record_mac:'#13#10 +
       'This alert is returned if a record is received with an incorrect'#13#10 +
       'MAC. This alert also SHOULD be returned if a TLSCiphertext'#13#10 +
       'decrypted in an invalid way: either it wasn''t an even multiple of'#13#10 +
       'the block length, or its padding values, when checked, weren''t'#13#10 +
       'correct. This message is always fatal.';
    21: Msg := Msg + 'decryption_failed:'#13#10 +
       'This alert MAY be returned if a TLSCiphertext decrypted in an'#13#10 +
       'invalid way: either it wasn''t an even multiple of the block'#13#10 +
       'length, or its padding values, when checked, weren''t correct.'#13#10 +
       'This message is always fatal.'#13#10 +
       #13#10 +
       'NB: Differentiating between bad_record_mac and decryption_failed'#13#10 +
       'alerts may permit certain attacks against CBC mode as used in'#13#10 +
       'TLS. It is preferable to uniformly use the bad_record_mac alert'#13#10 +
       'to hide the specific type of the error.';
    22: Msg := Msg + 'record_overflow:'#13#10 +
       'A TLSCiphertext record was received which had a length more than'#13#10 +
       '2^14+2048 bytes, or a record decrypted to a TLSCompressed record'#13#10 +
       'with more than 2^14+1024 bytes. This message is always fatal.';
    30: Msg := Msg + 'decompression_failure:'#13#10 +
       'The decompression function received improper input (e.g. data'#13#10 +
       'that would expand to excessive length). This message is always'#13#10 +
       'fatal.';
    40: Msg := Msg + 'handshake_failure:'#13#10 +
       'Reception of a handshake_failure alert message indicates that the'#13#10 +
       'sender was unable to negotiate an acceptable set of security'#13#10 +
       'parameters given the options available. This is a fatal error.';
    41: Msg := Msg + 'no_certificate_RESERVED:'#13#10 +
       'This alert was used in SSLv3 but not in TLS. It should not be'#13#10 +
       'sent by compliant implementations.';
    42: Msg := Msg + 'bad_certificate:'#13#10 +
       'A certificate was corrupt, contained signatures that did not'#13#10 +
       'verify correctly, etc. This alert code is e.g. generated by the'#13#10 +
       'client if the server certificate is not allowed to be used for'#13#10 +
       'server authentication';
    43: Msg := Msg + 'unsupported_certificate:'#13#10 +
       'A certificate was of an unsupported type.';
    44: Msg := Msg + 'certificate_revoked:'#13#10 +
       'A certificate was revoked by its signer.';
    45: Msg := Msg + 'certificate_expired:'#13#10 +
       'A certificate has expired or is not currently valid.';
    46: Msg := Msg + 'certificate_unknown:'#13#10 +
       'Some other (unspecified) issue arose in processing the'#13#10 +
       'certificate, rendering it unacceptable.';
    47: Msg := Msg + 'illegal_parameter:'#13#10 +
       'A field in the handshake was out of range or inconsistent with'#13#10 +
       'other fields. This is always fatal.';
    48: Msg := Msg + 'unknown_ca:'#13#10 +
       'A valid certificate chain or partial chain was received, but the'#13#10 +
       'certificate was not accepted because the CA certificate could not'#13#10 +
       'be located or couldn''t be matched with a known, trusted CA.  This'#13#10 +
       'message is always fatal.';
    49: Msg := Msg + 'access_denied:'#13#10 +
       'A valid certificate was received, but when access control was'#13#10 +
       'applied, the sender decided not to proceed with negotiation.'#13#10 +
       'This message is always fatal.';
    50: Msg := Msg + 'decode_error:'#13#10 +
       'A message could not be decoded because some field was out of the'#13#10 +
       'specified range or the length of the message was incorrect. This'#13#10 +
       'message is always fatal.';
    51: Msg := Msg + 'decrypt_error:'#13#10 +
       'A handshake cryptographic operation failed, including being'#13#10 +
       'unable to correctly verify a signature, decrypt a key exchange,'#13#10 +
       'or validate a finished message.';
    60: Msg := Msg + 'export_restriction:'#13#10 +
       'A negotiation not in compliance with export restrictions was'#13#10 +
       'detected; for example, attempting to transfer a 1024 bit'#13#10 +
       'ephemeral RSA key for the RSA_EXPORT handshake method. This'#13#10 +
       'message is always fatal.';
    70: Msg := Msg + 'protocol_version:'#13#10 +
       'The protocol version the client has attempted to negotiate is'#13#10 +
       'recognized, but not supported. (For example, old protocol'#13#10 +
       'versions might be avoided for security reasons). This message is'#13#10 +
       'always fatal.';
    71: Msg := Msg + 'insufficient_security:'#13#10 +
       'Returned instead of handshake_failure when a negotiation has'#13#10 +
       'failed specifically because the server requires ciphers more'#13#10 +
       'secure than those supported by the client. This message is always'#13#10 +
       'fatal.';
    80: Msg := Msg + 'internal_error:'#13#10 +
       'An internal error unrelated to the peer or the correctness of the'#13#10 +
       'protocol makes it impossible to continue (such as a memory'#13#10 +
       'allocation failure). This message is always fatal.';
    90: Msg := Msg + 'user_canceled:'#13#10 +
       'This handshake is being canceled for some reason unrelated to a'#13#10 +
       'protocol failure. If the user cancels an operation after the'#13#10 +
       'handshake is complete, just closing the connection by sending a'#13#10 +
       'close_notify is more appropriate. This alert should be followed'#13#10 +
       'by a close_notify. This message is generally a warning.';
    100: Msg := Msg + 'no_renegotiation:'#13#10 +
       'Sent by the client in response to a hello request or by the'#13#10 +
       'server in response to a client hello after initial handshaking.'#13#10 +
       'Either of these would normally lead to renegotiation; when that'#13#10 +
       'is not appropriate, the recipient should respond with this alert'#13#10 +
       'at that point, the original requester can decide whether to'#13#10 +
       'proceed with the connection. One case where this would be'#13#10 +
       'appropriate would be where a server has spawned a process to'#13#10 +
       'satisfy a request; the process might receive security parameters'#13#10 +
       '(key length, authentication, etc.) at startup and it might be'#13#10 +
       'difficult to communicate changes to these parameters after that'#13#10 +
       'point. This message is always a warning.';
  else
    Msg := Msg + 'Unknown alert code: No description available.';
  end;
  if AlertCode <> 0 then begin
    if Fatal then
      MessageDlg(Msg,mtWarning,[mbAbort],0)
    else
      Fatal := MessageDlg(Msg,mtWarning,[mbIgnore,mbAbort],0) = mrAbort;
  end;
end;

end.
