{Note: The password is 'abc'}
unit SMIMEMailSendMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, SecUtils, Asn1, MpCms, MpX509, SecComp, StreamSecII,
  ResourceFile, TlsInternalServer, SsCertMgr;

const
  CharSet = 'iso-8859-1'; // Localize

type
  TfrmSMIMEMailSendMain = class(TForm)
    Label1: TLabel;
    edtFrom: TEdit;
    Label2: TLabel;
    edtTo: TEdit;
    Label3: TLabel;
    edtSubject: TEdit;
    Label4: TLabel;
    edtHost: TEdit;
    Label5: TLabel;
    edtUsername: TEdit;
    Label6: TLabel;
    edtPassword: TEdit;
    memBody: TMemo;
    btnSend: TButton;
    PrivateKeyRing: TStreamSecII;
    MyCerts: TX509TrustedCertificates;
    PrivateKeyRingFile: TResourceFile;
    MyCertsFile: TResourceFile;
    Label7: TLabel;
    SignedData: TMPSignedData;
    EnvelopedData: TMPEnvelopedData;
    chbEncrypt: TCheckBox;
    TestRecipient: TResourceFile;
    KeysAndCerts: TSimpleTLSInternalServer;
    TrustedCerts: TTLSInternalServer;
    SimpleTLSInternalServer1: TSimpleTLSInternalServer;
    procedure PrivateKeyRingPassword(Sender: TObject;
      Password: ISecretKey);
    procedure btnSendClick(Sender: TObject);
    procedure SimpleTLSInternalServer1BeforeImportTLSCert(Sender: TObject;
      Cert: TASN1Struct; var ExplicitTrust, AllowExpired: Boolean);
  private
    function GetInnerMIMEEntity: string;
    procedure EnvelopeData(AMsgData: TStrings);
    procedure SignData(AMsgData: TStrings);
  public
    { Public declarations }
  end;

var
  frmSMIMEMailSendMain: TfrmSMIMEMailSendMain;

implementation

uses
  SMTPSend, SynaUtil,
  Pkix, SsBase64, X509Base, SsArc2;

{$R *.dfm}

procedure TfrmSMIMEMailSendMain.PrivateKeyRingPassword(Sender: TObject;
  Password: ISecretKey);
begin
  Password.SetLength(3);
  Password.SetKeyStrAt('abc',0);
end;

function TfrmSMIMEMailSendMain.GetInnerMIMEEntity: string;
begin
  if chbEncrypt.Checked then begin
    Result := 'From: ' + edtFrom.Text + #13#10 +
              'To: ' + edtTo.Text + #13#10 +
              'Date: ' + Rfc822DateTime(Now) + #13#10 +
              'Subject: ' + edtSubject.Text + #13#10 +
              'X-mailer: StreamSec StrSecII S/MIME'#13#10 +
              'Content-Type: text/plain;'#13#10 +
              #9'charset="' + CharSet + '"'#13#10 +
              'Content-Transfer-Encoding: base64'#13#10;
    Result := Result + StrToMime64('','',memBody.Text);
  end else begin
    Result := 'Content-Type: text/plain;'#13#10 +
              #9'charset="' + CharSet + '"'#13#10 +
              'Content-Transfer-Encoding: base64'#13#10;
    Result := Result + StrToMime64('','',memBody.Text);

(*
  Result := 'Content-Type: text/plain;'#13#10 +
            #9'charset="' + CharSet + '"'#13#10 +
            'Content-Transfer-Encoding: 8bit'#13#10 +
            #13#10 +
            memBody.Text;

  while Result[Length(Result)] = #10 do
    Delete(Result,Length(Result)-1,2);
  Result := Result + #13#10;
  *)
  end;
end;

procedure TfrmSMIMEMailSendMain.SignData(AMsgData: TStrings);
var
  Data, Signature: TStringStream;
  Idx: Integer;
  Cert: PASN1Struct;
  Boundary: string;
begin
  Idx := 0;
  repeat
    Cert := nil;
    MyCerts.FindCert([],[],[digitalSignature],[id_kp_emailProtection],Idx,Cert);
  until (Cert = nil) or X509Base.CheckEmailAddress(Cert^,edtFrom.Text);
  if Cert = nil then
    raise Exception.Create('Could not find email certificate');

  Data := TStringStream.Create(GetInnerMIMEEntity);
  try
    Signature := TStringStream.Create('');
    try
//      SignedData.SigningDescription := edtSubject.Text;    
      SignedData.DetachedSignature := True;
      SignedData.IncludeRootCertificate := True;
      SignedData.SignData(Cert^,Data,Signature);
      AMsgData.Clear;
      AMsgData.Add('content-class: urn:content-classes:message');
      AMsgData.Add('MIME-Version: 1.0');
      AMsgData.Add('Content-Type: multipart/signed;');
      Randomize;
      Boundary := '----=_NextPart_' + IntToStr(Random($7FFFFFFF));
      AMsgData.Add(#9'boundary="' + Boundary + '";');
      AMsgData.Add(#9'protocol="application/x-pkcs7-signature";');
      AMsgData.Add(#9'micalg=SHA1');
      AMsgData.Add('');
      AMsgData.Add('This is a multi-part message in MIME format.');
      AMsgData.Add('');
      AMsgData.Add('--' + Boundary);
      AMsgData.Add(Data.DataString);
      AMsgData.Add('--' + Boundary);
      AMsgData.Add('Content-Type: application/x-pkcs7-signature;');
      AMsgData.Add(#9'name="smime.p7s"');
      AMsgData.Add('Content-Transfer-Encoding: base64');
      AMsgData.Add('Content-Disposition: attachment;');
      AMsgData.Add(#9'filename="smime.p7s"');
      AMsgData.Add(StrToMime64('','',Signature.DataString));
      AMsgData.Add('--' + Boundary + '--');
    finally
      Signature.Free;
    end;
  finally
    Data.Free;
  end;
end;

procedure TfrmSMIMEMailSendMain.btnSendClick(Sender: TObject);
var
  Cursor: TCursor;
  MsgData: TStringList;
begin
  Cursor := Screen.Cursor;
  Screen.Cursor := crAppStart;
  try
    MsgData := TStringList.Create;
    try
      if chbEncrypt.Checked then
        EnvelopeData(MsgData)
      else
        SignData(MsgData);
      if not SMTPSend.SendToEx(edtFrom.Text,
                               edtTo.Text,
                               edtSubject.Text,
                               edtHost.Text,
                               MsgData,
                               edtUserName.Text,
                               edtPassWord.Text) then
        raise Exception.Create('Unable to send email');
    finally
      MsgData.Free;
    end;
  finally
    Screen.Cursor := Cursor;
  end;
end;

procedure TfrmSMIMEMailSendMain.EnvelopeData(AMsgData: TStrings);
var
  Data, Signature, Envelope: TStringStream;
  Idx: Integer;
  Cert: PASN1Struct;
  RCert: PASN1Struct;
  Boundary: string;
begin
  Idx := 0;
  repeat
    Cert := nil;
    MyCerts.FindCert([],[],[digitalSignature],[id_kp_emailProtection],Idx,Cert);
  until (Cert = nil) or X509Base.CheckEmailAddress(Cert^,edtFrom.Text);
  if Cert = nil then
    raise Exception.Create('Could not find email certificate');

  Idx := 0;
  repeat
    RCert := nil;
    EnvelopedData.CollectedCertificates.FindCert([],[rsaEncryption],[keyEncipherment],[id_kp_emailProtection],Idx,RCert);
  until (RCert = nil) or X509Base.CheckEmailAddress(RCert^,edtTo.Text);
  if RCert = nil then
    raise Exception.Create('Could not find email certificate');

  Data := TStringStream.Create(GetInnerMIMEEntity);
  try
    Signature := TStringStream.Create('');
    try
//      SignedData.SigningDescription := edtSubject.Text;
      SignedData.DetachedSignature := False;
      SignedData.IncludeRootCertificate := True;
      SignedData.SignData(Cert^,Data,Signature);

      AMsgData.Clear;
      AMsgData.Add('Content-Type: application/pkcs7-mime;');
      AMsgData.Add(#9'smime-type=signed-data;');
      AMsgData.Add(#9'name=smime.p7m');
      AMsgData.Add('Content-Transfer-Encoding: base64');
      AMsgData.Add('Content-Disposition: attachment; filename=smime.p7m');
      AMsgData.Add(StrToMime64('','',Signature.DataString));
      AMsgData.Add('.');
      
      Data.Size := 0;
      Data.WriteString(AMsgData.Text);
      Data.Position := 0;
      Envelope := TStringStream.Create('');
      try
        EnvelopedData.Encrypt(RCert^,Data,Envelope);

        AMsgData.Clear;
        AMsgData.Add('content-class: urn:content-classes:message');
        AMsgData.Add('MIME-Version: 1.0');
        AMsgData.Add('Content-Type: application/pkcs7-mime;');
        AMsgData.Add(#9'smime-type=enveloped-data;');
        AMsgData.Add(#9'name=smime.p7m');
        AMsgData.Add('Content-Transfer-Encoding: base64');
        AMsgData.Add('Content-Disposition: attachment; filename=smime.p7m');
        AMsgData.Add(StrToMime64('','',Envelope.DataString));
      finally
        Envelope.Free;
      end;
    finally
      Signature.Free;
    end;
  finally
    Data.Free;
  end;
end;

procedure TfrmSMIMEMailSendMain.SimpleTLSInternalServer1BeforeImportTLSCert(
  Sender: TObject; Cert: TASN1Struct; var ExplicitTrust,
  AllowExpired: Boolean);
begin
  AllowExpired := True;
  ExplicitTrust := True;
end;

end.
