unit TLSServerMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, SecComp, StreamSecII, MpX509, TlsInternalServer, StdCtrls,
  SecUtils, TlsSynaSock, TlsClass;

type
  TForm1 = class(TForm)
    Button1: TButton;
    OpenDialog1: TOpenDialog;
    SimpleTLSInternalServer1: TSimpleTLSInternalServer;
    SsPrivateKeyRingComponent1: TSsPrivateKeyRingComponent;
    memLog: TMemo;
    Button2: TButton;
    Button3: TButton;
    procedure SsPrivateKeyRingComponent1AdminPassword(Sender: TObject;
      Password: ISecretKey);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure SimpleTLSInternalServer1TLSIncomingAlert(Sender: TObject;
      Client: TCustomTLS_ContentLayer; var Fatal: Boolean;
      AlertCode: Integer);
    procedure SimpleTLSInternalServer1TLSOutgoingAlert(Sender: TObject;
      Client: TCustomTLS_ContentLayer; var Fatal: Boolean;
      AlertCode: Integer);
    procedure Button3Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
  private
    FServer: TTLSSynaServer;
    procedure NewConnect(Sender: TObject; Socket: TCustomTLSSynSock);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses
  TlsConst;

type
  TLogThread = class(TThread)
  private
    FLines: TStrings;
    FMessage: string;
  protected
    procedure DoAddMessage;
    procedure Execute; override;
  public
    constructor Create(ALines: TStrings; const AMessage: string);
  end;

{ TLogThread }

constructor TLogThread.Create(ALines: TStrings; const AMessage: string);
begin
  FLines := ALines;
  FMessage := AMessage;
  FreeOnTerminate := True;
  inherited Create(False);
end;

procedure TLogThread.DoAddMessage;
begin
  FLines.Add(FMessage);
end;

procedure TLogThread.Execute;
begin
  Synchronize(DoAddMessage);
end;

{$R *.dfm}

procedure TForm1.SsPrivateKeyRingComponent1AdminPassword(Sender: TObject;
  Password: ISecretKey);
begin
  Password.SetKeyStr('abc');
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  PW: iSecretKey;
begin
  PW := nil;
  if OpenDialog1.Execute then begin
    SimpleTLSInternalServer1.ImportFromPFX(OpenDialog1.FileName,PW);
    if pkaRSA in SimpleTLSInternalServer1.PublicKeyAlgorithms then begin
      SimpleTLSInternalServer1.Options.SignatureRSA := prPrefer;
      SimpleTLSInternalServer1.Options.KeyAgreementRSA := prAllowed;
      SimpleTLSInternalServer1.Options.KeyAgreementDHE := prPrefer;
      SimpleTLSInternalServer1.TLSSetupServer;
      Button2.Enabled := True;
    end;
  end;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  if not Assigned(FServer) then begin
    FServer := TTLSSynaServer.Create(Self);
    FServer.OnNewConnect := NewConnect;
    FServer.Port := 802;
    FServer.Start;
    Button2.Enabled := False;
    Button3.Enabled := True;
  end;
end;

procedure TForm1.SimpleTLSInternalServer1TLSIncomingAlert(Sender: TObject;
  Client: TCustomTLS_ContentLayer; var Fatal: Boolean; AlertCode: Integer);
begin
  TLogThread.Create(Form1.memLog.Lines,TlsConst.AlertMsg(AlertCode));
end;

procedure TForm1.SimpleTLSInternalServer1TLSOutgoingAlert(Sender: TObject;
  Client: TCustomTLS_ContentLayer; var Fatal: Boolean; AlertCode: Integer);
begin
  TLogThread.Create(Form1.memLog.Lines,TlsConst.AlertMsg(AlertCode));
end;

procedure TForm1.Button3Click(Sender: TObject);
begin
  if Assigned(FServer) then begin
    Button3.Enabled := False;
    FServer.Stop;
    FServer.Free;
    FServer := nil;
    Button2.Enabled := True;
  end;
end;

procedure TForm1.NewConnect(Sender: TObject; Socket: TCustomTLSSynSock);
var
  Ticks: Cardinal;
  Len: Integer;
  S: string;
begin
  Ticks := GetTickCount;
  repeat
    Len := Socket.ReceiveLength;
  until (Len > 0) or (GetTickCount - Ticks > 5000);
  if Len > 0 then begin
    SetLength(S,Len);
    Socket.ReceiveBuf(Pointer(S)^,Len);
    TLogThread.Create(memLog.Lines,Format('%s: %s',[Socket.RemoteHost,S]));
  end else
    TLogThread.Create(memLog.Lines,Format('Timeout waiting for %s',[Socket.RemoteHost]));
end;

procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := FServer = nil;
  if not CanClose then Beep;
end;

end.
 