{NOTE: This demo requires unit ScktComp. You must install the package containing
 this unit before running this demo.}
{.$DEFINE P7B}
unit BasicTLSClientDM;

interface

uses
  SysUtils, Classes, Forms, Windows, MpYarrow, MpX509, StreamSecII, ScktComp,
  Asn1, Tls, TlsClass, SecComp, TlsInternalServer;

type
  TdmBasicTLSClient = class(TDataModule)
    ClientSocket1: TClientSocket;
    MPYarrow1: TMPYarrow;
    StreamSecII1: TStreamSecII;
    CollectedCerts: TX509TrustedCertificates;
    CACerts: TX509TrustedCertificates;
    procedure ApplicationEvents1Message(var Msg: tagMSG;
      var Handled: Boolean);
    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);
    procedure ClientSocket1Connect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure StreamSecII1TLSIncomingAlert(Sender: TObject;
      Client: TCustomTLS_ContentLayer; var Fatal: Boolean;
      AlertCode: Integer);
    procedure StreamSecII1TLSOutgoingAlert(Sender: TObject;
      Client: TCustomTLS_ContentLayer; var Fatal: Boolean;
      AlertCode: Integer);
    procedure ClientSocket1Read(Sender: TObject; Socket: TCustomWinSocket);
    procedure StreamSecII1TLSChangeCipherSpec(Sender: TObject;
      Client: TCustomTLS_ContentLayer);
    procedure ClientSocket1Disconnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure CollectedCertsCertNotTrusted(Sender: TObject;
      Cert: TASN1Struct; var ExplicitTrust: Boolean);
    procedure CollectedCertsCertNotAccepted(Sender: TObject;
      Cert: TASN1Struct; Status: TCertStatusCode);
  private
    FLog: TStrings;
    FOnLog: TNotifyEvent;
    FTLSClient: TCustomTLS_ContentLayer;
    FSessionID: TSessionID;
    procedure LogEntry(const Entry: string);
    procedure SetOnLog(const Value: TNotifyEvent);
  public
    procedure Connect(const Service, URL: string);
    property Log: TStrings read FLog;
    property OnLog: TNotifyEvent read FOnLog write SetOnLog;
  end;

var
  dmBasicTLSClient: TdmBasicTLSClient;

implementation

uses
  SecUtils, ReadStrm, Pkix, X509Base, BasicTLSClientMain, Dialogs;

{$R *.dfm}

procedure TdmBasicTLSClient.ApplicationEvents1Message(var Msg: tagMSG;
  var Handled: Boolean);
begin
  // This is an acceptable way to collect entropy for random number generation:
  MPYarrow1.Accumulate(Msg,SizeOf(Msg));

  OnLog := frmBasicTLSClient.TLSLog;
end;

procedure TdmBasicTLSClient.Connect(const Service, URL: string);
var
  Port, Code: Integer;
begin
  Val(Service,Port,Code);
  if Code = 0 then
    ClientSocket1.Port := Port
  else
    ClientSocket1.Service := Service;
  ClientSocket1.Host := URL;
  ClientSocket1.Open;
end;

procedure TdmBasicTLSClient.DataModuleCreate(Sender: TObject);
var
  {$IFDEF P7B}
  Status: TCertStatusCodes;
  {$ENDIF}
  Ticks: Cardinal;
begin
  FLog := TStringList.Create;
  Ticks := GetTickCount;
  {$IFDEF P7B}
  CACerts.ImportFromCMSFile('RootCerts.p7b',True,Status);
  {$ELSE}
  CACerts.LoadFromFile('Root.scl');
  {$ENDIF}                             
  Ticks := GetTickCount - Ticks;
  ShowMessage(IntToStr(Ticks));
  {$IFDEF P7B}
  if Status <> [crcOK] then
    ShowMessage('Failure');
  {$ENDIF}
  Application.OnMessage := ApplicationEvents1Message;
end;

procedure TdmBasicTLSClient.DataModuleDestroy(Sender: TObject);
begin
  FLog.Free;
end;

procedure TdmBasicTLSClient.LogEntry(const Entry: string);
begin
  FLog.Add(DateTimeToStr(Now) + #13#10 + Entry + #13#10);
  if Assigned(FOnLog) then
    FOnLog(Self);
end;

procedure TdmBasicTLSClient.SetOnLog(const Value: TNotifyEvent);
begin
  FOnLog := Value;
end;

procedure TdmBasicTLSClient.ClientSocket1Connect(Sender: TObject;
  Socket: TCustomWinSocket);
var
  MS: TSecureMemoryStream;
begin
  LogEntry('Connecting to ' + Socket.RemoteHost + #13#10);
  MS := TSecureMemoryStream.Create;
  StreamSecII1.TLSConnect(Socket,MS);
  MS.Position := 0;
  Socket.SendStream(MS);
  StreamSecII1.FindTLSSession(Socket,FTLSClient);
end;

procedure TdmBasicTLSClient.StreamSecII1TLSIncomingAlert(Sender: TObject;
  Client: TCustomTLS_ContentLayer; var Fatal: Boolean; AlertCode: Integer);
var
  Msg: string;
begin
  if AlertCode = 0 then
    Msg := 'Incoming close notification'#13#10
  else if Fatal then
    Msg := 'Incoming fatal alert ' + IntToStr(AlertCode) + #13#10
  else
    Msg := 'Incoming warning ' + IntToStr(AlertCode) + #13#10;
  case AlertCode and $FF of
    0:  Msg := Msg + 'close_notify: The remote host has terminated the connection.';
    10: Msg := Msg + 'unexpected_message:'#13#10 +
       'An inappropriate message was received. This alert is always fatal'#13#10 +
       'and should never be observed in communication between proper'#13#10 +
       'implementations. Please notify StreamSec if you get this message.';
    20: Msg := Msg + 'bad_record_mac:'#13#10 +
       'This alert is returned if a record is received with an incorrect'#13#10 +
       'MAC. This alert also SHOULD be returned if a TLSCiphertext'#13#10 +
       'decrypted in an invalid way: either it wasn''t an even multiple of'#13#10 +
       'the block length, or its padding values, when checked, weren''t'#13#10 +
       'correct. This message is always fatal.';
    21: Msg := Msg + 'decryption_failed:'#13#10 +
       'This alert MAY be returned if a TLSCiphertext decrypted in an'#13#10 +
       'invalid way: either it wasn''t an even multiple of the block'#13#10 +
       'length, or its padding values, when checked, weren''t correct.'#13#10 +
       'This message is always fatal.'#13#10 +
       #13#10 +
       'NB: Differentiating between bad_record_mac and decryption_failed'#13#10 +
       'alerts may permit certain attacks against CBC mode as used in'#13#10 +
       'TLS. It is preferable to uniformly use the bad_record_mac alert'#13#10 +
       'to hide the specific type of the error.';
    22: Msg := Msg + 'record_overflow:'#13#10 +
       'A TLSCiphertext record was received which had a length more than'#13#10 +
       '2^14+2048 bytes, or a record decrypted to a TLSCompressed record'#13#10 +
       'with more than 2^14+1024 bytes. This message is always fatal.';
    30: Msg := Msg + 'decompression_failure:'#13#10 +
       'The decompression function received improper input (e.g. data'#13#10 +
       'that would expand to excessive length). This message is always'#13#10 +
       'fatal.';
    40: Msg := Msg + 'handshake_failure:'#13#10 +
       'Reception of a handshake_failure alert message indicates that the'#13#10 +
       'sender was unable to negotiate an acceptable set of security'#13#10 +
       'parameters given the options available. This is a fatal error.';
    41: Msg := Msg + 'no_certificate_RESERVED:'#13#10 +
       'This alert was used in SSLv3 but not in TLS. It should not be'#13#10 +
       'sent by compliant implementations.';
    42: Msg := Msg + 'bad_certificate:'#13#10 +
       'A certificate was corrupt, contained signatures that did not'#13#10 +
       'verify correctly, etc.';
    43: Msg := Msg + 'unsupported_certificate:'#13#10 +
       'A certificate was of an unsupported type.';
    44: Msg := Msg + 'certificate_revoked:'#13#10 +
       'A certificate was revoked by its signer.';
    45: Msg := Msg + 'certificate_expired:'#13#10 +
       'A certificate has expired or is not currently valid.';
    46: Msg := Msg + 'certificate_unknown:'#13#10 +
       'Some other (unspecified) issue arose in processing the'#13#10 +
       'certificate, rendering it unacceptable.';
    47: Msg := Msg + 'illegal_parameter:'#13#10 +
       'A field in the handshake was out of range or inconsistent with'#13#10 +
       'other fields. This is always fatal.';
    48: Msg := Msg + 'unknown_ca:'#13#10 +
       'A valid certificate chain or partial chain was received, but the'#13#10 +
       'certificate was not accepted because the CA certificate could not'#13#10 +
       'be located or couldn''t be matched with a known, trusted CA.  This'#13#10 +
       'message is always fatal.';
    49: Msg := Msg + 'access_denied:'#13#10 +
       'A valid certificate was received, but when access control was'#13#10 +
       'applied, the sender decided not to proceed with negotiation.'#13#10 +
       'This message is always fatal.';
    50: Msg := Msg + 'decode_error:'#13#10 +
       'A message could not be decoded because some field was out of the'#13#10 +
       'specified range or the length of the message was incorrect. This'#13#10 +
       'message is always fatal.';
    51: Msg := Msg + 'decrypt_error:'#13#10 +
       'A handshake cryptographic operation failed, including being'#13#10 +
       'unable to correctly verify a signature, decrypt a key exchange,'#13#10 +
       'or validate a finished message.';
    60: Msg := Msg + 'export_restriction:'#13#10 +
       'A negotiation not in compliance with export restrictions was'#13#10 +
       'detected; for example, attempting to transfer a 1024 bit'#13#10 +
       'ephemeral RSA key for the RSA_EXPORT handshake method. This'#13#10 +
       'message is always fatal.';
    70: Msg := Msg + 'protocol_version:'#13#10 +
       'The protocol version the client has attempted to negotiate is'#13#10 +
       'recognized, but not supported. (For example, old protocol'#13#10 +
       'versions might be avoided for security reasons). This message is'#13#10 +
       'always fatal.';
    71: Msg := Msg + 'insufficient_security:'#13#10 +
       'Returned instead of handshake_failure when a negotiation has'#13#10 +
       'failed specifically because the server requires ciphers more'#13#10 +
       'secure than those supported by the client. This message is always'#13#10 +
       'fatal.';
    80: Msg := Msg + 'internal_error:'#13#10 +
       'An internal error unrelated to the peer or the correctness of the'#13#10 +
       'protocol makes it impossible to continue (such as a memory'#13#10 +
       'allocation failure). This message is always fatal.';
    90: Msg := Msg + 'user_canceled:'#13#10 +
       'This handshake is being canceled for some reason unrelated to a'#13#10 +
       'protocol failure. If the user cancels an operation after the'#13#10 +
       'handshake is complete, just closing the connection by sending a'#13#10 +
       'close_notify is more appropriate. This alert should be followed'#13#10 +
       'by a close_notify. This message is generally a warning.';
    100: Msg := Msg + 'no_renegotiation:'#13#10 +
       'Sent by the client in response to a hello request or by the'#13#10 +
       'server in response to a client hello after initial handshaking.'#13#10 +
       'Either of these would normally lead to renegotiation; when that'#13#10 +
       'is not appropriate, the recipient should respond with this alert'#13#10 +
       'at that point, the original requester can decide whether to'#13#10 +
       'proceed with the connection. One case where this would be'#13#10 +
       'appropriate would be where a server has spawned a process to'#13#10 +
       'satisfy a request; the process might receive security parameters'#13#10 +
       '(key length, authentication, etc.) at startup and it might be'#13#10 +
       'difficult to communicate changes to these parameters after that'#13#10 +
       'point. This message is always a warning.';
  else
    Msg := Msg + 'Unknown alert code: No description available.';
  end;

  LogEntry(Msg + #13#10);
end;

procedure TdmBasicTLSClient.StreamSecII1TLSOutgoingAlert(Sender: TObject;
  Client: TCustomTLS_ContentLayer; var Fatal: Boolean; AlertCode: Integer);
var
  Msg: string;
begin
  if AlertCode = 0 then
    Msg := 'Outgoing close notification'#13#10
  else if Fatal then
    Msg := 'Outgoing fatal alert ' + IntToStr(AlertCode) + #13#10
  else
    Msg := 'Outgoing warning ' + IntToStr(AlertCode) + #13#10;
  case AlertCode and $FF of
    0:  Msg := Msg + 'close_notify: The connection is about to be terminated.';
    10: Msg := Msg + 'unexpected_message:'#13#10 +
       'An inappropriate message was received. This alert is always fatal'#13#10 +
       'and should never be observed in communication between proper'#13#10 +
       'implementations.';
    20: Msg := Msg + 'bad_record_mac:'#13#10 +
       'This alert is returned if a record is received with an incorrect'#13#10 +
       'MAC. This alert also SHOULD be returned if a TLSCiphertext'#13#10 +
       'decrypted in an invalid way: either it wasn''t an even multiple of'#13#10 +
       'the block length, or its padding values, when checked, weren''t'#13#10 +
       'correct. This message is always fatal.';
    21: Msg := Msg + 'decryption_failed:'#13#10 +
       'This alert MAY be returned if a TLSCiphertext decrypted in an'#13#10 +
       'invalid way: either it wasn''t an even multiple of the block'#13#10 +
       'length, or its padding values, when checked, weren''t correct.'#13#10 +
       'This message is always fatal.'#13#10 +
       #13#10 +
       'NB: Differentiating between bad_record_mac and decryption_failed'#13#10 +
       'alerts may permit certain attacks against CBC mode as used in'#13#10 +
       'TLS. It is preferable to uniformly use the bad_record_mac alert'#13#10 +
       'to hide the specific type of the error.';
    22: Msg := Msg + 'record_overflow:'#13#10 +
       'A TLSCiphertext record was received which had a length more than'#13#10 +
       '2^14+2048 bytes, or a record decrypted to a TLSCompressed record'#13#10 +
       'with more than 2^14+1024 bytes. This message is always fatal.';
    30: Msg := Msg + 'decompression_failure:'#13#10 +
       'The decompression function received improper input (e.g. data'#13#10 +
       'that would expand to excessive length). This message is always'#13#10 +
       'fatal.';
    40: Msg := Msg + 'handshake_failure:'#13#10 +
       'Reception of a handshake_failure alert message indicates that the'#13#10 +
       'sender was unable to negotiate an acceptable set of security'#13#10 +
       'parameters given the options available. This is a fatal error.';
    41: Msg := Msg + 'no_certificate_RESERVED:'#13#10 +
       'This alert was used in SSLv3 but not in TLS. It should not be'#13#10 +
       'sent by compliant implementations.';
    42: Msg := Msg + 'bad_certificate:'#13#10 +
       'A certificate was corrupt, contained signatures that did not'#13#10 +
       'verify correctly, etc.';
    43: Msg := Msg + 'unsupported_certificate:'#13#10 +
       'A certificate was of an unsupported type.';
    44: Msg := Msg + 'certificate_revoked:'#13#10 +
       'A certificate was revoked by its signer.';
    45: Msg := Msg + 'certificate_expired:'#13#10 +
       'A certificate has expired or is not currently valid.';
    46: Msg := Msg + 'certificate_unknown:'#13#10 +
       'Some other (unspecified) issue arose in processing the'#13#10 +
       'certificate, rendering it unacceptable.';
    47: Msg := Msg + 'illegal_parameter:'#13#10 +
       'A field in the handshake was out of range or inconsistent with'#13#10 +
       'other fields. This is always fatal.';
    48: Msg := Msg + 'unknown_ca:'#13#10 +
       'A valid certificate chain or partial chain was received, but the'#13#10 +
       'certificate was not accepted because the CA certificate could not'#13#10 +
       'be located or couldn''t be matched with a known, trusted CA.  This'#13#10 +
       'message is always fatal.';
    49: Msg := Msg + 'access_denied:'#13#10 +
       'A valid certificate was received, but when access control was'#13#10 +
       'applied, the sender decided not to proceed with negotiation.'#13#10 +
       'This message is always fatal.';
    50: Msg := Msg + 'decode_error:'#13#10 +
       'A message could not be decoded because some field was out of the'#13#10 +
       'specified range or the length of the message was incorrect. This'#13#10 +
       'message is always fatal.';
    51: Msg := Msg + 'decrypt_error:'#13#10 +
       'A handshake cryptographic operation failed, including being'#13#10 +
       'unable to correctly verify a signature, decrypt a key exchange,'#13#10 +
       'or validate a finished message.';
    60: Msg := Msg + 'export_restriction:'#13#10 +
       'A negotiation not in compliance with export restrictions was'#13#10 +
       'detected; for example, attempting to transfer a 1024 bit'#13#10 +
       'ephemeral RSA key for the RSA_EXPORT handshake method. This'#13#10 +
       'message is always fatal.';
    70: Msg := Msg + 'protocol_version:'#13#10 +
       'The protocol version the client has attempted to negotiate is'#13#10 +
       'recognized, but not supported. (For example, old protocol'#13#10 +
       'versions might be avoided for security reasons). This message is'#13#10 +
       'always fatal.';
    71: Msg := Msg + 'insufficient_security:'#13#10 +
       'Returned instead of handshake_failure when a negotiation has'#13#10 +
       'failed specifically because the server requires ciphers more'#13#10 +
       'secure than those supported by the client. This message is always'#13#10 +
       'fatal.';
    80: Msg := Msg + 'internal_error:'#13#10 +
       'An internal error unrelated to the peer or the correctness of the'#13#10 +
       'protocol makes it impossible to continue (such as a memory'#13#10 +
       'allocation failure). This message is always fatal.';
    90: Msg := Msg + 'user_canceled:'#13#10 +
       'This handshake is being canceled for some reason unrelated to a'#13#10 +
       'protocol failure. If the user cancels an operation after the'#13#10 +
       'handshake is complete, just closing the connection by sending a'#13#10 +
       'close_notify is more appropriate. This alert should be followed'#13#10 +
       'by a close_notify. This message is generally a warning.';
    100: Msg := Msg + 'no_renegotiation:'#13#10 +
       'Sent by the client in response to a hello request or by the'#13#10 +
       'server in response to a client hello after initial handshaking.'#13#10 +
       'Either of these would normally lead to renegotiation; when that'#13#10 +
       'is not appropriate, the recipient should respond with this alert'#13#10 +
       'at that point, the original requester can decide whether to'#13#10 +
       'proceed with the connection. One case where this would be'#13#10 +
       'appropriate would be where a server has spawned a process to'#13#10 +
       'satisfy a request; the process might receive security parameters'#13#10 +
       '(key length, authentication, etc.) at startup and it might be'#13#10 +
       'difficult to communicate changes to these parameters after that'#13#10 +
       'point. This message is always a warning.';
  else
    Msg := Msg + 'Unknown alert code: No description available.';
  end;

  LogEntry(Msg + #13#10);

end;

procedure TdmBasicTLSClient.ClientSocket1Read(Sender: TObject;
  Socket: TCustomWinSocket);
var
  InStrm, Data, OutStrm: TSecureMemoryStream;
begin
  LogEntry('Incoming message' + #13#10);
  InStrm := TSecureMemoryStream.Create;
  try
    Data := TSecureMemoryStream.Create;
    try
      InStrm.Size := Socket.ReceiveLength;
      if InStrm.Size > 0 then begin
        Socket.ReceiveBuf(InStrm.Memory^,InStrm.Size);
        OutStrm := TSecureMemoryStream.Create;

        StreamSecII1.TLSDecodeData(Socket,InStrm,Data,OutStrm);
        if OutStrm.Size > 0 then begin
          OutStrm.Position := 0;
          Socket.SendStream(OutStrm);
        end else if Assigned(FTLSClient) and (FTLSClient.Encrypted or
                    not FTLSClient.Active) then begin
          // This is just a demo of the handshake protocol.
          // The handshake is done so we close the connection:
          StreamSecII1.TLSClose(Socket,OutStrm);
          FTLSClient := nil;
          OutStrm.Position := 0;
          Socket.SendStream(OutStrm);
        end;

        if Assigned(FTLSClient) then
          FSessionID := FTLSClient.SessionID;
      end;
    finally
      Data.Free;
    end;
  finally
    InStrm.Free;
  end;
end;

procedure TdmBasicTLSClient.StreamSecII1TLSChangeCipherSpec(
  Sender: TObject; Client: TCustomTLS_ContentLayer);
var
  KS: Integer;
  PK: TX509PublicKey;
  Alg: TX509AlgorithmIdentifier;
  XName: TX501Name;
  Msg: string;
begin
  Msg := 'Connection is encrypted:'#13#10;
  if Client.ServerCertificate = nil then
    Msg := Msg + 'Warning: Anonymous connection - vulnerable to man in the middle attacks.'#13#10
  else begin
    KS := 0;
    ExtractSubjectPublicKeySize(Client.ServerCertificate^,KS);
    Msg := Msg + '  Server certificate:'#13#10 +
      '    Public Key Size (normalized ): ' +
      IntToStr(KS) + #13#10;
    PK := nil;
    try
      if ExtractSubjectPublicKey(Client.ServerCertificate^,PK) = E_OK then
        Msg := Msg + '    Public Key Algorithm: ' +
          GetObjectName(PK.Algorithm.Algorithm) + #13#10;
    finally
      PK.Free;
    end;
    Alg := nil;
    try
      if ExtractSignatureAlgorithmCertOrCRL(Client.ServerCertificate^,Alg) = E_OK then
        Msg := Msg + '    Signature Algorithm: ' +
          GetObjectName(Alg.Algorithm) + #13#10;
    finally
      Alg.Free;
    end;

    ExtractSubject(Client.ServerCertificate^,XName);
    Msg := Msg + '    Subject Name: ' + XName.commonName.Str + #13#10;
    XName.commonName.Str := '';
    ExtractIssuer(Client.ServerCertificate^,XName);
    Msg := Msg + '    Issuer Name: ' + XName.commonName.Str + #13#10;
  end;

  Msg := Msg + '  Symmetric encryption key size: ' +
    IntToStr(Client.Context^.key_size*8) + ' bits'#13#10;
  case Client.Context^.bulk_cipher_algorithm of
    bcaRC4,
    bcaRC4_40,
    bcaRC4_56: Msg := Msg + '  Symmetric encryption algorithm: RC4'#13#10;
    bcaDES,
    bcaDES40,
    bcaDES56: Msg := Msg + '  Symmetric encryption algorithm: DES'#13#10;
    bca3DES:  Msg := Msg + '  Symmetric encryption algorithm: TripleDES'#13#10;
    bcaAES128,
    bcaAES256:Msg := Msg + '  Symmetric encryption algorithm: AES'#13#10;
  end;
  case Client.Context^.mac_algorithm of
    haMD5:  Msg := Msg + '  MAC algorithm: MD5'#13#10;
    haSHA1: Msg := Msg + '  MAC algorithm: SHA1'#13#10;
  end;

  LogEntry(Msg + #13#10);
end;

procedure TdmBasicTLSClient.ClientSocket1Disconnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  if Assigned(FTLSClient) then begin
    FTLSClient.Free;
    FTLSClient := nil;
  end;
end;

procedure TdmBasicTLSClient.CollectedCertsCertNotTrusted(Sender: TObject;
  Cert: TASN1Struct; var ExplicitTrust: Boolean);
begin
  LogEntry('Certificate not trusted'#13#10);
  ExplicitTrust := True;
end;

procedure TdmBasicTLSClient.CollectedCertsCertNotAccepted(Sender: TObject;
  Cert: TASN1Struct; Status: TCertStatusCode);
begin
  if Status = crcExpired then
    LogEntry('Expired Certificate'#13#10);
end;

end.
