{NOTE: This demo requires unit ScktComp. You must install the package containing
 this unit before running this demo.}
{NOTE: The password for the private key ring used by TLSChatClientDM is 'abc'.}
unit TLSChatClientMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, ExtCtrls;
   
const
  WM_CLOSE_DELAYED = WM_USER + 1;

type
  TfrmTLSChatClient = class(TForm)
    Panel1: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    edtURL: TEdit;
    edtService: TEdit;
    Button1: TButton;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    Panel2: TPanel;
    btnSend: TButton;
    edtMessage: TEdit;
    edtNick: TEdit;
    Label3: TLabel;
    memChat: TMemo;
    memLog: TMemo;
    btnLoadMyCert: TButton;
    btnLoadMyKey: TButton;
    MyKeyOpenDlg: TOpenDialog;
    MyCertOpenDlg: TOpenDialog;
    Button2: TButton;
    Button3: TButton;
    procedure btnSendClick(Sender: TObject);
    procedure edtMessageKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure btnLoadMyCertClick(Sender: TObject);
    procedure btnLoadMyKeyClick(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
  private
    FMyCertPKOID: string;
    FPKI: string;
    procedure WMCloseDelayed(var Msg: TagMsg); message WM_CLOSE_DELAYED;
  public
    procedure TLSLog(Sender: TObject);
    procedure TLSRead(Sender: TObject);
  end;

var
  frmTLSChatClient: TfrmTLSChatClient;

implementation

uses
  SecUtils, PasswordMain, Pkix, Asn1, MpDL, MpIF, MpEC, X509Base, MpX509,
  TLSChatClientDM;

{$R *.dfm}

{ TftmTLSChatClient }

procedure TfrmTLSChatClient.TLSLog(Sender: TObject);
begin
  memLog.Lines.BeginUpdate;
  try
    memLog.Lines.Assign(dmTLSChatClient.Log);
  finally
    memLog.Lines.EndUpdate;
  end;
end;

procedure TfrmTLSChatClient.TLSRead(Sender: TObject);
begin
  memChat.Lines.Add(dmTLSChatClient.RecvStr);
end;

procedure TfrmTLSChatClient.btnSendClick(Sender: TObject);
begin
  dmTLSChatClient.SendStr(edtNick.Text + ': ' + edtMessage.Text);
  edtMessage.Text := '';
  ActiveControl := edtMessage;
end;

procedure TfrmTLSChatClient.edtMessageKeyDown(Sender: TObject;
  var Key: Word; Shift: TShiftState);
begin
  if Key = VK_RETURN then
    btnSendClick(nil);
end;

procedure TfrmTLSChatClient.btnLoadMyCertClick(Sender: TObject);
var
  FileName: string;
  FS: TFileStream;
  C: TASN1Struct;
  Status: TCertStatusCodes;
  StatusCode: TCertStatusCode;
  OK: Boolean;
  PK: TX509PublicKey;
begin
  if MyCertOpenDlg.Execute then begin
    FileName := MyCertOpenDlg.FileName;
    if LowerCase(ExtractFileExt(FileName)) = '.cer' then begin
      C := TASN1Struct.Create;
      try
        FS := TFileStream.Create(FileName,fmOpenRead);
        try
          C.LoadFromStream(FS,fmtDER);
        finally
          FS.Free;
        end;
        dmTLSChatClient.MyCerts.AddCertificate(C,True,StatusCode);
        OK := StatusCode = crcOK;
      finally
        C.Free;
      end;
    end else
      OK := dmTLSChatClient.MyCerts.ImportFromCMSFile(FileName,True,Status);
    if OK then begin
      C := dmTLSChatClient.MyCerts.Certs[dmTLSChatClient.MyCerts.Count-1];
      FPKI := ExtractSubjectKeyIdentifier(C);
      PK := nil;
      try
        ExtractSubjectPublicKey(C,PK);
        FMyCertPKOID := PK.Algorithm.Algorithm;
      finally
        PK.Free;
      end;
      btnLoadMyKey.Enabled := True;
    end;
  end;
end;

procedure TfrmTLSChatClient.btnLoadMyKeyClick(Sender: TObject);
var
  FileName: string;
  FS: TFileStream;
  IFPriv: TIFPrivateKey;
  DLPriv: TDLPrivateKey;
  Password: string;
  OK: Boolean;
begin
  if MyKeyOpenDlg.Execute then begin
    FileName := MyKeyOpenDlg.FileName;
    if LowerCase(ExtractFileExt(FileName)) = '.spk' then begin
      if FMyCertPKOID = rsaEncryption then begin
        FillChar(IFPriv,SizeOf(IFPriv),0);
        try
          Password := '';
          try
            if PasswordDlg(FileName,Password) then begin
              FS := TFileStream.Create(FileName,fmOpenRead);
              try
                OK := RSAPrivateKeyFromStream(FS,Pointer(Password)^,Length(Password),IFPriv);
              finally
                FS.Free;
              end;
            end else
              OK := False;
          finally
            ProtectClear(Pointer(Password)^,Length(Password));
          end;
          if OK then
            dmTLSChatClient.StreamSecII1.AddPrivateRSAKey(IFPriv,Pointer(FPKI),Length(FPKI),True);
        finally
          DisposeIFPrivateKey(IFPriv);
        end;
      end else if (FMyCertPKOID = id_dsa) or
                  (FMyCertPKOID = dhPublicNumber) then begin
        FillChar(DLPriv,SizeOf(DLPriv),0);
        try
          Password := '';
          try
            if PasswordDlg(FileName,Password) then begin
              FS := TFileStream.Create(FileName,fmOpenRead);
              try
                OK := DLPrivateKeyFromStream(FS,Pointer(Password)^,Length(Password),DLPriv);
              finally
                FS.Free;
              end;
            end else
              OK := False;
          finally
            ProtectClear(Pointer(Password)^,Length(Password));
          end;
          if OK then
            dmTLSChatClient.StreamSecII1.AddPrivateDLKey(DLPriv,
                                                         Pointer(FPKI),Length(FPKI),
                                                         FMyCertPKOID = id_dsa,
                                                         True);
        finally
          DisposeDLPrivateKey(DLPriv);
        end;
      end else
        OK := False;
    end else begin
      FS := TFileStream.Create(FileName,fmOpenRead);
      try
        Password := '';
        try
          if PasswordDlg(FileName,Password) then
            OK := dmTLSChatClient.StreamSecII1.LoadPrivateKeyRingFromStream(FS,Password)
          else
            OK := False;
        finally
          ProtectClear(Pointer(Password)^,Length(Password));
        end;
      finally
        FS.Free;
      end;
    end;
    if not OK then
      raise Exception.Create(FileName + ' was not opened')
    else begin
      FMyCertPKOID := '';
      FPKI := '';
      btnLoadMyKey.Enabled := False;
    end;
  end;
end;

procedure TfrmTLSChatClient.Button2Click(Sender: TObject);
var
  FileName: string;
  FS: TFileStream;
  C: TASN1Struct;
  Status: TCertStatusCodes;
  StatusCode: TCertStatusCode;
  OK: Boolean;
begin
  if MyCertOpenDlg.Execute then begin
    FileName := MyCertOpenDlg.FileName;
    if LowerCase(ExtractFileExt(FileName)) = '.cer' then begin
      C := TASN1Struct.Create;
      try
        FS := TFileStream.Create(FileName,fmOpenRead);
        try
          C.LoadFromStream(FS,fmtDER);
        finally
          FS.Free;
        end;
        dmTLSChatClient.CACerts.AddCertificate(C,True,StatusCode);
        OK := StatusCode = crcOK;
      finally
        C.Free;
      end;
    end else
      OK := dmTLSChatClient.CACerts.ImportFromCMSFile(FileName,True,Status);
    if not OK then
      raise Exception.Create('Illegal format or invalid certificate');
  end;
end;

procedure TfrmTLSChatClient.Button1Click(Sender: TObject);
begin
  dmTLSChatClient.Connect(edtService.Text,edtURL.Text);
end;

procedure TfrmTLSChatClient.WMCloseDelayed(var Msg: TagMsg);
begin
  dmTLSChatClient.ClientSocket1.Close;
end;

procedure TfrmTLSChatClient.Button3Click(Sender: TObject);
begin
  dmTLSChatClient.ClientSocket1.Close;
end;

end.
