unit ECTestMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls;

type
  TForm1 = class(TForm)
    Memo1: TMemo;
    Button1: TButton;
    procedure Button1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses
  MpArithTypes, MpArith, MpECArith, MpEC_NISTCurves, MpYarrow, MpKaratsuba;

{$R *.dfm}

procedure TForm1.Button1Click(Sender: TObject);
var
  EC: TECurve;

  procedure TestECurve(const EC: TECurve);
  var
    X, Y, Z, W: PMPInteger;
    I, QSize: Integer;
    QP0, QP1, QPSum0, QPSum1: Int64;

    function FormatMPInt(X: PMPInteger): string;
    var
      I: Integer;
    begin
      Result := '';
      for I := 0 to X.Size-1 do
        Result := Result + IntToHex(X.Data[I],8) + ' ';
    end;

  begin
    X := nil;
    Y := nil;
    Z := nil;
    W := nil;
    try
      QPSum0 := 0;
      QPSum1 := 0;
      QSize := MPMSB(EC.Q);
      for I := 1 to 1024 do begin
        MPRawRandom(X,QSize*2);
        MPCopy2(X,Y);
        MPCopy2(X,Z);
        MpMod(Y,EC.Q);
        EC.ModProc(Z,EC.Q);
        if MpCmpOffset(Y,Z) <> 0 then begin
          Memo1.Lines.Add('Mod failed');
          Memo1.Lines.Add(FormatMPInt(X));
          Memo1.Lines.Add(FormatMPInt(Y));
          Memo1.Lines.Add(FormatMPInt(Z));
          Abort;
        end;

        MPRawRandom(X,QSize-1);
        MPRawRandom(W,QSize-1);
        QueryPerformanceCounter(QP0);
        MpMulMod2(X,W,EC.Q,Y);
        QueryPerformanceCounter(QP1);
        QPSum0 := QPSum0 + (QP1 - QP0);
        QueryPerformanceCounter(QP0);
        EC.MulModProc(X,W,EC.Q,Z);
        QueryPerformanceCounter(QP1);  
        QPSum1 := QPSum1 + (QP1 - QP0);
        if MpCmpOffset(Y,Z) <> 0 then begin
          Memo1.Lines.Add(Format('MulMod failed after %d attempts',[I]));
          Memo1.Lines.Add(FormatMPInt(X));
          Memo1.Lines.Add(FormatMPInt(W));
          Memo1.Lines.Add(FormatMPInt(Y));
          Memo1.Lines.Add(FormatMPInt(Z));
          Abort;
        end;
      end;
      Memo1.Lines.Add(Format('%d / %d = %d',[QPSum1,QPSum0,Trunc(100*QPSum1/QPSum0)]) + '%');
    finally
      MpDealloc(X);
      MpDealloc(Y);
      MpDealloc(Z);
      MpDealloc(W);
    end;
  end;

begin
  Memo1.Clear;
  FillChar(EC,SizeOf(EC),0);
  try
    ECurveP192(EC);
    TestECurve(EC);
    Memo1.Lines.Add('Prime 192 OK');
     
    ECurveP224(EC);
    TestECurve(EC);
    Memo1.Lines.Add('Prime 224 OK');
                              
    ECurveP256(EC);
    TestECurve(EC);
    Memo1.Lines.Add('Prime 256 OK');
  finally
    MpECArith.ECDealloc(EC);
  end;
end;

end.
