unit SHATestMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
  TForm1 = class(TForm)
    Memo1: TMemo;
    btnTest: TButton;
    btnTestRipeMD: TButton;
    btnHMACTest: TButton;
    btnMD5Test: TButton;
    btnMD2Test: TButton;
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    procedure btnTestClick(Sender: TObject);
    procedure btnTestRipeMDClick(Sender: TObject);
    procedure btnHMACTestClick(Sender: TObject);
    procedure btnMD5TestClick(Sender: TObject);
    procedure btnMD2TestClick(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses
  SecUtils, ASN1, Ssl3;

{$R *.DFM}

// SHA-1 and SHA-2 Test Values from the FIPS draft #2:

procedure TForm1.btnTestClick(Sender: TObject);
var
  Msg, D: string;
begin
  Memo1.Clear;

  // SHA-1

  Msg := 'abc';
  with TSHA1.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('a9993e36 4706816a ba3e2571 7850c26c 9cd0d89d') then
    Memo1.Lines.Add('SHA-1 One Block Message: Good')
  else
    Memo1.Lines.Add('SHA-1 One Block Message: Fail');

  Msg := 'abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq';
  with TSHA1.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('84983e44 1c3bd26e baae4aa1 f95129e5 e54670f1') then
    Memo1.Lines.Add('SHA-1 Multi-Block Message: Good')
  else
    Memo1.Lines.Add('SHA-1 Multi-Block Message: Fail');

  Msg := StringOfChar('a',1000000);
  with TSHA1.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('34aa973c d4c4daa4 f61eeb2b dbad2731 6534016f') then
    Memo1.Lines.Add('SHA-1 Long Message: Good')
  else
    Memo1.Lines.Add('SHA-1 Long Message: Fail');

  Memo1.Lines.Add(Format('Cycles per block: %d',[TSHA1.CyclesOfCompress]));

  // SHA-256

  Msg := 'abc';
  with TSHA256.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('ba7816bf 8f01cfea 414140de 5dae2223 b00361a3 96177a9c b410ff61 f20015ad') then
    Memo1.Lines.Add('SHA-256 One Block Message: Good')
  else
    Memo1.Lines.Add('SHA-256 One Block Message: Fail');

  Msg := 'abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq';
  with TSHA256.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('248d6a61 d20638b8 e5c02693 0c3e6039 a33ce459 64ff2167 f6ecedd4 19db06c1') then
    Memo1.Lines.Add('SHA-256 Multi-Block Message: Good')
  else
    Memo1.Lines.Add('SHA-256 Multi-Block Message: Fail');

  Msg := StringOfChar('a',1000000);
  with TSHA256.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('cdc76e5c 9914fb92 81a1c7e2 84d73e67 f1809a48 a497200e 046d39cc c7112cd0') then
    Memo1.Lines.Add('SHA-256 Long Message: Good')
  else
    Memo1.Lines.Add('SHA-256 Long Message: Fail');

  Memo1.Lines.Add(Format('Cycles per block: %d',[TSHA256.CyclesOfCompress]));

  // SHA-512                                           

  Msg := 'abc';
  with TSHA512.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('ddaf35a193617aba cc417349ae204131 12e6fa4e89a97ea2' +
                 '0a9eeee64b55d39a 2192992a274fc1a8 36ba3c23a3feebbd' +
                 '454d4423643ce80e 2a9ac94fa54ca49f') then
    Memo1.Lines.Add('SHA-512 One Block Message: Good')
  else
    Memo1.Lines.Add('SHA-512 One Block Message: Fail');

  Msg := 'abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijk' +
         'lmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu';
  with TSHA512.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('8e959b75dae313da 8cf4f72814fc143f 8f7779c6eb9f7fa1' +
                 '7299aeadb6889018 501d289e4900f7e4 331b99dec4b5433a' +
                 'c7d329eeb6dd2654 5e96e55b874be909') then
    Memo1.Lines.Add('SHA-512 Multi-Block Message: Good')
  else
    Memo1.Lines.Add('SHA-512 Multi-Block Message: Fail');

  Msg := StringOfChar('a',1000000);
  with TSHA512.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('e718483d0ce76964 4e2e42c7bc15b463 8e1f98b13b204428' +
                 '5632a803afa973eb de0ff244877ea60a 4cb0432ce577c31b' +
                 'eb009c5c2c49aa2e 4eadb217ad8cc09b') then
    Memo1.Lines.Add('SHA-512 Long Message: Good')
  else
    Memo1.Lines.Add('SHA-512 Long Message: Fail');

  // SHA-384

  Msg := 'abc';
  with TSHA384.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('cb00753f45a35e8b b5a03d699ac65007 272c32ab0eded163' +
                 '1a8b605a43ff5bed 8086072ba1e7cc23 58baeca134c825a7') then
    Memo1.Lines.Add('SHA-384 One Block Message: Good')
  else
    Memo1.Lines.Add('SHA-384 One Block Message: Fail');

  Msg := 'abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijk' +
         'lmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu';
  with TSHA384.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('09330c33f71147e8 3d192fc782cd1b47 53111b173b3b05d2' +
                 '2fa08086e3b0f712 fcc7c71a557e2db9 66c3e9fa91746039') then
    Memo1.Lines.Add('SHA-384 Multi-Block Message: Good')
  else
    Memo1.Lines.Add('SHA-384 Multi-Block Message: Fail');

  Msg := StringOfChar('a',1000000);
  with TSHA384.Create(Pointer(Msg)^,Length(Msg)) do begin
    Done(nil);
    D := Digest;
    Free;
  end;
  if D = HexToOS('9d0e1809716474cb 086e834e310a4a1c ed149e9c00f24852' +
                 '7972cec5704c2a5b 07b8b3dc38ecc4eb ae97ddd87f3d8985') then
    Memo1.Lines.Add('SHA-384 Long Message: Good')
  else
    Memo1.Lines.Add('SHA-384 Long Message: Fail');
end;

// RipeMD160 Test Values from the RipeMD160 paper:

const
  RipeMDTestVals: array [0..8,0..1] of string = (
    ('','9c1185a5c5e9fc54612808977ee8f548b2258d31'),
    ('a','0bdc9d2d256b3ee9daae347be6f4dc835a467ffe'),
    ('abc','8eb208f7e05d987a9b044a8e98c6b087f15a0bfc'),
    ('message digest','5d0689ef49d2fae572b881b123a85ffa21595f36'),
    ('abcdefghijklmnopqrstuvwxyz','f71c27109c692c1b56bbdceb5b9d2865b3708dbc'),
    ('abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq',
     '12a053384a9c0c88e405a06c27dcf49ada62eb2b'),
    ('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789',
     'b0e20b6e3116640286ed3a87a5713079b21f5189'),
    ('8 times "1234567890"','9b752e45573d4b39f4dbd3323cab82bf63326bfb'),
    ('1 million "a"','52783243c1697bdbe16d37f97f68f08325dc1528'));

procedure TForm1.btnTestRipeMDClick(Sender: TObject);
var
  I: Integer;
  Msg, D: string;
begin
  Memo1.Clear;

  for I := 0 to 8 do begin
    if I < 7 then
      Msg := RipeMDTestVals[I,0]
    else if I = 7 then
      Msg := '1234567890' +
             '1234567890' +
             '1234567890' +
             '1234567890' +
             '1234567890' +
             '1234567890' +
             '1234567890' +
             '1234567890'
    else if I = 8 then
      Msg := StringOfChar('a',1000000);
    with TRipeMD160.Create(Pointer(Msg)^,Length(Msg)) do begin
      Done(nil);
      D := Digest;
      Free;
    end;
    if D = HexToOS(RipeMDTestVals[I,1]) then
      Memo1.Lines.Add(Format('RipeMD160 %d: Good',[I]))
    else
      Memo1.Lines.Add(Format('RipeMD160 %d: Fail',[I]));
  end;
end;

// HMAC Test Values from the FIPS draft:

procedure TForm1.btnHMACTestClick(Sender: TObject);
var
  Key, Msg, H: string;
begin
  SetLength(H,20);

  Key := HexToOS('0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b');
  Msg := 'Hi There';
  HMac(Pointer(Key)^,Length(Key),
       Pointer(Msg)^,Length(Msg),
       haSHA1,
       Pointer(H)^,Length(H));
  if H = HexToOS('b617318655057264e28bc0b6fb378c8ef146be00') then
    Memo1.Lines.Add('HMAC SHA-1 Test Case 1: Good')
  else
    Memo1.Lines.Add('HMAC SHA-1 Test Case 1: Fail');

  Key := HexToOS('aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa');
  Msg := StringOfChar(#$dd,50);
  HMac(Pointer(Key)^,Length(Key),
       Pointer(Msg)^,Length(Msg),
       haSHA1,
       Pointer(H)^,Length(H));
  if H = HexToOS('125d7342b9ac11cd91a39af48aa17b4f63f175d3') then
    Memo1.Lines.Add('HMAC SHA-1 Test Case 2: Good')
  else
    Memo1.Lines.Add('HMAC SHA-1 Test Case 2: Fail');

  Key := HexToOS('0102030405060708090a0b0c0d0e0f10111213141516171819');
  Msg := StringOfChar(#$cd,50);
  HMac(Pointer(Key)^,Length(Key),
       Pointer(Msg)^,Length(Msg),
       haSHA1,
       Pointer(H)^,Length(H));
  if H = HexToOS('4c9007f4026250c6bc8414f9bf50c86c2d7235da') then
    Memo1.Lines.Add('HMAC SHA-1 Test Case 3: Good')
  else
    Memo1.Lines.Add('HMAC SHA-1 Test Case 3: Fail');

  Key := HexToOS('0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c');
  Msg := 'Test With Truncation';
  HMac(Pointer(Key)^,Length(Key),
       Pointer(Msg)^,Length(Msg),
       haSHA1,
       Pointer(H)^,Length(H));
  if H = HexToOS('4c1a03424b55e07fe7f27be1d58bb9324a9a5a04') then
    Memo1.Lines.Add('HMAC SHA-1 Test Case 4: Good')
  else
    Memo1.Lines.Add('HMAC SHA-1 Test Case 4: Fail');

  SetLength(H,12);
  Key := HexToOS('0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c');
  Msg := 'Test With Truncation';
  HMac(Pointer(Key)^,Length(Key),
       Pointer(Msg)^,Length(Msg),
       haSHA1,
       Pointer(H)^,Length(H));
  if H = HexToOS('4c1a03424b55e07fe7f27be1') then
    Memo1.Lines.Add('HMAC SHA-1 Test Case 4 Trunc: Good')
  else
    Memo1.Lines.Add('HMAC SHA-1 Test Case 4 Trunc: Fail');
end;

// Test values from rfc1321:
const
  MD5TestVals: array [0..6,0..1] of string = (
    ('','d41d8cd98f00b204e9800998ecf8427e'),
    ('a','0cc175b9c0f1b6a831c399e269772661'),
    ('abc','900150983cd24fb0d6963f7d28e17f72'),
    ('message digest','f96b697d7cb7938d525a2f31aaf161d0'),
    ('abcdefghijklmnopqrstuvwxyz','c3fcd3d76192e4007dfb496cca67e13b'),
    ('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789',
     'd174ab98d277d9f5a5611c2c9f419d9f'),
    ('123456789012345678901234567890123456789012345678901234567890123456' +
     '78901234567890',
     '57edf4a22be3c955ac49da2e2107b67a'));

procedure TForm1.btnMD5TestClick(Sender: TObject);
var
  I: Integer;
  Msg, D: string;
begin
  Memo1.Clear;

  for I := 0 to 6 do begin
    Msg := MD5TestVals[I,0];
    with TMD5.Create(Pointer(Msg)^,Length(Msg)) do begin
      Done(nil);
      D := Digest;
      Free;
    end;
    if D = HexToOS(MD5TestVals[I,1]) then
      Memo1.Lines.Add(Format('MD5 %d: Good',[I]))
    else
      Memo1.Lines.Add(Format('MD5 %d: Fail',[I]));
  end;
end;

// Test values from rfc1319:
const
  MD2TestVals: array [0..6,0..1] of string = (
    ('','8350e5a3e24c153df2275c9f80692773'),
    ('a','32ec01ec4a6dac72c0ab96fb34c0b5d1'),
    ('abc','da853b0d3f88d99b30283a69e6ded6bb'),
    ('message digest','ab4f496bfb2a530b219ff33031fe06b0'),
    ('abcdefghijklmnopqrstuvwxyz','4e8ddff3650292ab5a4108c3aa47940b'),
    ('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789',
     'da33def2a42df13975352846c30338cd'),
    ('123456789012345678901234567890123456789012345678901234567890123456' +
     '78901234567890',
     'd5976f79d83d3a0dc9806c3c66f3efd8'));

procedure TForm1.btnMD2TestClick(Sender: TObject);
var
  I: Integer;
  Msg, D: string;
begin
  Memo1.Clear;

  for I := 0 to 6 do begin
    Msg := MD2TestVals[I,0];
    with TMD2.Create(Pointer(Msg)^,Length(Msg)) do begin
      Done(nil);
      D := Digest;
      Free;
    end;
    if D = HexToOS(MD2TestVals[I,1]) then
      Memo1.Lines.Add(Format('MD2 %d: Good',[I]))
    else
      Memo1.Lines.Add(Format('MD2 %d: Fail',[I]));
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  Msg, D: string;
begin                     
  Memo1.Clear;

  if SameText(OSToHex(DigestString(haSHA224,''),True),
              'd14a028c2a3a2bc9476102bb288234c415a2b01f828ea62ac5b3e42f') then
    Memo1.Lines.Add('Test 1 good')
  else
    Memo1.Lines.Add('Test 1 fail');

  if SameText(OSToHex(DigestString(haSHA224,'a'),True),
              'abd37534c7d9a2efb9465de931cd7055ffdb8879563ae98078d6d6d5') then
    Memo1.Lines.Add('Test 2 good')
  else
    Memo1.Lines.Add('Test 2 fail');

  if SameText(OSToHex(DigestString(haSHA224,'abc'),True),
              '23097d223405d8228642a477bda255b32aadbce4bda0b3f7e36c9da7') then
    Memo1.Lines.Add('Test 3 good')
  else
    Memo1.Lines.Add('Test 3 fail');

  if SameText(OSToHex(DigestString(haSHA224,'message digest'),True),
              '2cb21c83ae2f004de7e81c3c7019cbcb65b71ab656b22d6d0c39b8eb') then
    Memo1.Lines.Add('Test 4 good')
  else
    Memo1.Lines.Add('Test 4 fail');

  if SameText(OSToHex(DigestString(haSHA224,'abcdefghijklmnopqrstuvwxyz'),True),
              '45a5f72c39c5cff2522eb3429799e49e5f44b356ef926bcf390dccc2') then
    Memo1.Lines.Add('Test 5 good')
  else
    Memo1.Lines.Add('Test 5 fail');

  if SameText(OSToHex(DigestString(haSHA224,'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'),True),
              'bff72b4fcb7d75e5632900ac5f90d219e05e97a7bde72e740db393d9') then
    Memo1.Lines.Add('Test 6 good')
  else
    Memo1.Lines.Add('Test 6 fail');

  if SameText(OSToHex(DigestString(haSHA224,'12345678901234567890123456789012345678901234567890123456789012345678901234567890'),True),
              'b50aecbe4e9bb0b57bc5f3ae760a8e01db24f203fb3cdcd13148046e') then
    Memo1.Lines.Add('Test 7 good')
  else
    Memo1.Lines.Add('Test 7 fail');

  if SameText(OSToHex(DigestString(haSHA224,'abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq'),True),
              '75388b16512776cc5dba5da1fd890150b0c6455cb4f58b1952522525') then
    Memo1.Lines.Add('Test 8 good')
  else
    Memo1.Lines.Add('Test 8 fail');

  if SameText(OSToHex(DigestString(haSHA224,StringOfChar('a',1000000)),True),
              '20794655980c91d8bbb4c1ea97618a4bf03f42581948b2ee4ee7ad67') then
    Memo1.Lines.Add('Test 9 good')
  else
    Memo1.Lines.Add('Test 9 fail');

  if SameText(OSToHex(DigestString(haSHA224,'abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu'),True),
              'c97ca9a559850ce97a04a96def6d99a9e0e0e2ab14e6b8df265fc0b3') then
    Memo1.Lines.Add('Test 10 good')
  else
    Memo1.Lines.Add('Test 10 fail');

  if SameText(OSToHex(DigestString(haSHA224,'abcdbcdecdefdefgefghfghighijhijk'),True),
              'd7d147b9b2bc45479a494087ef8d212ef37910aa52462cce9c4dc8b7') then
    Memo1.Lines.Add('Test 11 good')
  else
    Memo1.Lines.Add('Test 11 fail');

  if SameText(OSToHex(DigestString(haSHA224,'abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu'),True),
              'c97ca9a559850ce97a04a96def6d99a9e0e0e2ab14e6b8df265fc0b3') then
    Memo1.Lines.Add('Test 12 good')
  else
    Memo1.Lines.Add('Test 12 fail');

  if SameText(OSToHex(DigestString(haSHA224,
                                   HexToOS('5d5e5f58595a5b54555657505152534c4d4e4f48494a4b44454647404142437c7d7e7f78797a' +
                                           '7b74757677707172736c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c' +
                                           '5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c5c' +
                                           '5c5c5c5c5c5c5c5c5c5c5c5c5c5c4f5ab933cac0555e3bd5e08f9fd0729ff44e402bc3a38092' +
                                           '4324047a2ef8f8ef6601120d40a1dc8f830e73cffd2e6f9d0562fb4ffe1f5010194b654be2fb' +
                                           'a5d4')),True),
              '80d14d51c69ccec49982f54ebd86f482bf2aedbc2ce9864e5e470c89') then
    Memo1.Lines.Add('Test 13 good')
  else
    Memo1.Lines.Add('Test 13 fail');
end;

procedure TForm1.Button2Click(Sender: TObject);
var
  HA: THashAlgorithm;
  Key, Msg, H: string;
begin
  for HA := Low(HA) to High(HA) do begin
    SetLength(H,FindHashClass(HA).DigestSize);
    Memo1.Lines.Add(FindHashClass(HA).ClassName);

    Key := HexToOS('0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b');
    Msg := 'Hi There';
    HMac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + ''',');

    Key := HexToOS('aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa');
    Msg := StringOfChar(#$dd,50);
    HMac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + ''',');

    Key := HexToOS('0102030405060708090a0b0c0d0e0f10111213141516171819');
    Msg := StringOfChar(#$cd,50);
    HMac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + ''',');

    Key := HexToOS('0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c');
    Msg := 'Test With Truncation';
    HMac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + ''',');

    SetLength(H,12);
    Key := HexToOS('0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c');
    Msg := 'Test With Truncation';
    HMac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + '''');
    Memo1.Lines.Add('');
  end;
end;

procedure SSL3MAC(const K; KLen: Integer;
                  const M; MLen: Integer;
                  HashAlgorithm: THashAlgorithm;
                  var Mac; MacLen: Integer);
var
  HC: THashClass;
  H: THash;
begin
  HC := FindHashClass(HashAlgorithm);
  if HC = nil then
    raise Exception.Create('Algorithm not supported');
  H := HC.Create(nil^,0);
  try
    SSL3MACBegin(K,KLen,H,False);
    H.HashData(M,MLen);
    SSL3MACEnd(K,KLen,H,Mac,MacLen);
  finally
    H.Free;
  end;
end;

procedure TForm1.Button3Click(Sender: TObject);
var
  HA: THashAlgorithm;
  Key, Msg, H: string;
begin
  for HA := haSHA1 to haMD5 do begin
    SetLength(H,FindHashClass(HA).DigestSize);
    Memo1.Lines.Add(FindHashClass(HA).ClassName);

    Key := HexToOS('0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b');
    Msg := 'Hi There';
    SSL3Mac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + ''',');

    Key := HexToOS('aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa');
    Msg := StringOfChar(#$dd,50);
    SSL3Mac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + ''',');

    Key := HexToOS('0102030405060708090a0b0c0d0e0f10111213141516171819');
    Msg := StringOfChar(#$cd,50);
    SSL3Mac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + ''',');

    Key := HexToOS('0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c');
    Msg := 'Test With Truncation';
    SSL3Mac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + ''',');

    SetLength(H,12);
    Key := HexToOS('0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c');
    Msg := 'Test With Truncation';
    SSL3Mac(Pointer(Key)^,Length(Key),
         Pointer(Msg)^,Length(Msg),
         HA,
         Pointer(H)^,Length(H));
    Memo1.Lines.Add('    ''' + OSToHex(H,True) + '''');
    Memo1.Lines.Add('');
  end;
end;

end.
